/*
 * Copyright (c) 2010 LAAS/CNRS                       --  Oct 2011
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 *
 *  Codels called by execution task lwrCmd
 *  File: lwrCmdCodels.cc
 *  Description:
 *  Creation date : Wed May 19 22:35:29 2010 
 *  Author: Xavier Broquere
 *  LAAS-CNRS, Toulouse FRANCE
 *------------------------------------------------------------*/

#ifdef LWR_FRI
#include "kukaFri-libs/friremote.h"
#endif
#include "server/lwrHeader.h"
extern "C" {
#include "Proto_gb.h"
#include "Proto_gbModeles.h"   // LWR MGI
}
#include "libeuler/pomEuler.h"
#include "coldman.h"
#include "softMotion/softMotion.h"
#include "softMotion/softMotionStruct.h"
#include "softMotion/softMotionStructGenom.h"
#include <pthread.h>
#include <sys/time.h>
#include <taskLib.h>
#include <h2timerLib.h>

#include "lwrConst.h"
#include "lwrFunctions.h"

#define MAX(a,b) ((a)>(b)?(a):(b)) //DScomment is it necessary to define MAX ??? It must be define elsewhere. Only used one time.


#ifdef LWR_FRI
friRemote * friInstLeft = NULL;
friRemote * friInstRight = NULL;
int lwrGetArmInf(friRemote *friInst, LWR_ARM_INST *arm, char* armName);
int lwrGetInf();
int lwrDoDataExchange();
#endif

static int LWR_NBJOINTS_ARMS_USER; // Number of dof used to compute the traj in articular space
static int LWR_CART_USER; // Number of dof used to compute the traj in cartesian space
static int LWR_NB_CONTROLLED_JNTS; // this is the number of controlled joint automatically computed in the Init request
static bool initIsDone = false;
static bool stopFriThread = false; // flag to exit the fri thread


static SM_LIMITS limitsTimeScale;
LWR_MOTION_STR currentMotion;

SM_TRAJ trajBuff1;
SM_TRAJ trajBuff2;
static int buffInUse=1;

SM_TRAJ_STR smTraj;

GENMANIP_POSITION_ARM_SAHAND_TRACK_STR  lwrPositionsPoster;

float  initialCartPos[LWR_CART_FRM_DIM];
static POSTER_ID lwr_poster_q = NULL;
static POSTER_ID posterMhpCollideId = NULL;
static POSTER_ID posterSAHandId = NULL;
static POSTER_ID posterHumDistId = NULL;

static FILE* fileptr = NULL;
static GEN_BOOL closingSaveFile = GEN_FALSE;

STATUS lwrCmd(); // main function

int lwrMonitorAndUpdateCntrlMode(LWR_ARM_INST *armInst, char* message) ;
int lwrDetectChock(LWR_ARM_INST *armInst, double TC, double threshold) ;

#ifdef LWR_FRI
static H2TIMER_ID threadFriCommTimerId;

int threadFriCommCode()
{
  if (h2timerPause(threadFriCommTimerId) != OK) {
    printf("threadFriCommCode: h2timerPause error\n");
    //lwrCmdSuspend (FALSE);
  }
  if(SDI_F->status.simulation == GEN_TRUE) {    
    lwrCmd();
  } else {
    if(stopFriThread) {  //?? move to reat time
      return OK;
    }
    if ((SDI_F->armLeft.is_used == GEN_TRUE) && (friInstLeft == NULL)) {
      return OK;
    }
    if((SDI_F->armRight.is_used == GEN_TRUE) && (friInstRight == NULL)) {
      return OK;
    }
    lwrCmd(); // Main function (collision checking, comptute the next command ...)
    lwrDoDataExchange();
  }
  // WARNING  names ("R7ArmLeft") come from jido.geom for POM...
  lwrUpdateArmStatus(friInstLeft , &SDI_F->armLeft , &SDI_F->pomArmLeftPos , (char *)"R7ArmLeft"); //?? move to reat time
  lwrUpdateArmStatus(friInstRight, &SDI_F->armRight, &SDI_F->pomArmRightPos, (char *)"R7ArmRight");//?? move to reat time
  lwrGetInf();  //?? move real time
  lwrUpdateMotion();  //?? move real time
  return 0;
}

void *threadFriCommFct(void *) {
  do{
    threadFriCommCode();
  } while(stopFriThread == false);
  printf("threadFriComm comm exited\n");
  return 0;
}
#endif

int lwrGetInf() {
#ifdef LWR_FRI
  /************************/
  /* Info about LEFT ARM  */
  /************************/
  if(SDI_F->armLeft.is_used == GEN_TRUE && friInstLeft != NULL) {
    lwrGetArmInf(friInstLeft, &(SDI_F->armLeft), (char *)"Left");
  }    
  /*************************/
  /* Info about RIGHT ARM  */
  /*************************/
  if(SDI_F->armRight.is_used ==  GEN_TRUE && friInstRight != NULL) {
    lwrGetArmInf(friInstRight, &(SDI_F->armRight), (char *)"Right");
  }
#endif
  return 0;
}

#ifdef LWR_FRI

friRemote * getFriRemote(LWR_ARM_ID_ENUM arm_id) {
  if(arm_id == LWR_ARM_LEFT) {
    return friInstLeft;
  } else if (arm_id == LWR_ARM_RIGHT) {
    return friInstRight;
  } else {
    printf("ERROR: getFriRemote wrong arm id: %d\n", arm_id);
    return NULL;
  }
}

// send and receive data, communication with fri
// this function wait for both arms if they are both activated
int lwrDoDataExchange() {
  bool flagLeft = false, flagRight = false, flag = false;
  int sockLeft = 0, sockRight = 0;
  fd_set rfds;
  int retval = -1;

  if(friInstLeft == NULL && SDI_F->armLeft.is_used == GEN_TRUE) {
    return 0;
  }
  if( friInstRight == NULL && SDI_F->armRight.is_used == GEN_TRUE ) {
    return 0;
  }
  if( SDI_F->armLeft.is_used == GEN_TRUE &&  SDI_F->armRight.is_used == GEN_TRUE) {
    do{    
      FD_ZERO(&rfds);
      FD_SET(SDI_F->armLeft.sock, &rfds);
      FD_SET(SDI_F->armRight.sock, &rfds);    
      retval = select( MAX(SDI_F->armLeft.sock, SDI_F->armRight.sock)+1 , &rfds, NULL, NULL, NULL);
      if (retval == -1) {
        perror("select()");
      } else if (retval) {
        if(FD_ISSET(SDI_F->armLeft.sock, &rfds)){
          flagLeft = true;
          friInstLeft->doDataExchange();
        }
        if(FD_ISSET(SDI_F->armRight.sock, &rfds)){
          flagRight = true;
          friInstRight->doDataExchange();
        }
      }
    } while(flagLeft==false || flagRight==false );

  } else if ( SDI_F->armLeft.is_used == GEN_TRUE) {
    friInstLeft->doDataExchange();
  } else if ( SDI_F->armRight.is_used == GEN_TRUE) {
    friInstRight->doDataExchange();
  } else {
    printf("WARNING: lwrDoDataExchange: Any arm is used\n");
  }
  return 0;
}

// param
// pointer of the FRI struct, pointer of the arm struct in genom, arm name
// import data from the friremote class into the genom poster, structures ...
int lwrGetArmInf(friRemote *friInst, LWR_ARM_INST *arm, char* armName) {
  GEN_BOOL power_on;
  // Arm power on/off
  power_on = friInst->isPowerOn()?GEN_TRUE:GEN_FALSE;
  if (arm->power_on != power_on) {
    if (friInst->isPowerOn() == false) {
      printf("WARN: Arm %s power is off\n",armName);
    }
    arm->power_on = power_on;
  }
  //  FRI mode monitor/command
  if (arm->fri_state != friInst->getState()) {
    if (friInst->getState() != FRI_STATE_CMD) {
      printf("WARN: FRI %s is in monitor mode\n",armName);
    }
    arm->fri_state = friInst->getState();
  }
  // Quality change: report as a warning as it can imply errors on following commands
  if (friInst->getQuality() != arm->fri_quality) {
    printf("%s: Arm %s Communication quality has %s (new value is %d)\n",
           (friInst->getQuality() > arm->fri_quality)?"INFO":"WARN", armName,
           (friInst->getQuality() > arm->fri_quality)?"improved":"worsened",
           friInst->getQuality());
    arm->fri_quality = friInst->getQuality();
  }

  memcpy(arm->estExtTcpFT,  friInst->getMsrEstExtTcpFT(), sizeof(float)*FRI_CART_VEC);
  memcpy(arm->estExtJntTrq,  friInst->getMsrEstExtJntTrq(), sizeof(float)*LWR_JOINTS);
  memcpy(arm->msrJntTrq,  friInst->getMsrJntTrq(), sizeof(float)*LWR_JOINTS);


  /* */

  //TODO use c++ class deque to store the history (for example)
  for(int v=0; v<(JNT_TRQ_NB_TIC_MEAN-1); v++) {
    for(int w=0; w< LWR_JOINTS; w++) {  
      arm->estExtJntTrqHistory[w][v] = arm->estExtJntTrqHistory[w][v+1];
    }
  }
  for(int w=0; w< LWR_JOINTS; w++) {  
    arm->estExtJntTrqHistory[w][JNT_TRQ_NB_TIC_MEAN-1] = arm->estExtJntTrq[w];
  }

  for(int v=0; v< LWR_JOINTS; v++) { 
    arm->estMeanExtJntTrq[v] =  0.0;
    for(int w=0; w<JNT_TRQ_NB_TIC_MEAN; w++) { 
      arm->estMeanExtJntTrq[v] +=  arm->estExtJntTrqHistory[v][w];
    }
    arm->estMeanExtJntTrq[v] = arm->estMeanExtJntTrq[v] / (double)JNT_TRQ_NB_TIC_MEAN;
  }
  /* */
  for(int v=0; v<(LWR_FORCES_WINDOW-1); v++) {
    for(int u=0; u < LWR_CART_VEC; u++) {
      arm->estExtTcpFTHistory[v][u] = arm->estExtTcpFTHistory[v+1][u];
    }
  }
  for(int w=0; w< LWR_CART_VEC; w++) {  
    arm->estExtTcpFTHistory[LWR_FORCES_WINDOW-1][w] = arm->estExtTcpFT[w];
  }
  /* */



  //  FRI command scheme POSITION/CART_IMP/JOINT_IMP/GRAV_COMP
  arm->fri_scheme = (LWR_CTRL_SCHEME_ENUM)friInst->getCurrentControlScheme();
  arm->timeStamp = friInst->getTimestamp();
  return 0;
}

ACTIVITY_EVENT
requestFriStateCmd(int *report, LWR_ARM_ID_ENUM arm_id,  LWR_CTRL_SCHEME_ENUM ctrlScheme)
{
  static int iter = 0;
  static int max_iter = MAX_REQ_CMD;

  friRemote * friInst = getFriRemote(arm_id);
  if(friInst == NULL) {
    printf("ERROR requestFriStateCmd: friInst == NULL\n");
    return ETHER;
  }

  if (SDI_F->status.simulation == GEN_TRUE) {
    /* No verifications required on simulation mode */
    return EXEC;
  }

  if (friInst->getState() != FRI_STATE_CMD)
    syncFriWithMeasuredPose(friInst, ctrlScheme);

  if (iter == max_iter) {
    /* FRI doesn't enter in command state after enough tries*/
    friInst->setToKRLInt(CMD_CHANNEL, FRI_IDLE);
    printf("ERROR: FRI doesn't start after %d iterations on arm %d \n", iter, arm_id);
    *report = S_lwr_NOT_INITIALIZED;
    iter = 0;
    return ETHER;
  }
  else if (ctrlScheme != (LWR_CTRL_SCHEME_ENUM)friInst->getCurrentControlScheme()) {
    /* Not in desired control scheme; request an scheme change */
    if (iter == 0) {
      max_iter = 3*MAX_REQ_CMD;    // provide extra time for the switch
      printf("WARN: Switching control scheme to %s. This can take several seconds\n",
             friSchemeStr(ctrlScheme));
    }
    
    switchCtrlScheme(friInst, ctrlScheme);
    
    iter++;
    return START;
  }
  else if (friInst->isPowerOn() == false) {
    /* Arm power is off; request switching it on */
    friInst->setToKRLInt(CMD_CHANNEL, POWER_ON);
    iter++;
    return START;
  }
  else if (friInst->getState() != FRI_STATE_CMD) {
    /* FRI is still not ready to accept commands; retry */
    friInst->setToKRLInt(CMD_CHANNEL, FRI_START);
    iter++;
    return START;
  }
  else {
    /* all ready; proceed to send command */
    friInst->setToKRLInt(CMD_CHANNEL, FRI_IDLE);
    *report = OK;
    iter = 0;
    return EXEC;
  }
}

ACTIVITY_EVENT
requestFriStateIdle(LWR_ARM_ID_ENUM arm_id, int *report)
{
  friRemote * friInst = getFriRemote(arm_id);
  if(friInst == NULL) {
    return ETHER;
  }

  /* Force entering on idle state */
  if(friInst->getFrmKRLInt(CMD_CHANNEL) != FRI_IDLE) {
    friInst->setToKRLInt(CMD_CHANNEL, FRI_IDLE);
    return END;
  }

  return ETHER;
}

#endif  /* ifndef LWR_FRI */

/*------------------------------------------------------------------------
 *
 * lwrCmdEnd  --  Termination codel
 *
 * Description: 
 * 
 * Returns:    OK or ERROR
 */
STATUS
lwrCmdEnd(void)
{
#ifdef LWR_FRI
  if(SDI_F->armLeft.is_used == GEN_TRUE) {
    printf("CLOSING CONNECTION FOR LWR LEFT ... \n");
    /* Send close connection code to KRL */
    //do {
    //  friInstLeft->setToKRLInt(CMD_CHANNEL, FRI_CLOSE);
    //  usleep(friInstLeft->getSampleTime()*1000);
    //}
    //while (friInstLeft->getFrmKRLInt(CMD_CHANNEL) != FRI_CLOSE);

    stopFriThread = true;
    //usleep(friInstLeft->getSampleTime()*2000);
    delete friInstLeft;
    friInstLeft = NULL;
    printf(" LWR LEFT CLOSED!\n");
  } 
  if(SDI_F->armRight.is_used == GEN_TRUE) {
    printf("CLOSING CONNECTION FOR LWR RIGHT IN ... \n");
    /* Send close connection code to KRL */
    //do {
    //  friInstRight->setToKRLInt(CMD_CHANNEL, FRI_CLOSE);
    //  usleep(friInstRight->getSampleTime()*1000);
    //}
    //while (friInstRight->getFrmKRLInt(CMD_CHANNEL) != FRI_CLOSE);

    stopFriThread = true;
    //usleep(friInstRight->getSampleTime()*2000);
    delete friInstRight;
    friInstRight = NULL;
    printf(" LWR RIGHT CLOSED!\n");
  }
  // if(threadFriComm) {
  //   pthread_exit();
  // }
#endif
  return OK;
}

/*------------------------------------------------------------------------
 *
 * lwrCmdPerm  --  1st codel of permanent activity (called BEFORE other activities)
 *
 * Description: 
 * 
 * Reports:      OK
 * 
 * Returns:    OK or ERROR
 */
static FILE* file = 0;
int pbFriCompteur = 0;

STATUS
lwrCmdPerm(int *report)
{
  if (!initIsDone) {
    return OK;
  }

#ifdef LWR_FRI
  if(closingSaveFile == GEN_TRUE) {
    closingSaveFile = GEN_FALSE;
    if(fileptr) {
      fclose(fileptr);
      printf("File lwrMotion.dat is written !\n");
      fileptr = NULL;
    }
  }
 // the command is done in the fri tread
#else
  // simulation mode
  lwrCmd();
  lwrUpdateArmStatus( &SDI_F->armLeft , &SDI_F->pomArmLeftPos , (char *)"R7ArmLeft");
  lwrUpdateArmStatus( &SDI_F->armRight, &SDI_F->pomArmRightPos, (char *)"R7ArmRight");
  lwrGetInf();
  lwrUpdateMotion();
#endif
  // generate the string for coldman-genom
  lwrGenerateDigestStatus();  // not real time. But Cmd is not real time
  return OK;
}

STATUS lwrCmd() {

  bool flagLeft = false;
  bool flagRight = false;

  if(SDI_F->armLeft.cntrl_mode == LWR_XOPEN_LOOP_CNTRL && SDI_F->armRight.cntrl_mode == LWR_XOPEN_LOOP_CNTRL) {
    currentMotion.duration = currentMotion.traj_x.getDuration();
  } else {
       currentMotion.duration = currentMotion.traj.getDuration();
  }

  if((posterHumDistId != NULL) && (SDI_F->humanDistMode == GEN_TRUE)) {
    if(lwrGENHUM_HUMAN_DISTANCEPosterRead(posterHumDistId, &SDI_F->humanDist) == ERROR) {
      printf("lwrGENHUM_HUMAN_DISTANCEPosterRead CANNOT read poster\n");
    } else {
      //SDI_F->timeScale = 1.0 - SDI_F->humanDist.costDistRobToHum;
      SDI_F->timeScale = SDI_F->humanDist.costDistRobToHum;
    }
  }

  // currentMotion is the global structure of the curent motion to execute
  if (currentMotion.type != LWR_NOT_MOVE) {

    if (currentMotion.type != LWR_NOT_MOVE) {
      if(SDI_F->saveMotionFile == GEN_TRUE) {
        // motion file is only for debug
        if(currentMotion.curTime == 0.0) {
          if ((fileptr = fopen("lwrMotion.dat","wb"))==NULL) {
            printf("ERROR Open File lwrMotion.dat\n");
          }
        } /*  if(currentMotion.curTime == 0.0) */
      } /* if(SDI_F->saveMotionFile == GEN_TRUE) */
    }

    /* check for predicted auto collision with colman library */
    if (SDI_F->status.collisionMode == GEN_TRUE) {
      for (int i = 0; i < SDI_F->motion.size; i++) {
        if (coldmanSetRobotQi(i, SDI_F->motion.q[i],
                              SDI_F->motion.vq[i],
                              SDI_F->motion.aq[i]) != 0) {
          printf("ERROR coldmanSetRobotQi on axis %d\n",i);
	  setMotionError(LWR_OUT_OF_BOUNDS);   // this will set state as LWR_ABORTED
        }
      }

      /* predict collision within 0.03 sec */
      if (coldmanPredictCollision(0.03) == 1) {
        printf("WARN: Coldman has reported collision risk. Motion aborted\n");
        setMotionError(LWR_COLLISION_RISK);   // this will set state as LWR_ABORTED
      } 
    } /*  (SDI_F->status.collisionMode == GEN_TRUE) */


    if(SDI_F->status.collisionEnvMode == GEN_TRUE) {	  
      if(lwrGENMANIP_COLLISION_STATE_STRPosterRead(posterMhpCollideId, &SDI_F->collisionEnvState) == ERROR) {
        printf("Collision reported cannot read poster of mhp \n");
      } else {
        if(SDI_F->collisionEnvState.state == GENMANIP_COLLISION_COLLISION) {
          printf("Collision reported by env module: lwr stoped \n");
          //setMotionError(LWR_COLLISION_RISK);
          setMotionEnd();
        }
      }
    } /*  (SDI_F->status.collisionEnvMode == GEN_TRUE) */

    // Add here your custom watch


#ifndef LWR_FRI
    // simulation
    if(SDI_F->armLeft.is_used == GEN_TRUE){	
      lwrCommandArm(SDI_F->isTrqCntrlSofMotion, LWR_ARM_LEFT, &SDI_F->armLeft, &(SDI_F->motion), &currentMotion);
    }
    if(SDI_F->armRight.is_used == GEN_TRUE){
      lwrCommandArm(SDI_F->isTrqCntrlSofMotion, LWR_ARM_RIGHT, &SDI_F->armRight, &(SDI_F->motion), &currentMotion);
    }
    
#else
    if(SDI_F->armLeft.is_used == GEN_TRUE) {
      if(friInstLeft->getState() != FRI_STATE_CMD) {
        flagLeft = true;
      }
    } else {
      flagLeft = true;
    }

    if(SDI_F->armRight.is_used == GEN_TRUE) {
      if(friInstRight->getState() != FRI_STATE_CMD) {
        flagRight = true;
      }
    } else {
      flagRight = true;
    }
    
    if ( flagRight && flagLeft && (SDI_F->status.simulation != GEN_TRUE)) {
      // The LWR has revoqued the FRI command state due to an exceptional situation
      printf("ERROR: FRI command state revoqued. Motion aborted flagRight: %d flagLeft: %d\n", flagRight, flagLeft);
      printf("curent currentMotion.curTime %f currentMotion.duration %f\n", currentMotion.curTime, currentMotion.duration);
      setMotionError(LWR_FRI_MONITOR_MODE);
    }
    else {


      if(SDI_F->armLeft.is_used == GEN_TRUE){	
        lwrCommandArm(friInstLeft, SDI_F->status.simulation, SDI_F->isTrqCntrlSofMotion, LWR_ARM_LEFT, &SDI_F->armLeft, &(SDI_F->motion), &currentMotion);
      }

      if(SDI_F->armRight.is_used == GEN_TRUE){
        lwrCommandArm(friInstRight, SDI_F->status.simulation, SDI_F->isTrqCntrlSofMotion, LWR_ARM_RIGHT, &SDI_F->armRight, &(SDI_F->motion), &currentMotion);
      }


#endif
      if(SDI_F->saveMotionFile == GEN_TRUE) {
        /*save motion into the file for debug */
        if(fileptr) {
          fprintf(fileptr,"%f %f %f ", currentMotion.curTime, SDI_F->armRight.timeStamp, SDI_F->humanDist.costDistRobToHum);
          for(int u=0; u<LWR_NBJOINTS_ARMS_USER; u++) {
            // if(u < LWR_JOINTS) {
            //   fprintf(fileptr,"%f %f %f %f %f %f %f %f ", SDI_F->armLeft.estMeanExtJntTrq[u],
            //           SDI_F->armLeft.currentJntVals[u],  SDI_F->armLeft.currentMsrJntVals[u],
            //           SDI_F->motion.vq[u],  SDI_F->motion.aq[u],
            //           SDI_F->armLeft.currentCmdVelJntVals[u],SDI_F->armLeft.currentCmdAccJntVals[u],
            //           currentMotion.q_V_max[u]);
            // } else {
            //   fprintf(fileptr,"%f %f %f %f %f %f %f %f ", SDI_F->armRight.estMeanExtJntTrq[u-LWR_JOINTS],
            //           SDI_F->armLeft.currentJntVals[u],  SDI_F->armLeft.currentMsrJntVals[u],
            //           SDI_F->motion.vq[u],  SDI_F->motion.aq[u],
            //           SDI_F->armRight.currentCmdVelJntVals[u-LWR_JOINTS], SDI_F->armRight.currentCmdAccJntVals[u-LWR_JOINTS],
            //           currentMotion.q_V_max[u]);
            // }

            //save the motion data for only the right arm
            if(u >= LWR_JOINTS) {
              fprintf(fileptr,"%f %f %f %f %f %f %f %f ", SDI_F->armRight.estMeanExtJntTrq[u-LWR_JOINTS],
                      SDI_F->armRight.currentJntVals[u-LWR_JOINTS],  SDI_F->armRight.currentMsrJntVals[u-LWR_JOINTS],
                      SDI_F->motion.vq[u],  SDI_F->motion.aq[u],
                      SDI_F->armRight.currentCmdVelJntVals[u-LWR_JOINTS], SDI_F->armRight.currentCmdAccJntVals[u-LWR_JOINTS],
                      currentMotion.limits_q[u].maxVel);

            }

          }
          fprintf(fileptr, "%f %f %f %f %f %f %f", SDI_F->armRight.currPose.x, SDI_F->armRight.currPose.y, SDI_F->armRight.currPose.z, SDI_F->armRight.currPose.rx, SDI_F->armRight.currPose.ry, SDI_F->armRight.currPose.rz, SDI_F->armRight.currPose.a);

          fprintf(fileptr," \n");
        }
      }


      if(SDI_F->armLeft.cntrl_mode == LWR_OPEN_LOOP_CNTRL || SDI_F->armLeft.cntrl_mode == LWR_XOPEN_LOOP_CNTRL || SDI_F->armLeft.cntrl_mode == LWR_TRAJECTORY_CNTRL ||
         SDI_F->armRight.cntrl_mode == LWR_OPEN_LOOP_CNTRL || SDI_F->armRight.cntrl_mode == LWR_XOPEN_LOOP_CNTRL || SDI_F->armRight.cntrl_mode == LWR_TRAJECTORY_CNTRL ) {
        if (currentMotion.curTime < currentMotion.duration) {

          /* Update the hand command */
          if(SDI_F->armRight.isHandUsed == GEN_TRUE) {
            if(SDI_F->armRight.toolCfg.tool_type == LWR_DLR_HIT) {
              SDI_F->sahand.handRight = SDI_F->armRight.hand;
            }
       	  }

          if(SDI_F->armLeft.isHandUsed == GEN_TRUE) {
            if(SDI_F->armLeft.toolCfg.tool_type == LWR_DLR_HIT) {
              SDI_F->sahand.handLeft = SDI_F->armLeft.hand;
            }
       	  }

          double timeScaleLoc , tsCalc;
          if((currentMotion.traj.getTsVec()).size() > 0) {
            /* Limits timeScale inthe case where tsVec is computed */
            tsCalc = (currentMotion.traj.getTsVec())[(int)(currentMotion.curTime / LWR_PERIOD)];
            if(tsCalc < 0.01) {
              tsCalc = 0.01;
            }
            timeScaleLoc = MIN(tsCalc, SDI_F->timeScale);

            // printf("times  %f  %f %f %f \n",tsCalc, timeScaleLoc, currentMotion.timeScaleCond.v , currentMotion.curTime);

          } else {
	    timeScaleLoc =  SDI_F->timeScale;
          }

	  if(timeScaleLoc > 0 && timeScaleLoc > SDI_F->timeScaleMax) {
	    timeScaleLoc =  SDI_F->timeScaleMax;
	  } else if (timeScaleLoc < 0 && timeScaleLoc < -SDI_F->timeScaleMax) {
	    timeScaleLoc =  -SDI_F->timeScaleMax;
	  }

	  //printf("before smooth: max: scale: loc: %g %g %g  ", SDI_F->timeScaleMax, SDI_F->timeScale, timeScaleLoc);
	  
	  //limitsTimeScale.maxVel = SDI_F->timeScaleMax;

          sm_ComputeSmoothedStepVel(timeScaleLoc, (double)LWR_PERIOD, limitsTimeScale, &currentMotion.timeScaleCond);
	  //printf(" timeScaleCond.v %g \n", currentMotion.timeScaleCond.v);
	  
	  // NEXT LINE TO BE REMOVED
	  //currentMotion.timeScaleCond.v =  SDI_F->timeScaleMax;

          currentMotion.curTime = currentMotion.curTime + (currentMotion.timeScaleCond.v * LWR_PERIOD);
          SDI_F->posForMove3d.tau = currentMotion.curTime;

	  //printf("after smooth: max: scale: loc: %g %g %g \n", SDI_F->timeScaleMax, SDI_F->timeScale, timeScaleLoc );

          /* Check curTime validity (it can be > 0 since there is the timeScale factor */
          if(currentMotion.curTime < 0.0 || currentMotion.curTime >= currentMotion.duration ) {
            setMotionEnd();
            SDI_F->posForMove3d.tau = 0.0;
            if(SDI_F->saveMotionFile == GEN_TRUE) {
              if(fileptr) {
                closingSaveFile = GEN_TRUE;
              }
            }
            printf("Motion completed\n");
          }
        } else {

          if(SDI_F->armLeft.cntrl_mode == LWR_FORCE_CNTRL || SDI_F->armRight.cntrl_mode == LWR_FORCE_CNTRL) {
            SDI_F->armLeft.cntrl_mode = LWR_FORCE_CNTRL ;
            SDI_F->armRight.cntrl_mode = LWR_FORCE_CNTRL;
            printf("WARN: All Arm switch to LWR_FORCE_CNTRL \n");
          }  else {
            setMotionEnd();
            if(SDI_F->saveMotionFile == GEN_TRUE) {
              if(fileptr) {
                closingSaveFile = GEN_TRUE;
              }
            }
            printf("Motion completed\n");
          }
        }
      } 
      
      
      if(SDI_F->armLeft.cntrl_mode == LWR_FORCE_CNTRL && SDI_F->armRight.cntrl_mode == LWR_FORCE_CNTRL) {
        if(SDI_F->armLeft.isStabilized == GEN_TRUE &&  SDI_F->armRight.isStabilized == GEN_TRUE) {
          setMotionEnd();
          if(SDI_F->saveMotionFile == GEN_TRUE) {
            if(fileptr) {
              closingSaveFile = GEN_TRUE;
            }
          }
          printf("Motion completed\n");
        }
      }
      
      
      if(SDI_F->armLeft.cntrl_mode == LWR_XCLOSED_LOOP_CNTRL || SDI_F->armRight.cntrl_mode == LWR_XCLOSED_LOOP_CNTRL ) {
	  // Check if goal is reached 
	  // compute the dist between the current pose of R0_R7 and currentMotion.th_R0_R7

      Gb_th th_R0_R7_cur, th_Base_R8_cur;
      Gb_dep dep_R0_R7_cur, dep_R0_R7_goto;
        fri2th_cartPos(SDI_F->armRight.currentCartPos, &th_Base_R8_cur);
        lwr_get_th_R0_R7_from_Gb_th_Base_R8(&th_Base_R8_cur, &th_R0_R7_cur);

        Gb_th_dep(&th_R0_R7_cur, &dep_R0_R7_cur);
        Gb_th_dep(&(currentMotion.th_R0_R7_Goto), &dep_R0_R7_goto);

        double dist = sqrt(pow(dep_R0_R7_goto.x - dep_R0_R7_cur.x, 2) +
                    pow(dep_R0_R7_goto.y - dep_R0_R7_cur.y, 2) +
                    pow(dep_R0_R7_goto.z - dep_R0_R7_cur.z, 2) +
                    pow(dep_R0_R7_goto.rx*dep_R0_R7_goto.a - dep_R0_R7_cur.rx*dep_R0_R7_cur.a, 2) +
                    pow(dep_R0_R7_goto.ry*dep_R0_R7_goto.a - dep_R0_R7_cur.ry*dep_R0_R7_cur.a, 2) +
                    pow(dep_R0_R7_goto.rz*dep_R0_R7_goto.a - dep_R0_R7_cur.rz*dep_R0_R7_cur.a, 2) );
	  
	  double cart_thresh = 0.005;
	    if(dist < cart_thresh) {

	      // printf("******* GOAL REACHED *********\n");
	      if(  SDI_F->gotoClosedLoopMode == GEN_FALSE) {
		setMotionEnd();
		if(SDI_F->saveMotionFile == GEN_TRUE) {
		  if(fileptr) {
		    closingSaveFile = GEN_TRUE;
		  }
		}
		printf("Motion completed\n");
	      } else {
		//printf("Goal reached, waiting for another destination\n");
	      }
	    } else {
	      printf("Cart goal not reached\n");
	    }
	    
	  }

      if(currentMotion.type  != LWR_NOT_MOVE && SDI_F->isMonitoringForces == GEN_TRUE) {
        lwrMonitorAndUpdateCntrlMode(&(SDI_F->armLeft), (char*)"LEFT");
        lwrMonitorAndUpdateCntrlMode(&(SDI_F->armRight), (char*)"RIGHT");
      }


#ifdef LWR_FRI
    } /* else { --> flagRight && flagLeft && (SDI_F->status.simulation != GEN_TRUE) */
#endif
  } else {
    if(SDI_F->status.sahandUpdate == GEN_TRUE) {	  
      if(lwrGENMANIP_SAHAND_STRPosterRead(posterSAHandId, &SDI_F->sahandPos) == ERROR) {
        printf("Collision reported cannot read poster of sahand \n");
      } else {
        SDI_F->sahand.handRight.thumbRotation  = SDI_F->sahandPos.handRight.thumbRotation ;
        SDI_F->sahand.handRight.thumb.base     = SDI_F->sahandPos.handRight.thumb.base    ;
        SDI_F->sahand.handRight.thumb.phaProx  = SDI_F->sahandPos.handRight.thumb.phaProx ;
        SDI_F->sahand.handRight.thumb.phaMid   = SDI_F->sahandPos.handRight.thumb.phaMid  ;
        SDI_F->sahand.handRight.thumb.phaDist  = SDI_F->sahandPos.handRight.thumb.phaDist ;
        SDI_F->sahand.handRight.fore.base      = SDI_F->sahandPos.handRight.fore.base     ;
        SDI_F->sahand.handRight.fore.phaProx   = SDI_F->sahandPos.handRight.fore.phaProx  ;
        SDI_F->sahand.handRight.fore.phaMid    = SDI_F->sahandPos.handRight.fore.phaMid   ;
        SDI_F->sahand.handRight.fore.phaDist   = SDI_F->sahandPos.handRight.fore.phaDist  ;
        SDI_F->sahand.handRight.middle.base    = SDI_F->sahandPos.handRight.middle.base   ;
        SDI_F->sahand.handRight.middle.phaProx = SDI_F->sahandPos.handRight.middle.phaProx;
        SDI_F->sahand.handRight.middle.phaMid  = SDI_F->sahandPos.handRight.middle.phaMid ;
        SDI_F->sahand.handRight.middle.phaDist = SDI_F->sahandPos.handRight.middle.phaDist ;
        SDI_F->sahand.handRight.ring.base      = SDI_F->sahandPos.handRight.ring.base     ;
        SDI_F->sahand.handRight.ring.phaProx   = SDI_F->sahandPos.handRight.ring.phaProx  ;
        SDI_F->sahand.handRight.ring.phaMid    = SDI_F->sahandPos.handRight.ring.phaMid   ;
        SDI_F->sahand.handRight.ring.phaDist   = SDI_F->sahandPos.handRight.ring.phaDist  ;
        SDI_F->sahand.cpt = SDI_F->sahandPos.cpt;
      }
    } /*  (SDI_F->status.sahandUpdate == GEN_TRUE) */


  }
  return OK;
}

int lwrMonitorAndUpdateCntrlMode(LWR_ARM_INST *armInst, char* message) {
  double minForce[LWR_JOINTS];
  double coef = 5.0;
  double force = 0.0;
  minForce[0] = coef *1.0;
  minForce[1] = coef *1.0;
  minForce[2] = coef *1.0;
  minForce[3] = coef *1.0;
  minForce[4] = coef *0.5;
  minForce[5] = coef *0.5;
  minForce[6] = coef *0.5;
  unsigned int i = 0;

  double forceThreshold, dist;

  if (currentMotion.type != LWR_NOT_MOVE) {
    if(currentMotion.curTime <= SDI_F->monitorForceParams.dti) {
      forceThreshold = SDI_F->monitorForceParams.thri;
    } else if(currentMotion.curTime >= (currentMotion.duration - SDI_F->monitorForceParams.dtf) ) {
      forceThreshold = SDI_F->monitorForceParams.thrf;
    } else {
      forceThreshold = SDI_F->monitorForceParams.thr;
    }
  } else {
    forceThreshold = SDI_F->monitorForceParams.thr;
  }

  if( lwrDetectChock(armInst, SDI_F->monitorForceParams.TC, forceThreshold) == 1) {
#ifdef LWR_FRI
    dist = lwrComputeDistToEndTraj(armInst);
#endif
    printf("WARN: Monitoring External Torque has reported collision. Motion aborted distanceToGoal %f\n", dist);
    setMotionError(LWR_COLLISION_RISK);   // this will set state as LWR_ABORTED
  }
  return 0;
  
  if(armInst->is_used == GEN_FALSE) {
    armInst->isStabilized = GEN_TRUE;
    return 0;
  }
  for( i=0 ; i<LWR_JOINTS; i++) { 
    force = armInst->estMeanExtJntTrq[i];

    if(ABS(force) > ABS(minForce[i]) ) {
      //if(armInst->cntrl_mode != LWR_FORCE_CNTRL) {
      //  printf("WARN: Arm %s switch to LWR_FORCE_CNTRL \n", message);
      //}
      //armInst->cntrl_mode = LWR_FORCE_CNTRL;
      setMotionEnd();
      if(SDI_F->saveMotionFile == GEN_TRUE) {
        if(fileptr) {
          closingSaveFile = GEN_TRUE;
        }
      }
      printf("Motion stopped due to external forces\n");

      break;
    }
  }
  return 0;
}



static double Rold[LWR_JOINTS];

/* If collision detected return 1 */
int lwrDetectChock(LWR_ARM_INST *armInst, double TC, double threshold) {
  double R[LWR_JOINTS];
  double Text=0.0, Textold=0.0;
  
  
  for(int i=0; i<LWR_JOINTS; i++) {
    Text = armInst->estExtJntTrqHistory[i][JNT_TRQ_NB_TIC_MEAN-1] ;
    Textold =  armInst->estExtJntTrqHistory[i][JNT_TRQ_NB_TIC_MEAN-2];
    R[i] = (1/(2.0*TC+LWR_PERIOD))*( 2*TC*Text - 2*TC*Textold - Rold[i]*(LWR_PERIOD - 2*TC));
    if(ABS(R[i]) >  armInst->maxJntTrq[i]* threshold) {
      printf("WARN: CHOCK DETECTED ON JOINT %d\n", i);
      return 1;
    }
  }
  return 0;
}

int lwrUpdateMotion() {
  SDI_F->motion.size = LWR_NBJOINTS_ARMS_USER;
  if(SDI_F->armLeft.is_used > 0 && SDI_F->armLeft.toolCfg.tool_type == LWR_DLR_HIT) {
    SDI_F->motion.size += 13;
  }
  if(SDI_F->armRight.is_used > 0 && SDI_F->armRight.toolCfg.tool_type == LWR_DLR_HIT) {
    SDI_F->motion.size += 13;
  }
  SDI_F->motion.q[0] =  SDI_F->armLeft.pos.q.q1;
  SDI_F->motion.q[1] =  SDI_F->armLeft.pos.q.q2;
  SDI_F->motion.q[2] =  SDI_F->armLeft.pos.q.q3;
  SDI_F->motion.q[3] =  SDI_F->armLeft.pos.q.q4;
  SDI_F->motion.q[4] =  SDI_F->armLeft.pos.q.q5;
  SDI_F->motion.q[5] =  SDI_F->armLeft.pos.q.q6;
  SDI_F->motion.q[6] =  SDI_F->armLeft.pos.q.q7;
  SDI_F->motion.q[7]  =  SDI_F->armRight.pos.q.q1;
  SDI_F->motion.q[8]  =  SDI_F->armRight.pos.q.q2;
  SDI_F->motion.q[9]  =  SDI_F->armRight.pos.q.q3;
  SDI_F->motion.q[10] =  SDI_F->armRight.pos.q.q4;
  SDI_F->motion.q[11] =  SDI_F->armRight.pos.q.q5;
  SDI_F->motion.q[12] =  SDI_F->armRight.pos.q.q6;
  SDI_F->motion.q[13] =  SDI_F->armRight.pos.q.q7;

  SDI_F->motion.vq[0] = SDI_F->armLeft.pos.vq.q1;
  SDI_F->motion.vq[1] = SDI_F->armLeft.pos.vq.q2;
  SDI_F->motion.vq[2] = SDI_F->armLeft.pos.vq.q3;
  SDI_F->motion.vq[3] = SDI_F->armLeft.pos.vq.q4;
  SDI_F->motion.vq[4] = SDI_F->armLeft.pos.vq.q5;
  SDI_F->motion.vq[5] = SDI_F->armLeft.pos.vq.q6;
  SDI_F->motion.vq[6] = SDI_F->armLeft.pos.vq.q7;
  SDI_F->motion.vq[7]  = SDI_F->armRight.pos.vq.q1;
  SDI_F->motion.vq[8]  = SDI_F->armRight.pos.vq.q2;
  SDI_F->motion.vq[9]  = SDI_F->armRight.pos.vq.q3;
  SDI_F->motion.vq[10] = SDI_F->armRight.pos.vq.q4;
  SDI_F->motion.vq[11] = SDI_F->armRight.pos.vq.q5;
  SDI_F->motion.vq[12] = SDI_F->armRight.pos.vq.q6;
  SDI_F->motion.vq[13] = SDI_F->armRight.pos.vq.q7;

  SDI_F->motion.aq[0] = SDI_F->armLeft.pos.aq.q1;
  SDI_F->motion.aq[1] = SDI_F->armLeft.pos.aq.q2;
  SDI_F->motion.aq[2] = SDI_F->armLeft.pos.aq.q3;
  SDI_F->motion.aq[3] = SDI_F->armLeft.pos.aq.q4;
  SDI_F->motion.aq[4] = SDI_F->armLeft.pos.aq.q5;
  SDI_F->motion.aq[5] = SDI_F->armLeft.pos.aq.q6;
  SDI_F->motion.aq[6] = SDI_F->armLeft.pos.aq.q7;
  SDI_F->motion.aq[7]  = SDI_F->armRight.pos.aq.q1;
  SDI_F->motion.aq[8]  = SDI_F->armRight.pos.aq.q2;
  SDI_F->motion.aq[9]  = SDI_F->armRight.pos.aq.q3;
  SDI_F->motion.aq[10] = SDI_F->armRight.pos.aq.q4;
  SDI_F->motion.aq[11] = SDI_F->armRight.pos.aq.q5;
  SDI_F->motion.aq[12] = SDI_F->armRight.pos.aq.q6;
  SDI_F->motion.aq[13] = SDI_F->armRight.pos.aq.q7;

  if(SDI_F->armRight.toolCfg.tool_type == LWR_DLR_HIT) {
    int indexHandRight = LWR_NBJOINTS_ARMS_USER; // 14 = 2*7
    SDI_F->motion.q[indexHandRight]    = SDI_F->sahand.handRight.thumbRotation ;
    SDI_F->motion.q[indexHandRight+1]  = SDI_F->sahand.handRight.thumb.base    ;
    SDI_F->motion.q[indexHandRight+2]  = SDI_F->sahand.handRight.thumb.phaProx ;
    SDI_F->motion.q[indexHandRight+3]  = SDI_F->sahand.handRight.thumb.phaMid  ;
    SDI_F->motion.q[indexHandRight+4]  = SDI_F->sahand.handRight.fore.base     ;
    SDI_F->motion.q[indexHandRight+5]  = SDI_F->sahand.handRight.fore.phaProx  ;
    SDI_F->motion.q[indexHandRight+6]  = SDI_F->sahand.handRight.fore.phaMid   ;
    SDI_F->motion.q[indexHandRight+7]  = SDI_F->sahand.handRight.middle.base   ;
    SDI_F->motion.q[indexHandRight+8]  = SDI_F->sahand.handRight.middle.phaProx;
    SDI_F->motion.q[indexHandRight+9]  = SDI_F->sahand.handRight.middle.phaMid ;
    SDI_F->motion.q[indexHandRight+10] = SDI_F->sahand.handRight.ring.base     ;
    SDI_F->motion.q[indexHandRight+11] = SDI_F->sahand.handRight.ring.phaProx  ;
    SDI_F->motion.q[indexHandRight+12] = SDI_F->sahand.handRight.ring.phaMid   ;
  }
  return 0;
}
#ifdef LWR_FRI  
int lwrInitArm(LWR_INIT_ARM_STR arm, friRemote *friInst, LWR_ARM_INST *armInst, char* message) {

  armInst->is_used = GEN_FALSE;

  armInst->currentJntVals[0] = 0.0;
  armInst->currentJntVals[1] = 0.0;
  armInst->currentJntVals[2] = 0.0;
  armInst->currentJntVals[3] = 0.0;
  armInst->currentJntVals[4] = 0.0;
  armInst->currentJntVals[5] = 0.0;
  armInst->currentJntVals[6] = 0.0;
  
  /* Setup default cartesian stiffness and damping factors */
  armInst->cartStiff.x  = LWR_DEF_X_STIFFNESS;
  armInst->cartStiff.y  = LWR_DEF_Y_STIFFNESS;
  armInst->cartStiff.z  = LWR_DEF_Z_STIFFNESS;
  armInst->cartStiff.rx = LWR_DEF_A_STIFFNESS;
  armInst->cartStiff.ry = LWR_DEF_B_STIFFNESS;
  armInst->cartStiff.rz = LWR_DEF_C_STIFFNESS;
  armInst->cartDamp.x   = LWR_DEF_X_DAMPING;
  armInst->cartDamp.y   = LWR_DEF_Y_DAMPING;
  armInst->cartDamp.z   = LWR_DEF_Z_DAMPING;
  armInst->cartDamp.rx  = LWR_DEF_A_DAMPING;
  armInst->cartDamp.ry  = LWR_DEF_B_DAMPING;
  armInst->cartDamp.rz  = LWR_DEF_C_DAMPING;

  armInst->maxJntTrq[0] = LWR_A1_MAX_TORQUE;
  armInst->maxJntTrq[1] = LWR_A2_MAX_TORQUE;
  armInst->maxJntTrq[2] = LWR_E1_MAX_TORQUE;
  armInst->maxJntTrq[3] = LWR_A3_MAX_TORQUE;
  armInst->maxJntTrq[4] = LWR_A4_MAX_TORQUE;
  armInst->maxJntTrq[5] = LWR_A5_MAX_TORQUE;
  armInst->maxJntTrq[6] = LWR_A6_MAX_TORQUE;


  armInst->s_date0_tv_sec  = 0.0;
  armInst->s_date0_tv_usec = 0.0;
  armInst->date1 = -1;
  armInst->date2 = -1;

  if(arm.is_used == GEN_TRUE) {

    LWR_NB_CONTROLLED_JNTS += LWR_JOINTS;

    /* Setup transformation matrix from reference system 7 to R tool */
    armInst->toolCfg.distance  = -1;
    armInst->toolCfg.tool_type = arm.tool_type;

    if(armInst->toolCfg.tool_type ==  LWR_DLR_HIT) {
      // if SAHAND is used
      LWR_NB_CONTROLLED_JNTS += 17;
    }

    setupTh7Tool(friInst, armInst->toolCfg.tool_type, armInst->toolCfg.distance, &(armInst->th_7_tool), &(armInst->th_8_tool));

    /* Setup transformation matrix from world to base reference systems */
    armInst->baseCfg.base_type = arm.base_type;
    setupThBase0(friInst, armInst->baseCfg.base_type, &(armInst->th_base_0));

    // Ideally, this should avoid a FRI interpolation error on the first command
    if (friInst->getQuality() >= FRI_QUALITY_BAD){
      syncFriWithMeasuredPose(friInst);
    }

    armInst->sock = friInst->getSock();
    printf("INFO: FRI %s connected at socket %d\n", message, friInst->getSock());
    armInst->is_used = GEN_TRUE;
    printf("INFO: FRI %s pointer %p\n", message,friInst);
  }
  return 0;
}
#else
int lwrInitArm(LWR_INIT_ARM_STR arm,  LWR_ARM_INST *armInst, char* message) {

  armInst->is_used = GEN_FALSE;

  armInst->currentJntVals[0] = 0.0;
  armInst->currentJntVals[1] = 0.0;
  armInst->currentJntVals[2] = 0.0;
  armInst->currentJntVals[3] = 0.0;
  armInst->currentJntVals[4] = 0.0;
  armInst->currentJntVals[5] = 0.0;
  armInst->currentJntVals[6] = 0.0;
  
  /* Setup default cartesian stiffness and damping factors */
  armInst->cartStiff.x  = LWR_DEF_X_STIFFNESS;
  armInst->cartStiff.y  = LWR_DEF_Y_STIFFNESS;
  armInst->cartStiff.z  = LWR_DEF_Z_STIFFNESS;
  armInst->cartStiff.rx = LWR_DEF_A_STIFFNESS;
  armInst->cartStiff.ry = LWR_DEF_B_STIFFNESS;
  armInst->cartStiff.rz = LWR_DEF_C_STIFFNESS;
  armInst->cartDamp.x   = LWR_DEF_X_DAMPING;
  armInst->cartDamp.y   = LWR_DEF_Y_DAMPING;
  armInst->cartDamp.z   = LWR_DEF_Z_DAMPING;
  armInst->cartDamp.rx  = LWR_DEF_A_DAMPING;
  armInst->cartDamp.ry  = LWR_DEF_B_DAMPING;
  armInst->cartDamp.rz  = LWR_DEF_C_DAMPING;

  armInst->maxJntTrq[0] = LWR_A1_MAX_TORQUE;
  armInst->maxJntTrq[1] = LWR_A2_MAX_TORQUE;
  armInst->maxJntTrq[2] = LWR_E1_MAX_TORQUE;
  armInst->maxJntTrq[3] = LWR_A3_MAX_TORQUE;
  armInst->maxJntTrq[4] = LWR_A4_MAX_TORQUE;
  armInst->maxJntTrq[5] = LWR_A5_MAX_TORQUE;
  armInst->maxJntTrq[6] = LWR_A6_MAX_TORQUE;

  armInst->s_date0_tv_sec  = 0.0;
  armInst->s_date0_tv_usec = 0.0;
  armInst->date1 = -1;
  armInst->date2 = -1;

  if(arm.is_used == GEN_TRUE) {

    LWR_NB_CONTROLLED_JNTS += LWR_JOINTS;
    /* create an instance of FRI interface; the connection socket is automatically opened */

    /* Setup transformation matrix from reference system 7 to R tool */
    armInst->toolCfg.distance  = -1;
    armInst->toolCfg.tool_type = arm.tool_type;

    if(armInst->toolCfg.tool_type ==  LWR_DLR_HIT) {
      LWR_NB_CONTROLLED_JNTS += 17;
    }

    setupTh7Tool( armInst->toolCfg.tool_type, armInst->toolCfg.distance, &(armInst->th_7_tool), &(armInst->th_8_tool));

    /* Setup transformation matrix from world to base reference systems */
    armInst->baseCfg.base_type = arm.base_type;
    setupThBase0( armInst->baseCfg.base_type, &(armInst->th_base_0));

    armInst->is_used = GEN_TRUE;
  }
  return 0;
}

#endif


/*------------------------------------------------------------------------
 * Init
 *
 * Description: 
 *
 * Reports:      OK
 *              S_lwr_FILE_NOT_FOUND
 *              S_lwr_UNKNOW_ERROR
 */

/* lwrInitMain  -  codel EXEC of Init
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrInitMain(LWR_INIT_STR *initStr, int *report)
{
  int errcode = 0;
  printf("init lwr\n");
  if (initIsDone) {
    printf("Init is already done! --> Do nothing\n");
    return ETHER;
  }

  LWR_NB_CONTROLLED_JNTS = 0;

  if(initStr->armLeft.is_used == GEN_FALSE && initStr->armRight.is_used == GEN_FALSE) {
    printf("ERROR: There is no arm used \n");  
    SDI_F->status.simulation = GEN_TRUE;
    return ETHER;
  }
#ifdef LWR_FRI  
  if(initStr->armLeft.is_used == GEN_TRUE) {
    printf("INFO: connecting to the left arm ...\n"); 
    friInstLeft = new friRemote(initStr->armLeft.lwr_port, initStr->armLeft.lwr_host.name);
    printf("INFO: FRI %s connected at %s:%d\n", (char*)"LEFT", initStr->armLeft.lwr_host.name, initStr->armLeft.lwr_port);
  }
  lwrInitArm( initStr->armLeft, friInstLeft, &(SDI_F->armLeft), (char*)"LEFT"); 

  if(initStr->armRight.is_used == GEN_TRUE) {
    printf("INFO: connecting to the right arm ...\n"); 
    friInstRight = new friRemote(initStr->armRight.lwr_port, initStr->armRight.lwr_host.name);
    printf("INFO: FRI %s connected at %s:%d\n", (char*)"RIGHT", initStr->armRight.lwr_host.name, initStr->armRight.lwr_port);
  }
  lwrInitArm( initStr->armRight,friInstRight, &(SDI_F->armRight), (char*)"RIGHT");  
#else 
  lwrInitArm( initStr->armLeft,  &(SDI_F->armLeft), (char*)"LEFT"); 
  lwrInitArm( initStr->armRight, &(SDI_F->armRight), (char*)"RIGHT");  
#endif
  /***********************************/
  /*  Initialize global structures   */                 
  /***********************************/

  memset(&currentMotion, 0, sizeof(currentMotion));
  SDI_F->status.sahandUpdate = GEN_FALSE;  
  SDI_F->status.collisionMode = GEN_TRUE;
  SDI_F->status.collisionEnvMode = GEN_FALSE;
  SDI_F->status.simulation = GEN_FALSE;
  SDI_F->gotoClosedLoopMode = GEN_TRUE;

  currentMotion.traj = trajBuff1;
  buffInUse = 1;

  SDI_F->timeScale = 1.0;
  SDI_F->timeScaleMax = 1.0;

  LWR_NBJOINTS_ARMS_USER =  2*LWR_JOINTS;
  LWR_CART_USER = 2*7;

  /* Setup v, a, j limits */
  SDI_F->speedLimit = LWR_DEF_SPEED_LIMIT;
  setupMaxVAJ(SDI_F->speedLimit, currentMotion.limits_q, currentMotion.limits_x);

  /* Setup collision detection manager */
  if (coldmanEnvInit(initStr->world_path.name, initStr->world_file.name) != 0) {
    *report = S_lwr_FILE_NOT_FOUND;
    return ETHER;
  }
  printf("INFO: coldman model loaded from file %s/%s\n", initStr->world_path.name, initStr->world_file.name);

  //TODO Check if all sampling time are equal (armLeft, armRight and LWR_PERIOD)
#ifdef LWR_FRI  
  printf("##########################################################\n");
  printf("### INFO: Now, launch the FRI on the arm(s) controller ###\n");
  printf("##########################################################\n");
  // The next line is to wait the Kuka initialisation
  lwrDoDataExchange();

  if ((threadFriCommTimerId = h2timerAlloc()) == NULL) {
    printf("ERROR: lwrInitMain: h2timerAlloc\n");
    return ETHER;
  }
  if (h2timerStart (threadFriCommTimerId, LWR_TIC, 0) != OK) {
    printf("ERROR: lwrInitMain: h2timerStart\n");
    return ETHER;
  }

  /* create the FriComm thread */
  // taskSpawn : create a "vxWorks" thread
  if (taskSpawn("lwrDataExchangeTask",
                1 /* priorite */,
                VX_FP_TASK,
                4096 /*size*/,
                (FUNCPTR)threadFriCommFct) == ERROR) {
    printf("lwrInitMain: cannot spawn lwrDataExchangeTask\n");
    return ETHER;
  }
#endif

  if(SDI_F->armLeft.is_used == GEN_TRUE && SDI_F->armRight.is_used == GEN_TRUE) {
    SDI_F->armLeft.indexInTraj =  0;
    SDI_F->armLeft.indexCartInTraj =  0;
    SDI_F->armRight.indexInTraj = LWR_JOINTS;
    SDI_F->armRight.indexCartInTraj = 6;
    SDI_F->armRight.indexHandInTraj = 2*LWR_JOINTS;
    SDI_F->armLeft.isHandUsed = GEN_FALSE;
    SDI_F->armRight.isHandUsed = GEN_TRUE;   

  } else if (SDI_F->armLeft.is_used == GEN_TRUE) {
    SDI_F->armLeft.indexInTraj =  0;
    SDI_F->armLeft.indexCartInTraj =  0;
    SDI_F->armLeft.isHandUsed = GEN_FALSE;
    SDI_F->armRight.isHandUsed = GEN_FALSE;   
  } else if (SDI_F->armRight.is_used == GEN_TRUE) {
    SDI_F->armRight.indexInTraj =  0;
    SDI_F->armRight.indexCartInTraj =  6;
    SDI_F->armRight.indexHandInTraj = LWR_JOINTS;
    SDI_F->armLeft.isHandUsed = GEN_FALSE;
    SDI_F->armRight.isHandUsed = GEN_TRUE;   
  } else {
    printf("ERROR: No arm is used\n");
    return ETHER;
  }

  SDI_F->armLeft.isLeft = GEN_TRUE;
  SDI_F->armRight.isLeft = GEN_FALSE;
  SDI_F->isTrqCntrlSofMotion = GEN_FALSE;
  SDI_F->isMonitoringForces = GEN_FALSE;


  SDI_F->monitorForceParams.dti = 0.5; /* the duration at the begining of the trajectory to keep the threshold high */
  SDI_F->monitorForceParams.dtf = 0.5; /* the duration at the begining of the trajectory to keep the threshold low */
  SDI_F->monitorForceParams.thri = 0.1; /* the threshold high */
  SDI_F->monitorForceParams.thrf = 0.003; /* the threshold low */
  SDI_F->monitorForceParams.thr = 0.03; /* the normal threshold */
  SDI_F->monitorForceParams.TC = 0.1; /* time constantfor the DLR forceFilter */

  currentMotion.timeScaleCond.a = 0.0;
  currentMotion.timeScaleCond.v = SDI_F->timeScale;
  currentMotion.timeScaleCond.x = 0.0;


  SDI_F->timeScaleCntrts.jerk = 9.0;
  SDI_F->timeScaleCntrts.acc = 3.0;

  limitsTimeScale.maxJerk =  SDI_F->timeScaleCntrts.jerk;
  limitsTimeScale.maxAcc  =  SDI_F->timeScaleCntrts.acc;
  limitsTimeScale.maxVel  = SDI_F->timeScaleMax;

  SDI_F->saveMotionFile = GEN_FALSE;
  initIsDone = true;
  printf("INFO %d joints are needed for the trajectories that comes from motion planners\n", LWR_NB_CONTROLLED_JNTS);
  printf("lwr Init OK\n");
  return ETHER;
}


/*------------------------------------------------------------------------
 * SetCartStiff
 *
 * Description: Set current cartesian stiffness values
 *
 * Reports:      OK
 *              S_lwr_OutOfRange
 */

/* lwrSetCartStiffStart  -  codel START of SetCartStiff
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrSetCartStiffStart(LWR_CART_STIFF *cartStiff, int *report)
{
  /* TODO verify bounds */
  return EXEC;
}

/* lwrSetCartStiffMain  -  codel EXEC of SetCartStiff
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrSetCartStiffMain(LWR_CART_STIFF *cartStiff, int *report)
{
#ifdef LWR_FRI
  friRemote * friInst = getFriRemote(cartStiff->arm_id);
  if(friInst == NULL) {
    return END;
  }

  friInst->setToKRLReal(0, cartStiff->cartStiff.x);
  friInst->setToKRLReal(1, cartStiff->cartStiff.y);
  friInst->setToKRLReal(2, cartStiff->cartStiff.z);
  friInst->setToKRLReal(3, cartStiff->cartStiff.rx);
  friInst->setToKRLReal(4, cartStiff->cartStiff.ry);
  friInst->setToKRLReal(5, cartStiff->cartStiff.rz);
  friInst->setToKRLInt(CMD_CHANNEL, CART_STIFF);

  printf("Cartesian stiffness set to: %f, %f, %f, %f, %f, %f\n",
         cartStiff->cartStiff.x,  cartStiff->cartStiff.y,  cartStiff->cartStiff.z,
         cartStiff->cartStiff.rx, cartStiff->cartStiff.ry, cartStiff->cartStiff.rz);
#endif
  return END;
}

/* lwrSetCartStiffEnd  -  codel END of SetCartStiff
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrSetCartStiffEnd(LWR_CART_STIFF *cartStiff, int *report)
{
#ifdef LWR_FRI
  friRemote * friInst = getFriRemote(cartStiff->arm_id);
  if(friInst == NULL) {
    return END;
  }
  // Return arm script to idle status
  friInst->setToKRLInt(CMD_CHANNEL, FRI_IDLE);
#endif
  return ETHER;
}

/*------------------------------------------------------------------------
 * SetCartDamp
 *
 * Description: Set current cartesian damping values
 *
 * Reports:      OK
 *              S_lwr_OutOfRange
 */

/* lwrSetCartDampStart  -  codel START of SetCartDamp
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrSetCartDampStart(LWR_CART_DAMP *cartDamp, int *report)
{
  /* TODO verify bounds */
  return EXEC;
}

/* lwrSetCartDampMain  -  codel EXEC of SetCartDamp
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrSetCartDampMain(LWR_CART_DAMP *cartDamp, int *report)
{
#ifdef LWR_FRI
  friRemote * friInst = getFriRemote(cartDamp->arm_id);
  if(friInst == NULL) {
    return END;
  }

  friInst->setToKRLReal(0, cartDamp->cartDamp.x);
  friInst->setToKRLReal(1, cartDamp->cartDamp.y);
  friInst->setToKRLReal(2, cartDamp->cartDamp.z);
  friInst->setToKRLReal(3, cartDamp->cartDamp.rx);
  friInst->setToKRLReal(4, cartDamp->cartDamp.ry);
  friInst->setToKRLReal(5, cartDamp->cartDamp.rz);
  friInst->setToKRLInt(CMD_CHANNEL, CART_DAMP);

  printf("Cartesian damping set to: %f, %f, %f, %f, %f, %f\n",
         cartDamp->cartDamp.x,  cartDamp->cartDamp.y,  cartDamp->cartDamp.z,
         cartDamp->cartDamp.rx, cartDamp->cartDamp.ry, cartDamp->cartDamp.rz);
#endif
  return END;
}

/* lwrSetCartDampEnd  -  codel END of SetCartDamp
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrSetCartDampEnd(LWR_CART_DAMP *cartDamp, int *report)
{
#ifdef LWR_FRI
  friRemote * friInst = getFriRemote(cartDamp->arm_id);
  if(friInst == NULL) {
    return END;
  }

  // Return arm script to idle status
  friInst->setToKRLInt(CMD_CHANNEL, FRI_IDLE);
#endif 
  return ETHER;
}



/*------------------------------------------------------------------------
 * GotoQ
 *
 * Description: 
 *
 * Reports:      OK
 *              S_lwr_NOT_INITIALIZED
 *              S_lwr_WRONG_VALUE
 *              S_lwr_SOFTMOTION_ERROR
 *              S_lwr_FRI_MONITOR_MODE
 *              S_lwr_COLLISION_RISK
 */

/* lwrGotoQStart  -  codel START of GotoQ
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrGotoQStart(LWR_GOTO_CONF_STR *gotoConf, int *report)
{

  currentMotion.timeScaleCond.a = 0.0;
  currentMotion.timeScaleCond.v = MIN(SDI_F->timeScale, SDI_F->timeScaleMax);
  currentMotion.timeScaleCond.x = 0.0;

  if(SDI_F->status.simulation == GEN_TRUE) {
    return EXEC;
  }
#ifdef LWR_FRI
  ACTIVITY_EVENT eventLeft, eventRight;
  if(SDI_F->armLeft.is_used == GEN_TRUE) {
    eventLeft = requestFriStateCmd(report, LWR_ARM_LEFT, LWR_CTRL_POSITION);
  } else {
    eventLeft = EXEC;
  }
  if(SDI_F->armRight.is_used == GEN_TRUE) {
    eventRight = requestFriStateCmd(report, LWR_ARM_RIGHT, LWR_CTRL_POSITION);
  } else {
    eventRight = EXEC;
  }

  if(eventLeft == EXEC && eventRight == EXEC) {
    printf("lwrGotoQStart OK\n");
    return EXEC;
  }
  if(eventLeft == ETHER || eventRight == ETHER) { 
    return ETHER;
  }
  if(eventLeft == START || eventRight == START) {
    return START;
  }
#else
   // simultation mode
  return EXEC;
#endif
}

/* lwrGotoQMain  -  codel EXEC of GotoQ
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrGotoQMain(LWR_GOTO_CONF_STR *gotoConf, int *report)
{
  /* Send new absolute angles for all robot joints */


  std::vector<SM_COND> ic_cond, fc_cond;
  std::vector<SM_LIMITS> limits;

  memcpy(&(SDI_F->gotoConf), gotoConf, sizeof(LWR_GOTO_CONF_STR));


  ic_cond.resize( LWR_NBJOINTS_ARMS_USER);
  fc_cond.resize( LWR_NBJOINTS_ARMS_USER);
  limits.resize( LWR_NBJOINTS_ARMS_USER);

  for(int i=0; i<LWR_NBJOINTS_ARMS_USER; i++) {
     ic_cond[i].x = SDI_F->motion.q[i] ;
     ic_cond[i].v = SDI_F->motion.vq[i] ;
     ic_cond[i].a = SDI_F->motion.aq[i] ;
  }

  if(SDI_F->armLeft.is_used == GEN_TRUE) {
    printf("lwrGotoQAbsMain Arm Left is_used\n");
     fc_cond[0].x = SDI_F->gotoConf.confLeft.q1;
     fc_cond[1].x = SDI_F->gotoConf.confLeft.q2;
     fc_cond[2].x = SDI_F->gotoConf.confLeft.q3;
     fc_cond[3].x = SDI_F->gotoConf.confLeft.q4;
     fc_cond[4].x = SDI_F->gotoConf.confLeft.q5;
     fc_cond[5].x = SDI_F->gotoConf.confLeft.q6;
     fc_cond[6].x = SDI_F->gotoConf.confLeft.q7;
  } else {
    fc_cond[0].x = ic_cond[0].x;
    fc_cond[1].x = ic_cond[1].x;
    fc_cond[2].x = ic_cond[2].x;
    fc_cond[3].x = ic_cond[3].x;
    fc_cond[4].x = ic_cond[4].x;
    fc_cond[5].x = ic_cond[5].x;
    fc_cond[6].x = ic_cond[6].x;
  }
  if(SDI_F->armRight.is_used == GEN_TRUE) {
    printf("lwrGotoQAbsMain Arm Right is_used\n");
    fc_cond[7].x  = SDI_F->gotoConf.confRight.q1;
    fc_cond[8].x  = SDI_F->gotoConf.confRight.q2;
    fc_cond[9].x  = SDI_F->gotoConf.confRight.q3;
    fc_cond[10].x = SDI_F->gotoConf.confRight.q4;
    fc_cond[11].x = SDI_F->gotoConf.confRight.q5;
    fc_cond[12].x = SDI_F->gotoConf.confRight.q6;
    fc_cond[13].x = SDI_F->gotoConf.confRight.q7;
  } else {
    fc_cond[7].x  =  ic_cond[7].x  ;
    fc_cond[8].x  =  ic_cond[8].x  ;
    fc_cond[9].x  =  ic_cond[9].x  ;
    fc_cond[10].x =  ic_cond[10].x ;
    fc_cond[11].x =  ic_cond[11].x ;
    fc_cond[12].x =  ic_cond[12].x ;
    fc_cond[13].x =  ic_cond[13].x ;
  }

  for(int i=0; i<LWR_NBJOINTS_ARMS_USER; i++) {
    fc_cond[i].v = 0.0;
    fc_cond[i].a = 0.0;
  }
  //for(int i=0; i<LWR_NBJOINTS_ARMS_USER; i++) {
  //  limits[i].maxJerk = currentMotion.q_J_max[i];
  //  limits[i].maxAcc = currentMotion.q_A_max[i];
  //  limits[i].maxVel = currentMotion.q_V_max[i];
  //}

  printf(" currentMotion.limits_q.size() %d \n", (int)currentMotion.limits_q.size());
    // Compute soft motion between initial and final conditions
    currentMotion.traj.clear();
  int resp = currentMotion.traj.computeTraj(ic_cond,fc_cond, currentMotion.limits_q, SM_TRAJ::SM_INDEPENDANT);
  
  if (resp != 0) {
    printf("ERROR: Q interpolation failed (sm_ComputeSoftMotionPointToPoint_gen funcion)\n");
    *report = S_lwr_SOFTMOTION_ERROR;
    return ETHER;
  } else {
    printf("INFO: computeTraj() OK\n");
  }

#ifdef LWR_DEBUG
  //currentMotion.traj.print();
#endif
  //  currentMotion.cntrlMode = LWR_OPEN_LOOP_CNTRL;
  SDI_F->armLeft.cntrl_mode = LWR_TRAJECTORY_CNTRL;
  SDI_F->armRight.cntrl_mode = LWR_TRAJECTORY_CNTRL;
  setMotionStart(LWR_POSITION);

  printf("lwrGotoQ EXEC OK\n");
  //printf("currentMotion.timeScaleCond.a %f currentMotion.timeScaleCond.v %f currentMotion.timeScaleCond.x %f\n",currentMotion.timeScaleCond.a, currentMotion.timeScaleCond.v, currentMotion.timeScaleCond.x);

  return END;
}

/* lwrGotoQEnd  -  codel END of GotoQ
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrGotoQEnd(LWR_GOTO_CONF_STR *gotoConf, int *report)
{
  if ((gotoConf->wait_done == GEN_TRUE) &&
      ((currentMotion.state == LWR_STARTED) || (currentMotion.state == LWR_MOVING)))
    return END;
  
  fillMotionReport(currentMotion.error, report);
  return ETHER;
}


//int lwrComputeSoftMotionCmd() {
//  double criticalLength;
//  SM_MOTION_MONO motion[LWR_NBJOINTS_ARMS_USER];
//  SM_LIMITS limitsGoto;
//  int TrajectoryType[LWR_NBJOINTS_ARMS_USER];
//  SM_TIMES T_Jerk[LWR_NBJOINTS_ARMS_USER];
//  double maxTime = 0.0;
//  double sum = 0.0;
//  double maxAxis = 0;
//
//  memset(motion, 0, LWR_NBJOINTS_ARMS_USER*sizeof(SM_MOTION_MONO));
//  for(int i=0; i<LWR_NBJOINTS_ARMS_USER; i++) {
//    motion[i].IC.x = SDI_F->motion.q[i] ;
//    motion[i].IC.v = SDI_F->motion.vq[i] ;
//    motion[i].IC.a = SDI_F->motion.aq[i] ;
//  }
//
//  if(SDI_F->armLeft.is_used == GEN_TRUE) {
//    printf("lwrGotoQAbsMain Arm Left is_used\n");
//    motion[0].FC.x = SDI_F->gotoConf.confLeft.q1;
//    motion[1].FC.x = SDI_F->gotoConf.confLeft.q2;
//    motion[2].FC.x = SDI_F->gotoConf.confLeft.q3;
//    motion[3].FC.x = SDI_F->gotoConf.confLeft.q4;
//    motion[4].FC.x = SDI_F->gotoConf.confLeft.q5;
//    motion[5].FC.x = SDI_F->gotoConf.confLeft.q6;
//    motion[6].FC.x = SDI_F->gotoConf.confLeft.q7;
//  } else {
//    motion[0].FC.x = motion[0].IC.x;
//    motion[1].FC.x = motion[1].IC.x;
//    motion[2].FC.x = motion[2].IC.x;
//    motion[3].FC.x = motion[3].IC.x;
//    motion[4].FC.x = motion[4].IC.x;
//    motion[5].FC.x = motion[5].IC.x;
//    motion[6].FC.x = motion[6].IC.x;
//  }
//  if(SDI_F->armRight.is_used == GEN_TRUE) {
//    printf("lwrGotoQAbsMain Arm Right is_used\n");
//    motion[7].FC.x  = SDI_F->gotoConf.confRight.q1;
//    motion[8].FC.x  = SDI_F->gotoConf.confRight.q2;
//    motion[9].FC.x  = SDI_F->gotoConf.confRight.q3;
//    motion[10].FC.x = SDI_F->gotoConf.confRight.q4;
//    motion[11].FC.x = SDI_F->gotoConf.confRight.q5;
//    motion[12].FC.x = SDI_F->gotoConf.confRight.q6;
//    motion[13].FC.x = SDI_F->gotoConf.confRight.q7;
//  } else {
//    motion[7].FC.x  =  motion[7].IC.x  ;
//    motion[8].FC.x  =  motion[8].IC.x  ;
//    motion[9].FC.x  =  motion[9].IC.x  ;
//    motion[10].FC.x =  motion[10].IC.x ;
//    motion[11].FC.x =  motion[11].IC.x ;
//    motion[12].FC.x =  motion[12].IC.x ;
//    motion[13].FC.x =  motion[13].IC.x ;
//  }
//
//  for(int i=0; i<LWR_NBJOINTS_ARMS_USER; i++) {
//    motion[i].FC.v = 0.0;
//    motion[i].FC.a = 0.0;
//  }
//
//
//  maxTime = 0.0;
//  sum = 0.0;
//  maxAxis = 0;
//
//  for(int i=0; i<LWR_NBJOINTS_ARMS_USER; i++) {
//
//    SM_STATUS resp = sm_ComputeSoftMotion(motion[i].IC, motion[i].FC, currentMotion.limits_q[i], &(T_Jerk[i]), &(TrajectoryType[i]));
//    if (resp != SM_OK) {
//      printf("ERROR sm_ComputeSoftMotion\n");
//    }
//    sm_sum_motionTimes(&(T_Jerk[i]), &sum);
//
//    if(sum > maxTime) {
//      maxTime = sum;
//      maxAxis = i;
//    }
//  }
//  for(int i=0; i<LWR_NBJOINTS_ARMS_USER; i++) {
//    sm_adjustMotionWith3seg(motion[i].IC, motion[i].FC, maxTime, &(motion[i]));
//  }
//}

/*------------------------------------------------------------------------
 * GotoQRel
 *
 * Description: Set relative deplacements for all joints
 *
 * Reports:      OK
 *              S_lwr_NOT_INITIALIZED
 *              S_lwr_WRONG_VALUE
 *              S_lwr_SOFTMOTION_ERROR
 *              S_lwr_FRI_MONITOR_MODE
 */

/* lwrGotoQRelStart  -  codel START of GotoQRel
   Returns:  START EXEC END ETHER FAIL ZOMBIE */

ACTIVITY_EVENT
lwrGotoQRelStart(LWR_GOTO_CONF_STR *gotoConf, int *report)
{
  if(SDI_F->status.simulation == GEN_TRUE) {
    return EXEC;
  }
  
#ifdef LWR_FRI
  ACTIVITY_EVENT eventLeft, eventRight;
  if(SDI_F->armLeft.is_used == GEN_TRUE) {
    eventLeft = requestFriStateCmd(report, LWR_ARM_LEFT, LWR_CTRL_POSITION);
  } else {
    eventLeft = EXEC;
  }
  if(SDI_F->armRight.is_used == GEN_TRUE) {
    eventRight = requestFriStateCmd(report, LWR_ARM_RIGHT, LWR_CTRL_POSITION);
  } else {
    eventRight = EXEC;
  }

  if(eventLeft == EXEC && eventRight == EXEC) {
    printf("lwrGotoQRelStart OK\n");
    return EXEC;
  }
  if(eventLeft == ETHER || eventRight == ETHER) { 
    return ETHER;
  }
  if(eventLeft == START || eventRight == START) {
    return START;
  }
#else
  return EXEC;
#endif
}

/* lwrGotoQRelMain  -  codel EXEC of GotoQRel
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrGotoQRelMain(LWR_GOTO_CONF_STR *gotoConf, int *report)
{
  /* Send new absolute angles for all robot joints */
  if(SDI_F->armLeft.is_used == GEN_TRUE) {
    gotoConf->confLeft.q1 += SDI_F->armLeft.currConf.q1;
    gotoConf->confLeft.q2 += SDI_F->armLeft.currConf.q2;
    gotoConf->confLeft.q3 += SDI_F->armLeft.currConf.q3;
    gotoConf->confLeft.q4 += SDI_F->armLeft.currConf.q4;
    gotoConf->confLeft.q5 += SDI_F->armLeft.currConf.q5;
    gotoConf->confLeft.q6 += SDI_F->armLeft.currConf.q6;
    gotoConf->confLeft.q7 += SDI_F->armLeft.currConf.q7;
  } else {
    gotoConf->confLeft.q1 = SDI_F->armLeft.currConf.q1;
    gotoConf->confLeft.q2 = SDI_F->armLeft.currConf.q2;
    gotoConf->confLeft.q3 = SDI_F->armLeft.currConf.q3;
    gotoConf->confLeft.q4 = SDI_F->armLeft.currConf.q4;
    gotoConf->confLeft.q5 = SDI_F->armLeft.currConf.q5;
    gotoConf->confLeft.q6 = SDI_F->armLeft.currConf.q6;
    gotoConf->confLeft.q7 = SDI_F->armLeft.currConf.q7;
  }

  if(SDI_F->armRight.is_used == GEN_TRUE) {
    gotoConf->confRight.q1 += SDI_F->armRight.currConf.q1;
    gotoConf->confRight.q2 += SDI_F->armRight.currConf.q2;
    gotoConf->confRight.q3 += SDI_F->armRight.currConf.q3;
    gotoConf->confRight.q4 += SDI_F->armRight.currConf.q4;
    gotoConf->confRight.q5 += SDI_F->armRight.currConf.q5;
    gotoConf->confRight.q6 += SDI_F->armRight.currConf.q6;
    gotoConf->confRight.q7 += SDI_F->armRight.currConf.q7;
  } else {
    gotoConf->confRight.q1 = SDI_F->armRight.currConf.q1;
    gotoConf->confRight.q2 = SDI_F->armRight.currConf.q2;
    gotoConf->confRight.q3 = SDI_F->armRight.currConf.q3;
    gotoConf->confRight.q4 = SDI_F->armRight.currConf.q4;
    gotoConf->confRight.q5 = SDI_F->armRight.currConf.q5;
    gotoConf->confRight.q6 = SDI_F->armRight.currConf.q6;
    gotoConf->confRight.q7 = SDI_F->armRight.currConf.q7;
  }
  return lwrGotoQMain(gotoConf, report);
}

/* lwrGotoQRelEnd  -  codel END of GotoQRel
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrGotoQRelEnd(LWR_GOTO_CONF_STR *gotoConf, int *report)
{
  if ((gotoConf->wait_done == GEN_TRUE) &&
      ((currentMotion.state == LWR_STARTED) || (currentMotion.state == LWR_MOVING)))
    return END;

  fillMotionReport(currentMotion.error, report);
  return ETHER;
}


/*------------------------------------------------------------------------
 * GotoXAbs
 *
 * Description: Send a new absolute position/orientation for the tool
 *
 * Reports:      OK
 *              S_lwr_NOT_INITIALIZED
 *              S_lwr_WRONG_VALUE
 *              S_lwr_SOFTMOTION_ERROR
 *              S_lwr_FRI_MONITOR_MODE
 *              S_lwr_MGI_CALC_ERROR
 *              S_lwr_MGI_SINGULARITY
 *              S_lwr_JNT_OUT_OF_RANGE
 *              S_lwr_COLLISION_RISK
 */

/* lwrGotoXAbsStart  -  codel START of GotoXAbs
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrGotoXAbsStart(LWR_GOTO_POSE_STR *gotoPose, int *report)
{
  if(SDI_F->status.simulation == GEN_TRUE) {
    return EXEC;
  }
  
#ifdef LWR_FRI
  ACTIVITY_EVENT eventLeft, eventRight;
  if(SDI_F->armLeft.is_used == GEN_TRUE) {
    eventLeft = requestFriStateCmd(report, LWR_ARM_LEFT, LWR_CTRL_POSITION);
  } else {
    eventLeft = EXEC;
  }
  if(SDI_F->armRight.is_used == GEN_TRUE) {
    eventRight = requestFriStateCmd(report, LWR_ARM_RIGHT, LWR_CTRL_POSITION);
  } else {
    eventRight = EXEC;
  }

  if(eventLeft == EXEC && eventRight == EXEC) {
    printf("lwrGotoXAbsStart OK\n");
    return EXEC;
  }
  if(eventLeft == ETHER || eventRight == ETHER) { 
    return ETHER;
  }
  if(eventLeft == START || eventRight == START) {
    if(eventLeft == START ) {
      printf("lwrGotoXAbsStart waiting for arm Left\n");
    }
    if(eventRight == START ) {
      printf("lwrGotoXAbsStart waiting for arm Right\n");
    }
    return START;
  }
#else
  return EXEC;
#endif
}

/* lwrGotoXAbsMain  -  codel EXEC of GotoXAbs
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrGotoXAbsMain(LWR_GOTO_POSE_STR *gotoPose, int *report)
{
  Gb_dep gotoDep = gotoPose->poseRight;
  LWR_GOTO_CONF_STR gotoConf;
  if (! normalize( gotoDep.rx,  gotoDep.ry,  gotoDep.rz,
                   &gotoDep.rx, &gotoDep.ry, &gotoDep.rz)) {
    printf("ERROR: null vector provided for tool orientation\n");
    *report = S_lwr_WRONG_VALUE;
    return ETHER;
  }

  Gb_dep dep_Rt;
  Gb_th  th_goto;
  Gb_th th_Rt, th_R7, th_7_tool_inv;
  Gb_th th_invBase;

  Gb_th th_7_8;

  th_7_8.vx.x =  1;
  th_7_8.vy.x =  0;
  th_7_8.vz.x =  0;
  th_7_8.vp.x =  0;

  th_7_8.vx.y =  0;
  th_7_8.vy.y =  1;
  th_7_8.vz.y =  0;
  th_7_8.vp.y =  0;

  th_7_8.vx.z =  0;
  th_7_8.vy.z =  0;
  th_7_8.vz.z =  1;
  th_7_8.vp.z = LWR_R7;

  Gb_dep_th(&gotoDep, &th_goto);
  Gb_th_print(&th_goto, "Th_goto");
  Gb_th_inverse(&SDI_F->armRight.th_base_0, &th_invBase);
  Gb_th_produit(&th_invBase, &th_goto, &th_Rt);


  Gb_th_inverse(&SDI_F->armRight.th_7_tool, &th_7_tool_inv);
  //Gb_th_inverse(&th_7_8, &th_8_tool_inv);
  Gb_th_produit(&th_Rt, &th_7_tool_inv, &th_R7);

  Gb_th_print(&th_Rt, "Trt");
  Gb_th_print(&th_R7, "Tr7");
  
  int e1, e2, e3;
  kukaLBR_gete1e2e3(LWR_R3, LWR_R5, &SDI_F->armRight.currConf, &e1, &e2, &e3);
  
  Gb_q7 q;
  Gb_statusMGI result =
      kukaLBR_mgi_q_e(&th_R7, &SDI_F->armRight.currConf, LWR_R3, LWR_R5, LWR_MGI_EPSILON,
                      e1, e2, e3, &q);
  
  switch (result) {
  case MGI_ERROR:
    printf("WARN: MGI calculation failed. Coldman model is probably wrong\n");
    return ETHER;
    break;
  case MGI_SINGULAR:
    printf("WARN: Singularity on MGI calculation. Colman model is probably wrong\n");
    return ETHER;
    break;
  case MGI_APPROXIMATE:
    printf("WARN: Approximate MGI calculation. Colman model can be inexact\n");
    return ETHER;
    break;
  case MGI_OK:
    printf("INFO: MGI OK\n");       
  }

  /* Send new absolute angles for all robot joints */
  if(SDI_F->armLeft.is_used == GEN_TRUE) {
    gotoConf.confLeft.q1 = SDI_F->armLeft.currConf.q1;
    gotoConf.confLeft.q2 = SDI_F->armLeft.currConf.q2;
    gotoConf.confLeft.q3 = SDI_F->armLeft.currConf.q3;
    gotoConf.confLeft.q4 = SDI_F->armLeft.currConf.q4;
    gotoConf.confLeft.q5 = SDI_F->armLeft.currConf.q5;
    gotoConf.confLeft.q6 = SDI_F->armLeft.currConf.q6;
    gotoConf.confLeft.q7 = SDI_F->armLeft.currConf.q7;
  } else {
    gotoConf.confLeft.q1 = SDI_F->armLeft.currConf.q1;
    gotoConf.confLeft.q2 = SDI_F->armLeft.currConf.q2;
    gotoConf.confLeft.q3 = SDI_F->armLeft.currConf.q3;
    gotoConf.confLeft.q4 = SDI_F->armLeft.currConf.q4;
    gotoConf.confLeft.q5 = SDI_F->armLeft.currConf.q5;
    gotoConf.confLeft.q6 = SDI_F->armLeft.currConf.q6;
    gotoConf.confLeft.q7 = SDI_F->armLeft.currConf.q7;
  }

  if(SDI_F->armRight.is_used == GEN_TRUE) {
    gotoConf.confRight.q1 = q.q1;
    gotoConf.confRight.q2 = q.q2;
    gotoConf.confRight.q3 = q.q3;
    gotoConf.confRight.q4 = q.q4;
    gotoConf.confRight.q5 = q.q5;
    gotoConf.confRight.q6 = q.q6;
    gotoConf.confRight.q7 = q.q7;
  } else {
    gotoConf.confRight.q1 = SDI_F->armRight.currConf.q1;
    gotoConf.confRight.q2 = SDI_F->armRight.currConf.q2;
    gotoConf.confRight.q3 = SDI_F->armRight.currConf.q3;
    gotoConf.confRight.q4 = SDI_F->armRight.currConf.q4;
    gotoConf.confRight.q5 = SDI_F->armRight.currConf.q5;
    gotoConf.confRight.q6 = SDI_F->armRight.currConf.q6;
    gotoConf.confRight.q7 = SDI_F->armRight.currConf.q7;
  }
  return lwrGotoQMain(&gotoConf, report);


  //    setMotionStart(LWR_CARTESIAN);
  //

  return END;
}

/* lwrGotoXAbsEnd  -  codel END of GotoXAbs
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrGotoXAbsEnd(LWR_GOTO_POSE_STR *gotoPose, int *report)
{
  if ((gotoPose->wait_done == GEN_TRUE) &&
      ((currentMotion.state == LWR_STARTED) || (currentMotion.state == LWR_MOVING)))
    return END;
  
  fillMotionReport(currentMotion.error, report);
  return ETHER;
}


/*------------------------------------------------------------------------
 * GotoXRel
 *
 * Description: Send a relative change on tool position/orientation
 *
 * Reports:      OK
 *              S_lwr_NOT_INITIALIZED
 *              S_lwr_WRONG_VALUE
 *              S_lwr_SOFTMOTION_ERROR
 *              S_lwr_FRI_MONITOR_MODE
 *              S_lwr_MGI_CALC_ERROR
 *              S_lwr_MGI_SINGULARITY
 *              S_lwr_JNT_OUT_OF_RANGE
 */

/* lwrGotoXRelStart  -  codel START of GotoXRel
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrGotoXRelStart(LWR_GOTO_POSE_STR *gotoPose, int *report)
{
  if(SDI_F->status.simulation == GEN_TRUE) {
    return EXEC;
  }
  
#ifdef LWR_FRI
  ACTIVITY_EVENT eventLeft, eventRight;
  if(SDI_F->armLeft.is_used == GEN_TRUE) {
    eventLeft = requestFriStateCmd(report, LWR_ARM_LEFT, LWR_CTRL_POSITION);
  } else {
    eventLeft = EXEC;
  }
  if(SDI_F->armRight.is_used == GEN_TRUE) {
    eventRight = requestFriStateCmd(report, LWR_ARM_RIGHT, LWR_CTRL_POSITION);
  } else {
    eventRight = EXEC;
  }

  if(eventLeft == EXEC && eventRight == EXEC) {
    printf("lwrGotoXAbsStart OK\n");
    return EXEC;
  }
  if(eventLeft == ETHER || eventRight == ETHER) { 
    return ETHER;
  }
  if(eventLeft == START || eventRight == START) {
    if(eventLeft == START ) {
      printf("lwrGotoXAbsStart waiting for arm Left\n");
    }
    if(eventRight == START ) {
      printf("lwrGotoXAbsStart waiting for arm Right\n");
    }
    return START;
  }
#else
  return EXEC;
#endif
}

/* lwrGotoXRelMain  -  codel EXEC of GotoXRel
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrGotoXRelMain(LWR_GOTO_POSE_STR *gotoPose, int *report)
{

  LWR_GOTO_CONF_STR gotoConf;
  Gb_dep gotoDep, currDep;
  //
  gotoDep = gotoPose->poseRight;
  //  //multip. por v3 en lugar de dep  >>>   solo aplicar la rotacion de th_base_0, no la transalt
  //  //GotoHome >> usar GotoQAbs, con una config. propia para cada base
  //
  //  // Convert the given pose (in base coordinates) into R0 system
  Gb_th_x_dep(&SDI_F->armRight.th_base_0, &gotoDep, &gotoDep);
  Gb_dep_print(&gotoDep);
  //
  //  // Get and normalize the current pose
  fri2dep_cartPos(SDI_F->armRight.currentCartPos, &currDep);
  normalize( currDep.rx,  currDep.ry,  currDep.rz,
             &currDep.rx, &currDep.ry, &currDep.rz);
  gotoDep.rx = currDep.rx;
  gotoDep.ry = currDep.ry;
  gotoDep.rz = currDep.rz;
  //
  //  // Undo the tool transformation (from R tool to R8 system)
  Gb_th th_Rt, th_R8, th_8_tool_inv;
  Gb_dep_th(&gotoDep, &th_Rt);
  Gb_th_inverse(&SDI_F->armRight.th_8_tool, &th_8_tool_inv);
  Gb_th_produit(&th_Rt, &th_8_tool_inv, &th_R8);

  Gb_th_dep(&th_R8, &gotoDep);

  Gb_dep_print(&gotoDep);
  // //  // Send a relative change on tool position/orientation
  //gotoDep.x += currDep.x;
  //gotoDep.y += currDep.y;
  //gotoDep.z += currDep.z;
  //gotoDep.a += currDep.a;
  //Gb_dep_print(&gotoDep);
  /////*
  // //  // Normalize rotation to add it to current rotation (already normalized)
  //if (! normalize( gotoDep.rx,  gotoDep.ry,  gotoDep.rz,
  //                 &gotoDep.rx, &gotoDep.ry, &gotoDep.rz) ||
  //    ! gotoDep.a)
  //  {
  //    // Null rotation vector and/or angle; retain the current rotation
  //    gotoDep.rx  = currDep.rx;
  //    gotoDep.ry  = currDep.ry;
  //    gotoDep.rz  = currDep.rz;
  //  }
  //else {
  //  gotoDep.rx += currDep.rx;
  //  gotoDep.ry += currDep.ry;
  //  gotoDep.rz += currDep.rz;
  //}

  int e1, e2, e3;
  kukaLBR_gete1e2e3(LWR_R3, LWR_R5, &SDI_F->armRight.currConf, &e1, &e2, &e3);

  Gb_q7 q;
  Gb_statusMGI result =
      kukaLBR_mgi_q_e(&th_R8, &SDI_F->armRight.currConf, LWR_R3, LWR_R5, LWR_MGI_EPSILON,
                      e1, e2, e3, &q);

  switch (result) {
  case MGI_ERROR:
    printf("WARN: MGI calculation failed. Coldman model is probably wrong\n");
    return ETHER;
    break;
  case MGI_SINGULAR:
    printf("WARN: Singularity on MGI calculation. Colman model is probably wrong\n");
    return ETHER;
    break;
  case MGI_APPROXIMATE:
    printf("WARN: Approximate MGI calculation. Colman model can be inexact\n");
    return ETHER;
    break;
  case MGI_OK:
    printf("INFO: MGI OK\n");       
  }
}

/* lwrGotoXRelEnd  -  codel END of GotoXRel
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrGotoXRelEnd(LWR_GOTO_POSE_STR *gotoPose, int *report)
{
  if ((gotoPose->wait_done == GEN_TRUE) &&
      ((currentMotion.state == LWR_STARTED) || (currentMotion.state == LWR_MOVING)))
    return END;
  
  fillMotionReport(currentMotion.error, report);
  return ETHER;
}

/*------------------------------------------------------------------------
 * Stop
 *
 * Description: Gently stop arm
 *
 * Reports:      OK
 */

/* lwrStopMain  -  codel EXEC of Stop
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrStopMain( double *dist, int *report)
{
  /* Gently stop arm by forcing the end of the current motion */
  setMotionEnd();
#ifdef LWR_FRI
  *dist = lwrComputeDistToEndTraj(&(SDI_F->armRight));
#endif
  *report = OK;
  printf("INFO: Motion stoped\n");

  return ETHER;
}



/*------------------------------------------------------------------------
 * TrackQ
 *
 * Description: 
 *
 * Reports:      OK
 *              S_lwr_NOT_INITIALIZED
 *              S_lwr_FRI_MONITOR_MODE
 *              S_lwr_POSTER_NOT_FOUND
 *              S_lwr_POSTER_READ_ERROR
 *              S_lwr_CMD_STOP_TRACK
 *              S_lwr_NB_POINTS
 *              S_lwr_ERROR_POSITION
 *              S_lwr_BAD_MODE
 *              S_lwr_FILE_NOT_FOUND
 *              S_lwr_ERROR_INIT_POSITION
 *              S_lwr_Q_LIMITED
 */

/* lwrTrackQStart  -  codel START of TrackQ
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrTrackQStart(LWR_TRACK_STR *trackStr, int *report)
{
  setMotionEnd();
  SDI_F->posForMove3d.tau = 0.0;
  if(SDI_F->saveMotionFile == GEN_TRUE) {
    if(fileptr) {
      closingSaveFile = GEN_TRUE;
    }
  }
  printf("Motion aborted by trackQ\n"); 

  if(SDI_F->status.simulation == GEN_TRUE) {
    return EXEC;
  }
  
#ifdef LWR_FRI
  ACTIVITY_EVENT eventLeft, eventRight;
  if(SDI_F->armLeft.is_used == GEN_TRUE) {
    eventLeft = requestFriStateCmd(report, LWR_ARM_LEFT, LWR_CTRL_POSITION);
  } else {
    eventLeft = EXEC;
  }
  if(SDI_F->armRight.is_used == GEN_TRUE) {
    eventRight = requestFriStateCmd(report, LWR_ARM_RIGHT, LWR_CTRL_POSITION);
  } else {
    eventRight = EXEC;
  }

  if(eventLeft == EXEC && eventRight == EXEC) {
    printf("lwrGotoQRelStart OK\n");
    return EXEC;
  }
  if(eventLeft == ETHER || eventRight == ETHER) { 
    return ETHER;
  }
  if(eventLeft == START || eventRight == START) {
    if(eventLeft == START ) {
      printf("lwrGotoQRelStart waiting for arm Left\n");
    }
    if(eventRight == START ) {
      printf("lwrGotoQRelStart waiting for arm Right\n");
    }
    return START;
  }
#else
  return EXEC;
#endif
}

#ifdef LWR_FRI
int * fctDataExchange()
{
  lwrDoDataExchange();
  return 0;
}
#endif

/* lwrTrackQExec  -  codel EXEC of TrackQ
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrTrackQExec(LWR_TRACK_STR *trackStr, int *report)
{

  switch(trackStr->trackMode) {
  case LWR_TRACK_FILE:

#ifdef LWR_FRI
    currentMotion.traj.load(trackStr->posterName.name, fctDataExchange);
#else
    printf("load file\n");
    currentMotion.traj.load(trackStr->posterName.name, NULL);
#endif
    printf("file loaded\n");
    break;

  case LWR_TRACK_POSTER:
    if(lwrSM_TRAJ_STRPosterFind (trackStr->posterName.name, &lwr_poster_q) != OK) {
      *report = S_lwr_POSTER_NOT_FOUND;
      return ETHER;
    }
    if(lwrSM_TRAJ_STRPosterRead (lwr_poster_q, &smTraj) != OK) {
      *report = S_lwr_POSTER_READ_ERROR;
      printf("ERROR: %s poster read failed", trackStr->posterName.name);
      return ETHER;
    }
    if(smTraj.nbAxis != LWR_NB_CONTROLLED_JNTS) {
      printf("ERROR: lwr::TrackQ traj.nbAxis (%d) != %d \n", smTraj.nbAxis,  LWR_NB_CONTROLLED_JNTS);
      return ETHER;
    } else { 
      printf("INFO: lwr::TrackQ there are %d axes in the trajectory \n", smTraj.nbAxis);
    }
    currentMotion.traj.clear();
    currentMotion.traj.importFromSM_TRAJ_STR(&smTraj);
    break;
  default:
    *report = S_lwr_BAD_MODE;
    return ETHER;
  }

  currentMotion.traj.computeTimeOnTraj();

  /* check initial position */
  Gb_q7 errorMax;
  std::vector<SM_COND> cond;
  errorMax.q1 = LWR_A1_MAX_VELOCITY/10.0;
  errorMax.q2 = LWR_A2_MAX_VELOCITY/10.0;
  errorMax.q3 = LWR_E1_MAX_VELOCITY/10.0;
  errorMax.q4 = LWR_A3_MAX_VELOCITY/10.0;
  errorMax.q5 = LWR_A4_MAX_VELOCITY/10.0;
  errorMax.q6 = LWR_A5_MAX_VELOCITY/10.0;
  errorMax.q7 = LWR_A6_MAX_VELOCITY/10.0;

  if(SDI_F->timeScale < 0.0) {
    SDI_F->timeScale = 1.0;
    printf("WARN: lwr::TrackQ set timeScale to 1.0\n");
  }
  currentMotion.traj.getMotionCond(0.0, cond);
  /* check Left arm */
  if(SDI_F->armLeft.is_used == GEN_TRUE) {
    if ((fabs(SDI_F->armLeft.currConf.q1 - cond[0+SDI_F->armLeft.indexInTraj].x) > errorMax.q1) ||
        (fabs(SDI_F->armLeft.currConf.q2 - cond[1+SDI_F->armLeft.indexInTraj].x) > errorMax.q2) ||
        (fabs(SDI_F->armLeft.currConf.q3 - cond[2+SDI_F->armLeft.indexInTraj].x) > errorMax.q3) ||
        (fabs(SDI_F->armLeft.currConf.q4 - cond[3+SDI_F->armLeft.indexInTraj].x) > errorMax.q4) ||
        (fabs(SDI_F->armLeft.currConf.q5 - cond[4+SDI_F->armLeft.indexInTraj].x) > errorMax.q5) ||
        (fabs(SDI_F->armLeft.currConf.q6 - cond[5+SDI_F->armLeft.indexInTraj].x) > errorMax.q6) ||
        (fabs(SDI_F->armLeft.currConf.q7 - cond[6+SDI_F->armLeft.indexInTraj].x) > errorMax.q7)) {
      *report = S_lwr_ERROR_INIT_POSITION;
      return ETHER;
    }
  }
  if(SDI_F->armRight.is_used == GEN_TRUE) {
    if ((fabs(SDI_F->armRight.currConf.q1 - cond[0+SDI_F->armRight.indexInTraj].x) > errorMax.q1) ||
        (fabs(SDI_F->armRight.currConf.q2 - cond[1+SDI_F->armRight.indexInTraj].x) > errorMax.q2) ||
        (fabs(SDI_F->armRight.currConf.q3 - cond[2+SDI_F->armRight.indexInTraj].x) > errorMax.q3) ||
        (fabs(SDI_F->armRight.currConf.q4 - cond[3+SDI_F->armRight.indexInTraj].x) > errorMax.q4) ||
        (fabs(SDI_F->armRight.currConf.q5 - cond[4+SDI_F->armRight.indexInTraj].x) > errorMax.q5) ||
        (fabs(SDI_F->armRight.currConf.q6 - cond[5+SDI_F->armRight.indexInTraj].x) > errorMax.q6) ||
        (fabs(SDI_F->armRight.currConf.q7 - cond[6+SDI_F->armRight.indexInTraj].x) > errorMax.q7)) {
      *report = S_lwr_ERROR_INIT_POSITION;
      return ETHER;
    }
  }

  /* compute the maximum timeScale allowed */
  std::vector<double> maxVel;
  maxVel.resize(7);    
  maxVel[0] = LWR_A1_MAX_VELOCITY;
  maxVel[1] = LWR_A2_MAX_VELOCITY;
  maxVel[2] = LWR_E1_MAX_VELOCITY;
  maxVel[3] = LWR_A3_MAX_VELOCITY;
  maxVel[4] = LWR_A4_MAX_VELOCITY;
  maxVel[5] = LWR_A5_MAX_VELOCITY;
  maxVel[6] = LWR_A6_MAX_VELOCITY;
  
  currentMotion.traj.computeMaxTimeScaleVector(maxVel, LWR_PERIOD, limitsTimeScale);
  currentMotion.timeScaleCond.a = 0.0;

  currentMotion.timeScaleCond.v = MIN((currentMotion.traj.getTsVec())[0], SDI_F->timeScale);
  currentMotion.timeScaleCond.v = MIN(currentMotion.timeScaleCond.v, SDI_F->timeScaleMax);

  currentMotion.timeScaleCond.x = 0.0;

  currentMotion.curTime = 0.0;
  SDI_F->armLeft.cntrl_mode = LWR_OPEN_LOOP_CNTRL;
  SDI_F->armRight.cntrl_mode = LWR_OPEN_LOOP_CNTRL;
  setMotionStart(LWR_POSITION);
  return END;
}

/* lwrTrackQEnd  -  codel END of TrackQ
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrTrackQEnd(LWR_TRACK_STR *trackStr, int *report)
{
  if ((currentMotion.state == LWR_STARTED) || (currentMotion.state == LWR_MOVING)){
    return END;
  }
  fillMotionReport(currentMotion.error, report);
  return ETHER;
}


/*------------------------------------------------------------------------
 * SetCollisionEnvPoster
 *
 * Description: 
 *
 * Reports:      OK
 *              S_lwr_CANNOT_FIND_POSTER
 *              S_lwr_CANNOT_READ_POSTER
 */

/* lwrSetCollisionEnvPosterMain  -  codel EXEC of SetCollisionEnvPoster
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrSetCollisionEnvPosterMain(GEN_STRING128 *collisionEnvPosterName, int *report)
{

  /* Find the external poster and try it by reading once */
  if(lwrGENMANIP_COLLISION_STATE_STRPosterFind(collisionEnvPosterName->name,
                                               &posterMhpCollideId) == ERROR) {
    posterMhpCollideId = NULL;
    *report =  S_lwr_CANNOT_FIND_POSTER;
  }
  else if(lwrGENMANIP_COLLISION_STATE_STRPosterRead(posterMhpCollideId,
                                                    &SDI_F->collisionEnvState) == ERROR) {
    posterMhpCollideId = NULL;
    *report =  S_lwr_CANNOT_READ_POSTER;
  }
  else {
    *report = OK;
  }

  return ETHER;
}

/*------------------------------------------------------------------------
 * SetScheme
 *
 * Description: 
 *
 * Reports:      OK
 */

/* lwrSetSchemeMain  -  codel EXEC of SetScheme
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrSetSchemeMain(LWR_FRI_SCHEME *fri_scheme, int *report)
{
  if(SDI_F->status.simulation == GEN_TRUE) {
    return EXEC;
  }
  
#ifdef LWR_FRI
  ACTIVITY_EVENT event;
  event = requestFriStateCmd(report, fri_scheme->arm_id, fri_scheme->fri_scheme);

  if(event == EXEC) {
    printf("lwrGotoXAbsStart OK\n");
    return END;
  }
  if(event == ETHER) { 
    return ETHER;
  }
  if(event == START) {
    printf("lwrGotoXAbsStart waiting for arm Left\n");
    return EXEC;
  }
#else
  return END;
#endif 
}

/* lwrSetSchemeEnd  -  codel END of SetScheme
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrSetSchemeEnd(LWR_FRI_SCHEME *fri_scheme, int *report)
{
  /* ... add your code here ... */
  return ETHER;
}


/*------------------------------------------------------------------------
 * SetCntrlMode
 *
 * Description: 
 *
 * Reports:      OK
 */

/* lwrSetCntrlModeStart  -  codel START of SetCntrlMode
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrSetCntrlModeStart(LWR_CNTRL_MODE_ENUM *cntrl_mode, int *report)
{
  if(SDI_F->status.simulation == GEN_TRUE) {
    return EXEC;
  }
  ACTIVITY_EVENT eventLeft, eventRight;  
#ifdef LWR_FRI

  SDI_F->armLeft.cntrl_mode = *cntrl_mode;
  SDI_F->armRight.cntrl_mode = *cntrl_mode;

  switch(*cntrl_mode) {
  case LWR_OPEN_LOOP_CNTRL:
  case LWR_TRAJECTORY_CNTRL:
  case LWR_GOAL_CNTRL:
  case LWR_EXTERNAL_POSITION_CNTRL:
  case LWR_EXTERNAL_VELOCITY_CNTRL:
  case LWR_FORCE_CNTRL:

    if(SDI_F->armLeft.is_used == GEN_TRUE) {
      eventLeft = requestFriStateCmd(report, LWR_ARM_LEFT, LWR_CTRL_POSITION);
    } else {
      eventLeft = EXEC;
    }
    if(SDI_F->armRight.is_used == GEN_TRUE) {
      eventRight = requestFriStateCmd(report, LWR_ARM_RIGHT, LWR_CTRL_POSITION);
    } else {
      eventRight = EXEC;
    }
    
    if(eventLeft == EXEC && eventRight == EXEC) {
      printf("lwrGotoQRelStart OK\n");
      return EXEC;
    }
    if(eventLeft == ETHER || eventRight == ETHER) { 
      return ETHER;
    }
    if(eventLeft == START || eventRight == START) {
      if(eventLeft == START ) {
        printf("lwrGotoQRelStart waiting for arm Left\n");
      }
      if(eventRight == START ) {
        printf("lwrGotoQRelStart waiting for arm Right\n");
      }
      return START;
    }
    break;
  default:

    break;
  }
#else
  return EXEC;
#endif
}

/* lwrSetCntrlModeMain  -  codel EXEC of SetCntrlMode
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrSetCntrlModeMain(LWR_CNTRL_MODE_ENUM *cntrl_mode, int *report)
{
  currentMotion.curTime = 0.0;
  setMotionStart(LWR_POSITION);
  return ETHER;
}

/* lwrSetCntrlModeEnd  -  codel END of SetCntrlMode
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrSetCntrlModeEnd(LWR_CNTRL_MODE_ENUM *cntrl_mode, int *report)
{
  /* ... add your code here ... */
  return ETHER;
}

/* lwrSetCntrlModeInter  -  codel INTER of SetCntrlMode
   Returns:  INTER ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrSetCntrlModeInter(LWR_CNTRL_MODE_ENUM *cntrl_mode, int *report)
{
  /* ... add your code here ... */
  return ETHER;
}



/*------------------------------------------------------------------------
 * SetSAHandPoster
 *
 * Description: 
 *
 * Reports:      OK
 *              S_lwr_CANNOT_FIND_POSTER
 *              S_lwr_CANNOT_READ_POSTER
 */

/* lwrSetSAHandPosterMain  -  codel EXEC of SetSAHandPoster
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrSetSAHandPosterMain(GEN_STRING128 *sahandPosterName, int *report)
{
  /* Find the external poster and try it by reading once */
  if(lwrGENMANIP_SAHAND_STRPosterFind(sahandPosterName->name,
                                      &posterSAHandId) == ERROR) {
    *report =  S_lwr_CANNOT_FIND_POSTER;
  }
  else if(lwrGENMANIP_SAHAND_STRPosterRead(posterSAHandId,
                                           &SDI_F->sahandPos) == ERROR) {
    *report =  S_lwr_CANNOT_READ_POSTER;
  }
  else {
    SDI_F->status.sahandUpdate = GEN_TRUE;  
    *report = OK;
  }
  return ETHER;
}

/*------------------------------------------------------------------------
 * SetHumanDistancePoster
 *
 * Description: 
 *
 * Reports:      OK
 *              S_lwr_CANNOT_FIND_POSTER
 *              S_lwr_CANNOT_READ_POSTER
 */

/* lwrSetHumanDistancePosterMain  -  codel EXEC of SetHumanDistancePoster
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrSetHumanDistancePosterMain(GEN_STRING128 *humanDistPosterName, int *report)
{
  /* Find the external poster and try it by reading once */
  if(lwrGENHUM_HUMAN_DISTANCEPosterFind(humanDistPosterName->name,
                                        &posterHumDistId) == ERROR) {
    *report =  S_lwr_CANNOT_FIND_POSTER;
  }
  else if(lwrGENHUM_HUMAN_DISTANCEPosterRead(posterHumDistId,
                                             &SDI_F->humanDist) == ERROR) {
    *report =  S_lwr_CANNOT_READ_POSTER;
  }
  else {
    *report = OK;
  }

  return ETHER;
}




/*------------------------------------------------------------------------
 * SaveForces
 *
 * Description: 
 *
 * Reports:      OK
 *              S_lwr_FILE_NAME_TOO_LONG
 *              S_lwr_FILE_IO_ERROR
 */

/* lwrSaveForcesMain  -  codel EXEC of SaveForces
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrSaveForcesMain(GEN_STRING128 *ForcesFileName, int *report)
{
  FILE* filePtr = 0;

#ifdef LWR_FRI
  if(friInstLeft) {
    friInstLeft->setToKRLInt(CMD_CHANNEL, FRI_IDLE);
  }
  if(friInstRight) {
    friInstRight->setToKRLInt(CMD_CHANNEL, FRI_IDLE);
  }
#endif
  if ((filePtr = fopen(ForcesFileName->name,"wb")) == NULL) {
    printf("ERROR Open File %s\n", ForcesFileName->name);
    *report = S_lwr_FILE_IO_ERROR;
    return ETHER;
  } else {
    fprintf(filePtr,"# i L_Fx L_Fy L_Fz L_Mx L_My L_Mz R_Fx R_Fy R_Fz R_Mx R_My R_Mz ;\n");
    for(int i=0; i < LWR_FORCES_WINDOW; i++) {
      fprintf(filePtr,"%d %g %g %g %g %g %g %g %g %g %g %g %g \n", i, 
              SDI_F->armLeft.estExtTcpFTHistory[i][0],
              SDI_F->armLeft.estExtTcpFTHistory[i][1],
              SDI_F->armLeft.estExtTcpFTHistory[i][2],
              SDI_F->armLeft.estExtTcpFTHistory[i][3],
              SDI_F->armLeft.estExtTcpFTHistory[i][4],
              SDI_F->armLeft.estExtTcpFTHistory[i][5],
              SDI_F->armRight.estExtTcpFTHistory[i][0],
              SDI_F->armRight.estExtTcpFTHistory[i][1],
              SDI_F->armRight.estExtTcpFTHistory[i][2],
              SDI_F->armRight.estExtTcpFTHistory[i][3],
              SDI_F->armRight.estExtTcpFTHistory[i][4],
              SDI_F->armRight.estExtTcpFTHistory[i][5]);
    }
    fclose(filePtr);
  }
  return ETHER;
}




/*------------------------------------------------------------------------
 * SetCollisionEnvMode
 *
 * Description: 
 *
 * Reports:   S_lwr_POSTER_NOT_SETTED    
 *            OK
 */

/* lwrSetCollisionEnvModeMain  -  codel EXEC of SetCollisionEnvMode
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrSetCollisionEnvModeMain(GEN_BOOL *collisionEnvMode, int *report)
{
  if(posterMhpCollideId == NULL) {
    SDI_F->status.collisionEnvMode = GEN_FALSE;
    printf("WARN: lwrSetCollisionEnvModeMain the collisionEnv poster is not setted\n");
    *report =  S_lwr_POSTER_NOT_SETTED;
  } else {
    SDI_F->status.collisionEnvMode = GEN_TRUE;
  }
  return ETHER;
}




/*------------------------------------------------------------------------
 * TrackTrajQ
 *
 * Description: 
 *
 * Reports:      OK
 *              S_lwr_NOT_INITIALIZED
 *              S_lwr_FRI_MONITOR_MODE
 *              S_lwr_POSTER_NOT_FOUND
 *              S_lwr_POSTER_READ_ERROR
 *              S_lwr_CMD_STOP_TRACK
 *              S_lwr_NB_POINTS
 *              S_lwr_ERROR_POSITION
 *              S_lwr_BAD_MODE
 *              S_lwr_FILE_NOT_FOUND
 *              S_lwr_ERROR_INIT_POSITION
 *              S_lwr_Q_LIMITED
 */

/* lwrTrackTrajQStart  -  codel START of TrackTrajQ
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrTrackTrajQStart(LWR_TRACK_STR *trackStr, int *report)
{
  if(SDI_F->status.simulation == GEN_TRUE) {
    return EXEC;
  }
  
#ifdef LWR_FRI
  ACTIVITY_EVENT eventLeft, eventRight;
  if(SDI_F->armLeft.is_used == GEN_TRUE) {
    eventLeft = requestFriStateCmd(report, LWR_ARM_LEFT, LWR_CTRL_POSITION);
  } else {
    eventLeft = EXEC;
  }
  if(SDI_F->armRight.is_used == GEN_TRUE) {
    eventRight = requestFriStateCmd(report, LWR_ARM_RIGHT, LWR_CTRL_POSITION);
  } else {
    eventRight = EXEC;
  }

  if(eventLeft == ETHER || eventRight == ETHER) { 
    return ETHER;
  }
  if(eventLeft == START || eventRight == START) {
    if(eventLeft == START ) {
      printf("lwrGotoQRelStart waiting for arm Left\n");
    }
    if(eventRight == START ) {
      printf("lwrGotoQRelStart waiting for arm Right\n");
    }
    return START;
  }
  if(eventLeft == EXEC && eventRight == EXEC) {
    printf("lwrTrackTrajQStart arms are ready\n");
    //return EXEC;
  }
#endif

  currentMotion.timeScaleCond.a = 0.0;
  currentMotion.timeScaleCond.v = SDI_F->timeScale;
  currentMotion.timeScaleCond.x = 0.0;

  switch(trackStr->trackMode) {
  case LWR_TRACK_FILE:
    printf("ERROR: lwrTrackTraj works only with poster \n");
    *report = S_lwr_BAD_MODE;
    return ETHER;
    break;

  case LWR_TRACK_POSTER:
    if(lwrSM_TRAJ_STRPosterFind (trackStr->posterName.name, &lwr_poster_q) != OK) {
      *report = S_lwr_POSTER_NOT_FOUND;
      return ETHER;
    }
    if(lwrSM_TRAJ_STRPosterRead (lwr_poster_q, &smTraj) != OK) {
      *report = S_lwr_POSTER_READ_ERROR;
      printf("ERROR: %s poster read failed", trackStr->posterName.name);
      return ETHER;
    }
    if(smTraj.nbAxis != LWR_NB_CONTROLLED_JNTS) {
      printf("ERROR: lwr::TrackQ traj.nbAxis (%d) != %d \n", smTraj.nbAxis,  LWR_NB_CONTROLLED_JNTS);
      return ETHER;
    }
    currentMotion.traj.clear();
    currentMotion.traj.importFromSM_TRAJ_STR(&smTraj);

    break;
  default:
    *report = S_lwr_BAD_MODE;
    return ETHER;
  }

  currentMotion.traj.computeTimeOnTraj();

  /* check initial position */
  Gb_q7 errorMax;
  std::vector<SM_COND> cond;
  errorMax.q1 = LWR_A1_MAX_VELOCITY/10.0;
  errorMax.q2 = LWR_A2_MAX_VELOCITY/10.0;
  errorMax.q3 = LWR_E1_MAX_VELOCITY/10.0;
  errorMax.q4 = LWR_A3_MAX_VELOCITY/10.0;
  errorMax.q5 = LWR_A4_MAX_VELOCITY/10.0;
  errorMax.q6 = LWR_A5_MAX_VELOCITY/10.0;
  errorMax.q7 = LWR_A6_MAX_VELOCITY/10.0;

  if(SDI_F->timeScale < 0.0) {
    SDI_F->timeScale = 1.0;
    printf("WARN: lwr::TrackQ set timeScale to 1.0\n");
  }
  currentMotion.traj.getMotionCond(0.0, cond);
  /* check Left arm */
  if(SDI_F->armLeft.is_used == GEN_TRUE) {
    if ((fabs(SDI_F->armLeft.currConf.q1 - cond[0+SDI_F->armLeft.indexInTraj].x) > errorMax.q1) ||
        (fabs(SDI_F->armLeft.currConf.q2 - cond[1+SDI_F->armLeft.indexInTraj].x) > errorMax.q2) ||
        (fabs(SDI_F->armLeft.currConf.q3 - cond[2+SDI_F->armLeft.indexInTraj].x) > errorMax.q3) ||
        (fabs(SDI_F->armLeft.currConf.q4 - cond[3+SDI_F->armLeft.indexInTraj].x) > errorMax.q4) ||
        (fabs(SDI_F->armLeft.currConf.q5 - cond[4+SDI_F->armLeft.indexInTraj].x) > errorMax.q5) ||
        (fabs(SDI_F->armLeft.currConf.q6 - cond[5+SDI_F->armLeft.indexInTraj].x) > errorMax.q6) ||
        (fabs(SDI_F->armLeft.currConf.q7 - cond[6+SDI_F->armLeft.indexInTraj].x) > errorMax.q7)) {
      *report = S_lwr_ERROR_INIT_POSITION;
      return ETHER;
    }
  }
  if(SDI_F->armRight.is_used == GEN_TRUE) {
    if ((fabs(SDI_F->armRight.currConf.q1 - cond[0+SDI_F->armRight.indexInTraj].x) > errorMax.q1) ||
        (fabs(SDI_F->armRight.currConf.q2 - cond[1+SDI_F->armRight.indexInTraj].x) > errorMax.q2) ||
        (fabs(SDI_F->armRight.currConf.q3 - cond[2+SDI_F->armRight.indexInTraj].x) > errorMax.q3) ||
        (fabs(SDI_F->armRight.currConf.q4 - cond[3+SDI_F->armRight.indexInTraj].x) > errorMax.q4) ||
        (fabs(SDI_F->armRight.currConf.q5 - cond[4+SDI_F->armRight.indexInTraj].x) > errorMax.q5) ||
        (fabs(SDI_F->armRight.currConf.q6 - cond[5+SDI_F->armRight.indexInTraj].x) > errorMax.q6) ||
        (fabs(SDI_F->armRight.currConf.q7 - cond[6+SDI_F->armRight.indexInTraj].x) > errorMax.q7)) {
      *report = S_lwr_ERROR_INIT_POSITION;
      return ETHER;
    }
  }


  /* compute the maximum timeScale allowed */
  std::vector<double> maxVel;
  maxVel.resize(7);    
  maxVel[0] = LWR_A1_MAX_VELOCITY;
  maxVel[1] = LWR_A2_MAX_VELOCITY;
  maxVel[2] = LWR_E1_MAX_VELOCITY;
  maxVel[3] = LWR_A3_MAX_VELOCITY;
  maxVel[4] = LWR_A4_MAX_VELOCITY;
  maxVel[5] = LWR_A5_MAX_VELOCITY;
  maxVel[6] = LWR_A6_MAX_VELOCITY;

  currentMotion.traj.computeMaxTimeScaleVector(maxVel, LWR_PERIOD, limitsTimeScale);

  currentMotion.timeScaleCond.a = 0.0;
  currentMotion.timeScaleCond.v = MIN((currentMotion.traj.getTsVec())[0], SDI_F->timeScale);
  currentMotion.timeScaleCond.x = 0.0;

  // currentMotion.traj.print();

  currentMotion.curTime = 0.0;
  SDI_F->armLeft.cntrl_mode = LWR_OPEN_LOOP_CNTRL;
  SDI_F->armRight.cntrl_mode = LWR_OPEN_LOOP_CNTRL;

  setMotionStart(LWR_POSITION);
  return EXEC;
}

/* lwrTrackTrajQMain  -  codel EXEC of TrackTrajQ
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrTrackTrajQMain(LWR_TRACK_STR *trackStr, int *report)
{
  if(currentMotion.type == LWR_NOT_MOVE) {
    return END;
  }

  if(lwrSM_TRAJ_STRPosterRead (lwr_poster_q, &smTraj) != OK) {
    *report = S_lwr_POSTER_READ_ERROR;
    printf("ERROR: %s poster read failed", trackStr->posterName.name);
    return ETHER;
  }
  if(smTraj.nbAxis != LWR_NB_CONTROLLED_JNTS) {
    printf("ERROR: lwr::TrackQ traj.nbAxis (%d) != 24 \n", smTraj.nbAxis);
    return ETHER;
  }
  

  if(smTraj.trajId != currentMotion.traj.getTrajId()) {
    if(smTraj.timePreserved > currentMotion.curTime) {

      SM_TRAJ &buffRef =  trajBuff1;
      if(buffInUse == 1) {
        buffRef = trajBuff2;
        buffInUse = 2;
      }
      else {
        buffInUse = 1;
      }

      /* replace the current trajectory */
      buffRef.importFromSM_TRAJ_STR(&smTraj);

      buffRef.computeTimeOnTraj();
      printf("New trajectory loaded duration %f currtime %f  \n", buffRef.getDuration(), currentMotion.curTime);

      /* compute the maximum timeScale allowed */
      std::vector<double> maxVel;
      maxVel.resize(7);    
      maxVel[0] = LWR_A1_MAX_VELOCITY;
      maxVel[1] = LWR_A2_MAX_VELOCITY;
      maxVel[2] = LWR_E1_MAX_VELOCITY;
      maxVel[3] = LWR_A3_MAX_VELOCITY;
      maxVel[4] = LWR_A4_MAX_VELOCITY;
      maxVel[5] = LWR_A5_MAX_VELOCITY;
      maxVel[6] = LWR_A6_MAX_VELOCITY;

      buffRef.computeMaxTimeScaleVector(maxVel, LWR_PERIOD, limitsTimeScale);

      currentMotion.traj = buffRef;

    }
  }
  return EXEC;
}


/* lwrTrackTrajQEnd  -  codel END of TrackTrajQ
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrTrackTrajQEnd(LWR_TRACK_STR *trackStr, int *report)
{
  printf("INFO lwrTrackTrajQEnd Motion Complete \n");
  return ETHER;
}

/* lwrTrackTrajQInter  -  codel INTER of TrackTrajQ
   Returns:  INTER ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrTrackTrajQInter(LWR_TRACK_STR *trackStr, int *report)
{
  /* ... add your code here ... */
  return ETHER;
}




/*------------------------------------------------------------------------
 * GetDep
 *
 * Description: 
 *
 * Reports:      OK
 */

/* lwrGetDepMain  -  codel EXEC of GetDep
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrGetDepMain(Gb_dep *currDep, int *report)
{
  Gb_dep dep;
  Gb_th th_R7, th_R8, th_Rt;

  fri2th_cartPos(SDI_F->armRight.currentCartPos, &th_R7);
  th_R7.vp.z = th_R7.vp.z- LWR_R1;
  Gb_th_print(&SDI_F->armRight.th_base_0,"base");
  Gb_th_print(&th_R7,"arm");
  
  Gb_th_produit(&SDI_F->armRight.th_base_0, &th_R7, &th_R8);
  //Gb_th_produit(&th_R8, &(SDI_F->armRight.th_8_tool), &th_Rt);
  Gb_th_dep(&th_R8, &dep);

  Gb_th_print(&th_R8,"th_R8");
  //Gb_th_print(&th_Rt, "th_Rt");
  printf("Displacement of the right arm\n");
  Gb_dep_print(&dep);
  *currDep = dep;
  return ETHER;
}

/*------------------------------------------------------------------------
 * GetDepDagger
 *
 * Description: 
 *
 * Reports:      OK
 */

/* lwrGetDepDaggerMain  -  codel EXEC of GetDep
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrGetDepDaggerMain(Gb_dep *currDepDagger, int *report)
{
  Gb_dep depDagger;
  Gb_th th_R0_R7, th_Base_R7, th_R7_Dagger, th_Base_Dagger;

  th_R7_Dagger.vx.x = 1.0;
  th_R7_Dagger.vx.y = 0.0;
  th_R7_Dagger.vx.z = 0.0;

  th_R7_Dagger.vy.x = 0.0;
  th_R7_Dagger.vy.y = 1.0;
  th_R7_Dagger.vy.z = 0.0;

  th_R7_Dagger.vz.x = 0.0;
  th_R7_Dagger.vz.y = 0.0;
  th_R7_Dagger.vz.z = 1.0;

  th_R7_Dagger.vp.x = -0.00142;
  th_R7_Dagger.vp.y = -0.00283;
  th_R7_Dagger.vp.z =  0.17981;


  fri2th_cartPos(SDI_F->armRight.currentCartPos, &th_R0_R7);
  th_R0_R7.vp.z = th_R0_R7.vp.z- LWR_R1;
  Gb_th_print(&SDI_F->armRight.th_base_0,"base");
  Gb_th_print(&th_R0_R7,"arm");
  
  Gb_th_produit(&SDI_F->armRight.th_base_0, &th_R0_R7, &th_Base_R7);

  Gb_th_produit(&th_Base_R7, &th_R7_Dagger, &th_Base_Dagger);
  Gb_th_dep(&th_Base_Dagger, &depDagger);

  Gb_th_print(&th_Base_Dagger,"th_Base_Dagger");
  //Gb_th_print(&th_Rt, "th_Rt");
  printf("Displacement of Dagger of the right arm\n");
  Gb_dep_print(&depDagger);
  *currDepDagger = depDagger;
  return ETHER;
}

/*------------------------------------------------------------------------
 * GetDepTestBoard
 *
 * Description: 
 *
 * Reports:      OK
 */

/* lwrGetDepTestBoardMain  -  codel EXEC of GetDep
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrGetDepTestBoardMain(Gb_dep *currDepTestBoard, int *report)
{
  Gb_dep depTestBoard;
  Gb_th th_R0_R7, th_Base_R7, th_R7_TestBoard, th_Base_TestBoard;

  th_R7_TestBoard.vx.x = -0.6816387600;
  th_R7_TestBoard.vx.y = -0.73168886887;
  th_R7_TestBoard.vx.z = 0.0;

  th_R7_TestBoard.vy.x = 0.0;
  th_R7_TestBoard.vy.y = 0.0;
  th_R7_TestBoard.vy.z = 1.0;

  th_R7_TestBoard.vz.x = -0.73168886887;
  th_R7_TestBoard.vz.y = 0.6816387600;
  th_R7_TestBoard.vz.z = 0.0;

  th_R7_TestBoard.vp.x =  0.00;
  th_R7_TestBoard.vp.y =  0.00;
  th_R7_TestBoard.vp.z =  0.170;


  fri2th_cartPos(SDI_F->armRight.currentCartPos, &th_R0_R7);
  th_R0_R7.vp.z = th_R0_R7.vp.z- LWR_R1;
  Gb_th_print(&SDI_F->armRight.th_base_0,"base");
  Gb_th_print(&th_R0_R7,"arm");

  Gb_th_produit(&SDI_F->armRight.th_base_0, &th_R0_R7, &th_Base_R7);

  Gb_th_produit(&th_Base_R7, &th_R7_TestBoard, &th_Base_TestBoard);
  Gb_th_dep(&th_Base_TestBoard, &depTestBoard);

  Gb_th_print(&th_Base_TestBoard,"th_Base_TestBoard");
  //Gb_th_print(&th_Rt, "th_Rt");
  printf("Displacement of TestBoard of the right arm\n");
  Gb_dep_print(&depTestBoard);
  *currDepTestBoard = depTestBoard;
  return ETHER;
}


/*------------------------------------------------------------------------
 * SetTimeScaleCntrts
 *
 * Description: 
 *
 * Reports:      OK
 */

/* lwrSetTimeScaleCntrtsMain  -  codel EXEC of SetTimeScaleCntrts
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrSetTimeScaleCntrtsMain(LWR_TIMESCALE_CNTRTS *timeScaleCntrts, int *report)
{

  SDI_F->timeScaleCntrts.jerk = timeScaleCntrts->jerk;
  SDI_F->timeScaleCntrts.acc = timeScaleCntrts->acc;

  limitsTimeScale.maxJerk =  SDI_F->timeScaleCntrts.jerk;
  limitsTimeScale.maxAcc  =  SDI_F->timeScaleCntrts.acc;
  limitsTimeScale.maxVel  = SDI_F->timeScaleMax;

  return ETHER;
}

/*------------------------------------------------------------------------
 * GotoXXAbs
 *
 * Description: 
 *
 * Reports:      OK
 *              S_lwr_NOT_INITIALIZED
 *              S_lwr_WRONG_VALUE
 *              S_lwr_SOFTMOTION_ERROR
 *              S_lwr_FRI_MONITOR_MODE
 *              S_lwr_MGI_CALC_ERROR
 *              S_lwr_MGI_SINGULARITY
 *              S_lwr_JNT_OUT_OF_RANGE
 *              S_lwr_COLLISION_RISK
 */

/* lwrGotoXXAbsStart  -  codel START of GotoXXAbs
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrGotoXXAbsStart(LWR_GOTO_POSE_STR *gotoPose, int *report)
{
  if(SDI_F->status.simulation == GEN_TRUE) {
    return EXEC;
  }

#ifdef LWR_FRI
  ACTIVITY_EVENT eventLeft, eventRight;
  if(SDI_F->armLeft.is_used == GEN_TRUE) {
    eventLeft = requestFriStateCmd(report, LWR_ARM_LEFT, LWR_CTRL_POSITION);
  } else {
    eventLeft = EXEC;
  }
  if(SDI_F->armRight.is_used == GEN_TRUE) {
    eventRight = requestFriStateCmd(report, LWR_ARM_RIGHT, LWR_CTRL_POSITION);
  } else {
    eventRight = EXEC;
  }

  if(eventLeft == EXEC && eventRight == EXEC) {
    printf("lwrGotoXAbsStart OK\n");
    return EXEC;
  }
  if(eventLeft == ETHER || eventRight == ETHER) {
    return ETHER;
  }
  if(eventLeft == START || eventRight == START) {
    if(eventLeft == START ) {
      printf("lwrGotoXAbsStart waiting for arm Left\n");
    }
    if(eventRight == START ) {
      printf("lwrGotoXAbsStart waiting for arm Right\n");
    }
      return START;
}
#else
  return EXEC;
#endif
  return ETHER;
}

/* lwrGotoXXAbsMain  -  codel EXEC of GotoXXAbs
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrGotoXXAbsMain(LWR_GOTO_POSE_STR *gotoPose, int *report)
{
  //frames : Rob : robot base
  //         Base : robot base Kuka
  //         R0 : robot base for gbM
  //         R7 : robot end for gbM
  //         R8 : flange (end for Kuka)
  //         Rt : tool frame
  // goal correspond at the pose of the tool in the robot frame
  Gb_dep gotoDep = gotoPose->poseRight;
  if (! normalize( gotoDep.rx,  gotoDep.ry,  gotoDep.rz,
                   &gotoDep.rx, &gotoDep.ry, &gotoDep.rz)) {
    printf("ERROR: wrong (maybe null) vector is provided for tool orientation\n");
    *report = S_lwr_WRONG_VALUE;
    return ETHER;
  }

  Gb_th  th_Rob_Rt;
  Gb_dep_th(&gotoDep, &th_Rob_Rt);

  Gb_th th_R7_R8, th_R8_R7, th_Rt_R8, th_Rob_R8, th_R0_Rob, 
    th_R0_R8, th_Base_R8_cur, th_R0_R8_cur, th_R0_R7, th_R0_R7_cur;
  Gb_quat quat_R0_R7_cur, quat_R0_R7;
  Gb_dep dep_R0_R7, dep_R0_R7_cur, dep_R7_cur_R7;
  // 7 is the center of R7(gbM) and 8 is flange frame
  th_R7_R8.vx.x =  1;
  th_R7_R8.vy.x =  0;
  th_R7_R8.vz.x =  0;
  th_R7_R8.vp.x =  0;
  th_R7_R8.vx.y =  0;
  th_R7_R8.vy.y =  1;
  th_R7_R8.vz.y =  0;
  th_R7_R8.vp.y =  0;
  th_R7_R8.vx.z =  0;
  th_R7_R8.vy.z =  0;
  th_R7_R8.vz.z =  1;
  th_R7_R8.vp.z = LWR_R7;

  Gb_th_inverse(&th_R7_R8, &th_R8_R7);
  
  //compute the goal position of the flange in the robot frame
  Gb_th_inverse(&SDI_F->armRight.th_8_tool,  &th_Rt_R8); 
  Gb_th_produit(&th_Rob_Rt, &th_Rt_R8, &th_Rob_R8);


  // verifier th_base
  Gb_th_inverse(&SDI_F->armRight.th_base_0, &th_R0_Rob);
  Gb_th_produit(&th_R0_Rob, &th_Rob_R8, &th_R0_R8);

  Gb_th_produit(&th_R0_R8, &th_R8_R7, &th_R0_R7);
  Gb_th_dep(&th_R0_R7, &dep_R0_R7);
  //printf("Dep_R0_R7 goal\n");
  //Gb_dep_print(&dep_R0_R7);

  // Get the dep of the flange in the R0 (J1) frame
  fri2th_cartPos(SDI_F->armRight.currentCartPos, &th_Base_R8_cur);
  
  th_R0_R8_cur = th_Base_R8_cur;
  th_R0_R8_cur.vp.z -= LWR_R1;
  Gb_th_produit(&th_R0_R8_cur, &th_R8_R7, &th_R0_R7_cur);
  Gb_th_dep(&th_R0_R7_cur, &dep_R0_R7_cur);
  

  // compute the relative dep
  Gb_th_quat(&th_R0_R7, &quat_R0_R7);
  Gb_th_quat(&th_R0_R7_cur, &quat_R0_R7_cur);
  Gb_quat_compute_relativeDep_to_interpole(&quat_R0_R7_cur, &quat_R0_R7, &dep_R7_cur_R7);
  
  printf("Dep_R7_cur_R7\n");
  Gb_dep_print(&dep_R7_cur_R7);

  //compute the softMotion trajectory in the cartesian frame
  std::vector<SM_COND> ic, fc;
  ic.resize(12);
  fc.resize(12);


  for(unsigned int o=0; o<12; o++) {
    ic[o].a = 0;
    ic[o].v = 0;
    ic[o].x = 0;
  }

  if(SDI_F->armLeft.is_used == GEN_TRUE) {
    printf("lwrGotoQAbsMain Arm Left is_used\n");
    fc[0].x = ic[0].x;
    fc[1].x = ic[1].x;
    fc[2].x = ic[2].x;
    fc[3].x = ic[3].x;
    fc[4].x = ic[4].x;
    fc[5].x = ic[5].x;
  } else {
    fc[0].x = ic[0].x;
    fc[1].x = ic[1].x;
    fc[2].x = ic[2].x;
    fc[3].x = ic[3].x;
    fc[4].x = ic[4].x;
    fc[5].x = ic[5].x;

  }
  if(SDI_F->armRight.is_used == GEN_TRUE) {
    printf("lwrGotoXXAbsMain Arm Right is_used\n");

    fc[6].a = 0;
    fc[6].v = 0;
    fc[6].x = dep_R7_cur_R7.x;
    fc[7].a = 0;
    fc[7].v = 0;
    fc[7].x = dep_R7_cur_R7.y;
    fc[8].a = 0;
    fc[8].v = 0;
    fc[8].x = dep_R7_cur_R7.z;
    fc[9].a = 0;
    fc[9].v = 0;
    fc[9].x = dep_R7_cur_R7.rx * dep_R7_cur_R7.a;
    fc[10].a = 0;
    fc[10].v = 0;
    fc[10].x = dep_R7_cur_R7.ry * dep_R7_cur_R7.a;
    fc[11].a = 0;
    fc[11].v = 0;
    fc[11].x = dep_R7_cur_R7.rz * dep_R7_cur_R7.a;
  } else{
    fc[6].x  =  ic[6].x  ;
    fc[7].x  =  ic[7].x  ;
    fc[8].x  =  ic[8].x  ;
    fc[9].x  =  ic[9].x  ;
    fc[10].x =  ic[10].x ;
    fc[11].x =  ic[11].x ;

    fc[6].v  =  0.0;
    fc[7].v  =  0.0;
    fc[8].v  =  0.0;
    fc[9].v  =  0.0;
    fc[10].v =  0.0;
    fc[11].v =  0.0;
    fc[6].a  =  0.0;
    fc[7].a  =  0.0;
    fc[8].a  =  0.0;
    fc[9].a  =  0.0;
    fc[10].a =  0.0;
    fc[11].a =  0.0;
  }

  currentMotion.traj_x.clear();
  
  currentMotion.traj_x.computeTraj(ic, fc, currentMotion.limits_x, SM_TRAJ::SM_INDEPENDANT);
  currentMotion.traj_x_init_R0_R7 = th_R0_R7_cur;

  currentMotion.traj_x.save("totototototo.traj");
  
  currentMotion.curTime = 0.0;
  SDI_F->armLeft.cntrl_mode = LWR_XOPEN_LOOP_CNTRL;
  SDI_F->armRight.cntrl_mode = LWR_XOPEN_LOOP_CNTRL;
  setMotionStart(LWR_POSITION);

  return END;
}


/* lwrGotoXXAbsEnd  -  codel END of GotoXXAbs
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrGotoXXAbsEnd(LWR_GOTO_POSE_STR *gotoPose, int *report)
{
  if ((currentMotion.state == LWR_STARTED) || (currentMotion.state == LWR_MOVING)){
    return END;
  }
  fillMotionReport(currentMotion.error, report);
  return ETHER;
}




/*
 *
 *
 *
 *
 *
 *    
 *
 *
 *
 */






/*------------------------------------------------------------------------
 * GotoClosedLoopXXAbs
 *
 * Description: 
 *
 * Reports:      OK
 *              S_lwr_NOT_INITIALIZED
 *              S_lwr_WRONG_VALUE
 *              S_lwr_SOFTMOTION_ERROR
 *              S_lwr_FRI_MONITOR_MODE
 *              S_lwr_MGI_CALC_ERROR
 *              S_lwr_MGI_SINGULARITY
 *              S_lwr_JNT_OUT_OF_RANGE
 *              S_lwr_COLLISION_RISK
 */

/* lwrGotoClosedLoopXXAbsStart  -  codel START of GotoClosedLoopXXAbs
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrGotoClosedLoopXXAbsStart(LWR_GOTO_POSE_STR *gotoPose, int *report)
{
  if(SDI_F->status.simulation == GEN_TRUE) {
    return EXEC;
  }

#ifdef LWR_FRI
  ACTIVITY_EVENT eventLeft, eventRight;
  if(SDI_F->armLeft.is_used == GEN_TRUE) {
    eventLeft = requestFriStateCmd(report, LWR_ARM_LEFT, LWR_CTRL_POSITION);
  } else {
    eventLeft = EXEC;
  }
  if(SDI_F->armRight.is_used == GEN_TRUE) {
    eventRight = requestFriStateCmd(report, LWR_ARM_RIGHT, LWR_CTRL_POSITION);
  } else {
    eventRight = EXEC;
  }

  if(eventLeft == EXEC && eventRight == EXEC) {
    printf("lwrGotoClosedLoopXXAbsStart OK\n");
    return EXEC;
  }
  if(eventLeft == ETHER || eventRight == ETHER) {
    return ETHER;
  }
  if(eventLeft == START || eventRight == START) {
    if(eventLeft == START ) {
      printf("lwrGotoClosedLoopXXAbsStart waiting for arm Left\n");
    }
    if(eventRight == START ) {
      printf("lwrGotoClosedLoopXXAbsStart waiting for arm Right\n");
    }
      return START;
}
#else
  return EXEC;
#endif
  return ETHER;
}

/* lwrGotoClosedLoopXXAbsMain  -  codel EXEC of GotoClosedLoopXXAbs
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrGotoClosedLoopXXAbsMain(LWR_GOTO_POSE_STR *gotoPose, int *report)
{
  //frames : Rob : robot base
  //         Base : robot base Kuka
  //         R0 : robot base for gbM
  //         R7 : robot end for gbM
  //         R8 : flange (end for Kuka)
  //         Rt : tool frame
  // goal correspond at the pose of the tool in the robot frame
  Gb_dep gotoDep = gotoPose->poseRight;
  if (! normalize( gotoDep.rx,  gotoDep.ry,  gotoDep.rz,
                   &gotoDep.rx, &gotoDep.ry, &gotoDep.rz)) {
    printf("ERROR: in codel lwrGotoClosedLoopXXAbsMain, wrong (maybe null) vector is provided for tool orientation\n");
*report = S_lwr_WRONG_VALUE;
  return ETHER;
  }

  Gb_th  th_Rob_Rt;
  Gb_dep_th(&gotoDep, &th_Rob_Rt);

  Gb_th th_R7_R8, th_R8_R7, th_Rt_R8, th_Rob_R8, th_R0_Rob, 
    th_R0_R8, th_Base_R8_cur, th_R0_R8_cur, th_R0_R7, th_R0_R7_cur;
  Gb_quat quat_R0_R7_cur, quat_R0_R7;
  Gb_dep dep_R0_R7, dep_R0_R7_cur, dep_R7_cur_R7;
  // 7 is the center of R7(gbM) and 8 is flange frame
  th_R7_R8.vx.x =  1;
  th_R7_R8.vy.x =  0;
  th_R7_R8.vz.x =  0;
  th_R7_R8.vp.x =  0;
  th_R7_R8.vx.y =  0;
  th_R7_R8.vy.y =  1;
  th_R7_R8.vz.y =  0;
  th_R7_R8.vp.y =  0;
  th_R7_R8.vx.z =  0;
  th_R7_R8.vy.z =  0;
  th_R7_R8.vz.z =  1;
  th_R7_R8.vp.z = LWR_R7;

  Gb_th_inverse(&th_R7_R8, &th_R8_R7);
  
  //compute the goal position of the flange in the robot frame
  Gb_th_inverse(&SDI_F->armRight.th_8_tool,  &th_Rt_R8); 
  Gb_th_produit(&th_Rob_Rt, &th_Rt_R8, &th_Rob_R8);


  // verifier th_base
  Gb_th_inverse(&SDI_F->armRight.th_base_0, &th_R0_Rob);
  Gb_th_produit(&th_R0_Rob, &th_Rob_R8, &th_R0_R8);

  Gb_th_produit(&th_R0_R8, &th_R8_R7, &th_R0_R7);
  
  currentMotion.th_R0_R7_Goto = th_R0_R7; //update goal
  currentMotion.curTime = 0.0;
  
  Gb_th_dep(&th_R0_R7, &dep_R0_R7);
  //printf("Dep_R0_R7 goal\n");
  //Gb_dep_print(&dep_R0_R7);

  //set control mode
  SDI_F->armLeft.cntrl_mode = LWR_XCLOSED_LOOP_CNTRL;
  SDI_F->armRight.cntrl_mode = LWR_XCLOSED_LOOP_CNTRL;
  setMotionStart(LWR_POSITION);

  //printf("In codel GotoClosedLoopXXAbs;\n");
  return END;
}


/* lwrGotoClosedLoopXXAbsEnd  -  codel END of GotoClosedLoopXXAbs
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrGotoClosedLoopXXAbsEnd(LWR_GOTO_POSE_STR *gotoPose, int *report)
{
  if ((currentMotion.state == LWR_STARTED) || (currentMotion.state == LWR_MOVING)){
    return END;
  }
  fillMotionReport(currentMotion.error, report);
  return ETHER;
}



/* lwrSet_dep_Rrob_Rtool_gotoMain  -  codel EXEC of Set_dep_Rrob_Rtool_goto
   Returns:  EXEC ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrSet_dep_Rrob_Rtool_gotoMain(LWR_GOTO_POSE_STR *gotoPose, int *report)
{
  //frames : Rob : robot base
  //         Base : robot base Kuka
  //         R0 : robot base for gbM
  //         R7 : robot end for gbM
  //         R8 : flange (end for Kuka)
  //         Rt : tool frame
  // goal correspond at the pose of the tool in the robot frame
  Gb_dep gotoDep = gotoPose->poseRight;
  if (! normalize( gotoDep.rx,  gotoDep.ry,  gotoDep.rz,
                   &gotoDep.rx, &gotoDep.ry, &gotoDep.rz)) {
    printf("ERROR: in codel lwrGotoClosedLoopXXAbsMain, wrong (maybe null) vector is provided for tool orientation\n");
*report = S_lwr_WRONG_VALUE;
  return ETHER;
  }

  Gb_th  th_Rob_Rt;
  Gb_dep_th(&gotoDep, &th_Rob_Rt);

  Gb_th th_R7_R8, th_R8_R7, th_Rt_R8, th_Rob_R8, th_R0_Rob, 
    th_R0_R8, th_Base_R8_cur, th_R0_R8_cur, th_R0_R7, th_R0_R7_cur;
  Gb_quat quat_R0_R7_cur, quat_R0_R7;
  Gb_dep dep_R0_R7, dep_R0_R7_cur, dep_R7_cur_R7;
  // 7 is the center of R7(gbM) and 8 is flange frame
  th_R7_R8.vx.x =  1;
  th_R7_R8.vy.x =  0;
  th_R7_R8.vz.x =  0;
  th_R7_R8.vp.x =  0;
  th_R7_R8.vx.y =  0;
  th_R7_R8.vy.y =  1;
  th_R7_R8.vz.y =  0;
  th_R7_R8.vp.y =  0;
  th_R7_R8.vx.z =  0;
  th_R7_R8.vy.z =  0;
  th_R7_R8.vz.z =  1;
  th_R7_R8.vp.z = LWR_R7;

  Gb_th_inverse(&th_R7_R8, &th_R8_R7);
  
  //compute the goal position of the flange in the robot frame
  Gb_th_inverse(&SDI_F->armRight.th_8_tool,  &th_Rt_R8); 
  Gb_th_produit(&th_Rob_Rt, &th_Rt_R8, &th_Rob_R8);


  // verifier th_base
  Gb_th_inverse(&SDI_F->armRight.th_base_0, &th_R0_Rob);
  Gb_th_produit(&th_R0_Rob, &th_Rob_R8, &th_R0_R8);

  Gb_th_produit(&th_R0_R8, &th_R8_R7, &th_R0_R7);
  
  currentMotion.th_R0_R7_Goto = th_R0_R7; //update goal
  
  return ETHER;
}

/*------------------------------------------------------------------------
 * AbortMotion
 *
 * Description: 
 *
 * Reports:      OK
 */

/* lwrAbortMotionMain  -  codel EXEC of AbortMotion
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrAbortMotionMain(int *report)
{
  setMotionEnd();
  SDI_F->posForMove3d.tau = 0.0;
  if(SDI_F->saveMotionFile == GEN_TRUE) {
    if(fileptr) {
      closingSaveFile = GEN_TRUE;
    }
  }
  printf("Motion aborted\n"); 
  return ETHER;
}



static int nb_loop_to_abort_motion = 0;
/*------------------------------------------------------------------------
 * AbortMotionSmooth
 *
 * Description: 
 *
 * Reports:      OK
 */

/* lwrAbortMotionSmoothStart  -  codel START of AbortMotionSmooth
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrAbortMotionSmoothStart(int *report)
{

  SDI_F->timeScale = 0.0;


  /* ... add your code here ... */
  return EXEC;
}

/* lwrAbortMotionSmoothMain  -  codel EXEC of AbortMotionSmooth
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrAbortMotionSmoothMain(int *report)
{


  static GEN_BOOL humanDistMode = SDI_F->humanDistMode;
  bool left_is_stopped = false , right_is_stopped = false;
  SDI_F->humanDistMode = GEN_FALSE;
  SDI_F->timeScale = 0.0;


  if(SDI_F->armLeft.is_used == GEN_TRUE) {
    left_is_stopped = lwrIsArmStopped(&(SDI_F->armLeft));
  } else {
    left_is_stopped = true;
  }

  if(SDI_F->armRight.is_used == GEN_TRUE) {
    right_is_stopped = lwrIsArmStopped(&(SDI_F->armRight));
  } else {
    right_is_stopped = true;
  }

  if(left_is_stopped && right_is_stopped) {
    if(humanDistMode == GEN_TRUE) {
      SDI_F->humanDistMode = GEN_TRUE;
    }
    return END;
  }
  nb_loop_to_abort_motion ++;
  return EXEC;
}

/* lwrAbortMotionSmoothEnd  -  codel END of AbortMotionSmooth
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrAbortMotionSmoothEnd(int *report)
{
  setMotionEnd();
  SDI_F->posForMove3d.tau = 0.0;
  if(SDI_F->saveMotionFile == GEN_TRUE) {
    if(fileptr) {
      closingSaveFile = GEN_TRUE;
    }
  }
  printf("Motion aborted within %f second\n", nb_loop_to_abort_motion*LWR_PERIOD); 
  return ETHER;
}



static POSTER_ID trackPosePosterId = NULL;

/*------------------------------------------------------------------------
 * TrackPosterPose
 *
 * Description: 
 *
 * Reports:      OK
 *              S_lwr_NOT_INITIALIZED
 *              S_lwr_WRONG_VALUE
 *              S_lwr_SOFTMOTION_ERROR
 *              S_lwr_FRI_MONITOR_MODE
 *              S_lwr_MGI_CALC_ERROR
 *              S_lwr_MGI_SINGULARITY
 *              S_lwr_JNT_OUT_OF_RANGE
 *              S_lwr_COLLISION_RISK
 */

/* lwrTrackPosterPoseStart  -  codel START of TrackPosterPose
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrTrackPosterPoseStart(GEN_STRING128 *posePosterName, int *report)
{
  if(SDI_F->status.simulation == GEN_TRUE) {
    return EXEC;
  }

#ifdef LWR_FRI
  ACTIVITY_EVENT eventLeft, eventRight;
  if(SDI_F->armLeft.is_used == GEN_TRUE) {
    eventLeft = requestFriStateCmd(report, LWR_ARM_LEFT, LWR_CTRL_POSITION);
  } else {
    eventLeft = EXEC;
  }
  if(SDI_F->armRight.is_used == GEN_TRUE) {
    eventRight = requestFriStateCmd(report, LWR_ARM_RIGHT, LWR_CTRL_POSITION);
  } else {
    eventRight = EXEC;
  }

  if(eventLeft == EXEC && eventRight == EXEC) {
    if(lwrGENMANIP_POSE_STRPosterFind (posePosterName->name, &trackPosePosterId) != OK) {
     *report = S_lwr_CANNOT_FIND_POSTER;
     return ETHER;
    }
    //lwrInitialize_th_R0_R7_inCurrentMotion
    Gb_th th_Base_R8_cur, th_R0_R7;
    fri2th_cartPos(SDI_F->armRight.currentCartPos, &th_Base_R8_cur);
    lwr_get_th_R0_R7_from_Gb_th_Base_R8(&th_Base_R8_cur, &th_R0_R7);
    currentMotion.th_R0_R7_Goto = th_R0_R7;
    SDI_F->armLeft.cntrl_mode = LWR_XCLOSED_LOOP_CNTRL;
    SDI_F->armRight.cntrl_mode = LWR_XCLOSED_LOOP_CNTRL;
    setMotionStart(LWR_POSITION);
    printf("lwrGotoClosedLoopXXAbsStart OK\n");
    return EXEC;
  }
  if(eventLeft == ETHER || eventRight == ETHER) {
    return ETHER;
  }
  if(eventLeft == START || eventRight == START) {
    if(eventLeft == START ) {
      printf("lwrGotoClosedLoopXXAbsStart waiting for arm Left\n");
    }
    if(eventRight == START ) {
      printf("lwrGotoClosedLoopXXAbsStart waiting for arm Right\n");
    }
    return START;
  }
#else
  //lwrInitialize_th_R0_R7_inCurrentMotion
  Gb_th th_Base_R8_cur, th_R0_R7;
  fri2th_cartPos(SDI_F->armRight.currentCartPos, &th_Base_R8_cur);
  lwr_get_th_R0_R7_from_Gb_th_Base_R8(&th_Base_R8_cur, &th_R0_R7);
  currentMotion.th_R0_R7_Goto = th_R0_R7;
  //set control mode
  SDI_F->armLeft.cntrl_mode = LWR_XCLOSED_LOOP_CNTRL;
  SDI_F->armRight.cntrl_mode = LWR_XCLOSED_LOOP_CNTRL;
  setMotionStart(LWR_POSITION);
  printf("lwrGotoClosedLoopXXAbsStart OK\n");
  return EXEC;
#endif
  return ETHER;
}

/* lwrTrackPosterPoseMain  -  codel EXEC of TrackPosterPose
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrTrackPosterPoseMain(GEN_STRING128 *posePosterName, int *report)
{

  GENMANIP_POSE_STR pose_Rrob_R8_poster;
  if(lwrGENMANIP_POSE_STRPosterRead (trackPosePosterId, &pose_Rrob_R8_poster) != OK) {
   *report = S_lwr_CANNOT_READ_POSTER;
   return END;
  }

  Gb_th  th_Rrob_Rt;
  Gb_th th_R7_R8, th_R8_R7, th_Rt_R8, th_Rrob_R8, th_R0_Rob,
  th_R0_R8, th_Base_R8_cur, th_R0_R8_cur, th_R0_R7, th_R0_R7_cur;
  Gb_quat quat_R0_R7_cur, quat_R0_R7;
  Gb_dep dep_Rrob_Rt, dep_R0_R7, dep_R0_R7_cur, dep_R7_cur_R7;
  //frames : Rob : robot base
  //         Base : robot base Kuka
  //         R0 : robot base for gbM
  //         R7 : robot end for gbM
  //         R8 : flange (end for Kuka)
  //         Rt : tool frame
  // goal correspond at the pose of the tool in the robot frame

  if(pose_Rrob_R8_poster.available == GEN_TRUE) {
    th_Rrob_Rt = pose_Rrob_R8_poster.th;

    // 7 is the center of R7(gbM) and 8 is flange frame
    th_R7_R8.vx.x =  1;
    th_R7_R8.vy.x =  0;
    th_R7_R8.vz.x =  0;
    th_R7_R8.vp.x =  0;
    th_R7_R8.vx.y =  0;
    th_R7_R8.vy.y =  1;
    th_R7_R8.vz.y =  0;
    th_R7_R8.vp.y =  0;
    th_R7_R8.vx.z =  0;
    th_R7_R8.vy.z =  0;
    th_R7_R8.vz.z =  1;
    th_R7_R8.vp.z = LWR_R7 + 0.05; // the robot will stay 5 cm in front of the object

    Gb_th_inverse(&th_R7_R8, &th_R8_R7);

    //compute the goal position of the flange in the robot frame
    Gb_th_inverse(&SDI_F->armRight.th_8_tool,  &th_Rt_R8);
    Gb_th_produit(&th_Rrob_Rt, &th_Rt_R8, &th_Rrob_R8);


    // verifier th_base
    Gb_th_inverse(&SDI_F->armRight.th_base_0, &th_R0_Rob);
    Gb_th_produit(&th_R0_Rob, &th_Rrob_R8, &th_R0_R8);

    Gb_th_produit(&th_R0_R8, &th_R8_R7, &th_R0_R7);
  } else {
    // set the goal pose as the current one
    fri2th_cartPos(SDI_F->armRight.currentCartPos, &th_Base_R8_cur);
    lwr_get_th_R0_R7_from_Gb_th_Base_R8(&th_Base_R8_cur, &th_R0_R7);
  }

  currentMotion.th_R0_R7_Goto = th_R0_R7; //update goal
  currentMotion.curTime = 0.0; // reinitialized but this time in the current mode does notting

  printf("In codel GotoClosedLoopXXAbs;\n");
  return EXEC;
}

/* lwrTrackPosterPoseEnd  -  codel END of TrackPosterPose
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrTrackPosterPoseEnd(GEN_STRING128 *posePosterName, int *report)
{
  setMotionEnd();
  SDI_F->posForMove3d.tau = 0.0;
  if(SDI_F->saveMotionFile == GEN_TRUE) {
    if(fileptr) {
      closingSaveFile = GEN_TRUE;
    }
  }
  printf("Motion TrackPosterPose terminated\n");
  return ETHER;
}

/*------------------------------------------------------------------------
 * ResetForceHistory
 *
 * Description: 
 *
 * Reports:      OK
 */

/* lwrResetForceHistoryMain  -  codel EXEC of ResetForceHistory
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrResetForceHistoryMain(int *report)
{

  for(int v=0; v< LWR_FORCES_WINDOW; v++) {
    for(int u=0; u < LWR_CART_VEC; u++) {
      SDI_F->armRight.estExtTcpFTHistory[v][u] = 0.0;
      SDI_F->armLeft.estExtTcpFTHistory[v][u] = 0.0;
    }
  }
  for(int v=0; v< JNT_TRQ_NB_TIC_MEAN ; v++) {
    for(int w=0; w< LWR_JOINTS; w++) {  
      SDI_F->armRight.estExtJntTrqHistory[w][v] = 0.0;
      SDI_F->armLeft.estExtJntTrqHistory[w][v] = 0.0;
    }
  }
  return ETHER;
}

