/*
 * Copyright (c) 2010 LAAS/CNRS                       --  Oct 2010
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
/*------------------------------------------------------------
 * 
 *                     --  Module  LWR  --
 * 
 *  Description:
 *  Creation date : Wed May 19 22:35:29 2010 
 *  Author: Xavier Broquere
 *  LAAS-CNRS, Toulouse FRANCE
 *------------------------------------------------------------*/

/**
 ** lwrCntrlTaskCodels.cc
 **
 ** Codels used by the control task lwrCntrlTask
 **
 ** Author: 
 ** Date: 
 **
 **/

#include <portLib.h>

#ifdef LWR_FRI
#include "kukaFri-libs/friremote.h"
#endif

#include "lwrConst.h"
#include "server/lwrHeader.h"

#include "lwrFunctions.h"


extern LWR_MOTION_STR currentMotion;
extern float currentCartStiff[LWR_CART_VEC];
extern float currentCartDamp[LWR_CART_VEC];

#ifdef LWR_FRI
extern friRemote * friInstLeft;
extern friRemote * friInstRight;
#endif

/*------------------------------------------------------------------------
 * lwrPrintInfoCntrl  -  control codel of CONTROL request PrintInfo
 *
 * Description:    
 * Report: OK
 *                 S_lwr_UNKNOW_ERROR
 *
 * Returns:    OK or ERROR
 */

STATUS
lwrPrintInfoCntrl(GEN_BOOL *printInfo, int *report)
{
  /* ... add your code here ... */
  return OK;
}

/*------------------------------------------------------------------------
 * lwrSetSpeedLimitCntrl  -  control codel of CONTROL request SpeedLimitSet
 *
 * Description:    Set speed limit (0 for no movement, 1 for max. speed)
 * Report: OK
 *                 S_lwr_OutOfRange
 *
 * Returns:    OK or ERROR
 */

STATUS
lwrSetSpeedLimitCntrl(double *speedLimit, int *report)
{
  /* The speed limit is a multiplicative factor of the maximum joints
     velocities, so 0 means no movement at all, and 1 maximum speed */
  if (*speedLimit < 0.0 || *speedLimit > 1.0) {
    *report = S_lwr_OutOfRange;
    return ERROR;
  }

  /* Setup v, a, j limits */
  SDI_F->speedLimit = *speedLimit;
  setupMaxVAJ(SDI_F->speedLimit, currentMotion.limits_q, currentMotion.limits_x);

  return OK;
}


/*------------------------------------------------------------------------
 * lwrConfigToolCntrl  -  control codel of CONTROL request ConfigTool
 *
 * Description:    
 * Report: OK
 *
 * Returns:    OK or ERROR
 */

STATUS
lwrConfigToolCntrl(LWR_TOOL_CONFIG_STR *toolCfg, int *report)
{
//  /* Setup transformation matrix from reference system 7 to R tool */
//  SDI_F->toolCfg = *toolCfg;
//  setupTh7Tool(SDI_F->toolCfg.tool_type, SDI_F->toolCfg.distance);
//
//  /* Notify the tool change to the LWR */
//  friInst->setToKRLInt(CMD_CHANNEL, SET_TOOL);
//  friInst->setToKRLInt(ARG_CHANNEL, SDI_F->toolCfg.tool_type);
  
  return OK;
}


/*------------------------------------------------------------------------
 * lwrSetSchemeCntrl  -  control codel of CONTROL request SetScheme
 *
 * Description:    
 * Report: OK
 *
 * Returns:    OK or ERROR
 */

STATUS
    lwrSetSchemeCntrl(LWR_CTRL_SCHEME_ENUM *fri_scheme, int *report)
{
 #ifdef LWR_FRI
  switchCtrlScheme(friInstLeft, *fri_scheme);
#endif
  return OK;
}







/*------------------------------------------------------------------------
 * lwrGetQCntrl  -  control codel of CONTROL request GetQ
 *
 * Description:    
 * Report: OK
 *
 * Returns:    OK or ERROR
 */

STATUS
lwrGetQCntrl(int *report)
{
  SDI_F->currentQ.confLeft = SDI_F->armLeft.currConf;
  SDI_F->currentQ.confRight = SDI_F->armRight.currConf;
  SDI_F->currentQ.wait_done = SDI_F->gotoConf.wait_done;
  return OK;
}


/*------------------------------------------------------------------------
 * lwrSetTimeScaleCntrl  -  control codel of CONTROL request SetTimeScale
 *
 * Description:    
 * Report: OK
 *
 * Returns:    OK or ERROR
 */

STATUS
lwrSetTimeScaleCntrl(double *timeScale, int *report)
{
  if(*timeScale < - SDI_F->timeScaleMax) {
    SDI_F->timeScale = - SDI_F->timeScaleMax;
    printf("WARN: lwrSetTimeScaleCntrl wrong timeScale < SDI_F->timeScaleMax:: setted to %g \n", SDI_F->timeScaleMax);
  }
  if(*timeScale > SDI_F->timeScaleMax) {
    SDI_F->timeScale = SDI_F->timeScaleMax;    
    printf("WARN: lwrSetTimeScaleCntrl wrong timeScale >  SDI_F->timeScaleMax: setted to %g\n",  SDI_F->timeScaleMax);
  }


  return OK;
}

/*------------------------------------------------------------------------
 * lwrSetTimeScaleMaxCntrl  -  control codel of CONTROL request SetTimeScale
 *
 * Description:    
 * Report: OK
 *
 * Returns:    OK or ERROR
 */

STATUS
lwrSetTimeScaleMaxCntrl(double *timeScaleMax, int *report)
{
  if(*timeScaleMax < 0) {
    SDI_F->timeScaleMax = 0.0;
    printf("WARN: lwrSetTimeScaleCntrl wrong timeScaleMax < 0.0: setted to 0.0\n");
  }
  if(*timeScaleMax > 1.0) {
    SDI_F->timeScaleMax = 1.0;
    printf("WARN: lwrSetTimeScaleCntrl wrong timeScale > 1.0: setted to 1.0\n");
  }
  return OK;
}



