/*
 * Copyright (c) 2010 LAAS/CNRS                       --  Oct 2010
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
/*------------------------------------------------------------
 * 
 *                     --  Module  LWR  --
 * 
 *  Description:
 *  Creation date : Wed May 19 22:35:29 2010 
 *  Author: Xavier Broquere
 *  LAAS-CNRS, Toulouse FRANCE
 *------------------------------------------------------------*/

#include <math.h>
#include <stdio.h>

#ifdef LWR_FRI
#include "kukaFri-libs/friremote.h"
#endif

#include "server/lwrHeader.h"

extern "C" {
#include "gbStruct.h"
#include "Proto_gb.h"
#include "Proto_gbModeles.h"
}


#include <sys/time.h>
#include "libeuler/pomEuler.h"

#include "lwrStruct.h"
#include "lwrFunctions.h"

extern LWR_MOTION_STR currentMotion;

double deg2rad(double deg)  { return (deg*M_PI)/180.0; }
double rad2deg(double rad)  { return (rad/M_PI)*180.0; }

void fri2th_cartPos(float cartPos[LWR_CART_FRM_DIM], Gb_th* th)
{
  th->vx.x = cartPos[ 0];
  th->vy.x = cartPos[ 1];
  th->vz.x = cartPos[ 2];
  th->vp.x = cartPos[ 3];

  th->vx.y = cartPos[ 4];
  th->vy.y = cartPos[ 5];
  th->vz.y = cartPos[ 6];
  th->vp.y = cartPos[ 7];

  th->vx.z = cartPos[ 8];
  th->vy.z = cartPos[ 9];
  th->vz.z = cartPos[10];
  th->vp.z = cartPos[11];
}


Gb_th fri2dep_cartPos(float cartPos[LWR_CART_FRM_DIM], Gb_dep* dep)
{
  Gb_th th;

  fri2th_cartPos(cartPos, &th);

  normalize( dep->rx,  dep->ry,  dep->rz,
             &dep->rx, &dep->ry, &dep->rz);

  Gb_th_dep(&th, dep);

  return th;
}

void th2fri_cartPos(Gb_th* th, float cartPos[LWR_CART_FRM_DIM])
{
  cartPos[ 0] = th->vx.x;
  cartPos[ 1] = th->vy.x;
  cartPos[ 2] = th->vz.x;
  cartPos[ 3] = th->vp.x;

  cartPos[ 4] = th->vx.y;
  cartPos[ 5] = th->vy.y;
  cartPos[ 6] = th->vz.y;
  cartPos[ 7] = th->vp.y;

  cartPos[ 8] = th->vx.z;
  cartPos[ 9] = th->vy.z;
  cartPos[10] = th->vz.z;
  cartPos[11] = th->vp.z;
}

Gb_th dep2fri_cartPos(Gb_dep* dep, float cartPos[LWR_CART_FRM_DIM])
{
  Gb_th th;
  Gb_dep_th(dep, &th);

  th2fri_cartPos(&th, cartPos);

  return th;
}

void lwr_get_th_R0_R7_from_Gb_th_Base_R8(const Gb_th *th_Rbase_R8, Gb_th *th_R0_R7)
{
  Gb_th th_R7_R8, th_R8_R7, th_R0_R8;
  th_R7_R8.vx.x =  1;
  th_R7_R8.vy.x =  0;
  th_R7_R8.vz.x =  0;
  th_R7_R8.vp.x =  0;
  th_R7_R8.vx.y =  0;
  th_R7_R8.vy.y =  1;
  th_R7_R8.vz.y =  0;
  th_R7_R8.vp.y =  0;
  th_R7_R8.vx.z =  0;
  th_R7_R8.vy.z =  0;
  th_R7_R8.vz.z =  1;
  th_R7_R8.vp.z = LWR_R7;

  Gb_th_inverse(&th_R7_R8, &th_R8_R7);
  th_R0_R8 = *th_Rbase_R8;
  th_R0_R8.vp.z -= LWR_R1;
  Gb_th_produit(&th_R0_R8, &th_R8_R7, th_R0_R7);
  return;
}


/*
 * Setup transformation matrix from reference system 7 to R tool. The second
 * parameter alows projecting the origin of R tool outside the physical tool
 * providing pivoting movements around a target object.
 */
#ifdef LWR_FRI
void setupTh7Tool(friRemote * friInst, LWR_TOOL_ENUM tool, double distance,  Gb_th *th_7_tool,  Gb_th *th_8_tool)
#else
void setupTh7Tool( LWR_TOOL_ENUM tool, double distance,  Gb_th *th_7_tool,  Gb_th *th_8_tool)
#endif
{
  Gb_th th_7_8;
  Gb_th th_R8_Rt1, th_Rt1_Rt;
  Gb_dep dep_Rt1_Rt;

  th_7_8.vx.x =  1;
  th_7_8.vy.x =  0;
  th_7_8.vz.x =  0;
  th_7_8.vp.x =  0;

  th_7_8.vx.y =  0;
  th_7_8.vy.y =  1;
  th_7_8.vz.y =  0;
  th_7_8.vp.y =  0;

  th_7_8.vx.z =  0;
  th_7_8.vy.z =  0;
  th_7_8.vz.z =  1;
  th_7_8.vp.z = LWR_R7;

  switch (tool) {
  case LWR_NO_TOOL:
    th_8_tool->vx.x =  1;
    th_8_tool->vy.x =  0;
    th_8_tool->vz.x =  0;
    th_8_tool->vp.x =  0;

    th_8_tool->vx.y =  0;
    th_8_tool->vy.y =  1;
    th_8_tool->vz.y =  0;
    th_8_tool->vp.y =  0;

    th_8_tool->vx.z =  0;
    th_8_tool->vy.z =  0;
    th_8_tool->vz.z =  1;
    //th_8_tool->vp.z = (distance < 0.0)?LWR_R8_NO_TOOL:distance;
    th_8_tool->vp.z = 0.0;

    printf("INFO: LWR configured with no tool\n");

    break;
  case LWR_GRIPPER:

    th_R8_Rt1.vx.x =  0;
    th_R8_Rt1.vy.x =  0;
    th_R8_Rt1.vz.x =  -1;
    th_R8_Rt1.vp.x =  0;
    th_R8_Rt1.vx.y =  0;
    th_R8_Rt1.vy.y =  1;
    th_R8_Rt1.vz.y =  0;
    th_R8_Rt1.vp.y =  0;
    th_R8_Rt1.vx.z =  1;
    th_R8_Rt1.vy.z =  0;
    th_R8_Rt1.vz.z =  0;
    th_R8_Rt1.vp.z = (distance < 0.0)?LWR_R8_GRIPPER:distance;
    
    Gb_dep_set(&dep_Rt1_Rt, 0, 0, 0, 1, 0, 0, -0.3927);
    Gb_dep_th(&dep_Rt1_Rt, &th_Rt1_Rt);

    Gb_th_produit(&th_R8_Rt1, &th_Rt1_Rt, th_8_tool);
    //th_8_tool->vx.x =  0;
    //th_8_tool->vy.x =  -0.38268342;
    //th_8_tool->vz.x =  0.9238795;
    //th_8_tool->vp.x =  0;
    //
    //th_8_tool->vx.y =  0;
    //th_8_tool->vy.y =  -0.9238795;
    //th_8_tool->vz.y =  -0.38268342;
    //th_8_tool->vp.y =  0;
    //
    //th_8_tool->vx.z =  1;
    //th_8_tool->vy.z =  0;
    //th_8_tool->vz.z =  0;
    //th_8_tool->vp.z = (distance < 0.0)?LWR_R8_GRIPPER:distance;
    
    printf("INFO: LWR configured to use the gripper\n");

    break;
  case LWR_DLR_HIT:
    th_8_tool->vx.x =  1;
    th_8_tool->vy.x =  0;
    th_8_tool->vz.x =  0;
    th_8_tool->vp.x =  0;

    th_8_tool->vx.y =  0;
    th_8_tool->vy.y =  1;
    th_8_tool->vz.y =  0;
    th_8_tool->vp.y =  0;

    th_8_tool->vx.z =  0;
    th_8_tool->vy.z =  0;
    th_8_tool->vz.z =  1;
    //th_8_tool->vp.z = (distance < 0.0)?LWR_R8_DLR_HIT:distance;
    th_8_tool->vp.z = 0.0;
    printf("INFO: LWR configured to use the DLR hand\n");

    break;
  default:
    printf("ERROR: Unrecogniced LWR tool: %d. No tool will be used instead\n", tool);
#ifdef LWR_FRI
    setupTh7Tool(friInst, LWR_NO_TOOL);
#else
    setupTh7Tool( LWR_NO_TOOL);
#endif
    return;
  }

  Gb_th_produit(&th_7_8, th_8_tool, th_7_tool);
#ifdef LWR_FRI
  friInst->setToKRLInt(CMD_CHANNEL, SET_TOOL);
  friInst->setToKRLInt(ARG_CHANNEL, tool);
#endif
  return;
}


/*
 * Setup transformation matrix from reference system 7 to R tool. The second
 * parameter alows projecting the origin of R tool outside the physical tool
 * providing pivoting movements around a target object.
 */
#ifdef LWR_FRI
void setupThBase0(friRemote * friInst, LWR_BASE_ENUM base, Gb_th *th_base_0)
#else
void setupThBase0( LWR_BASE_ENUM base, Gb_th *th_base_0)
#endif
{
  switch (base) {
  case LWR_NULL_BASE:
    th_base_0->vx.x =  1;
    th_base_0->vy.x =  0;
    th_base_0->vz.x =  0;
    th_base_0->vp.x =  0;

    th_base_0->vx.y =  0;
    th_base_0->vy.y =  1;
    th_base_0->vz.y =  0;
    th_base_0->vp.y =  0;

    th_base_0->vx.z =  0;
    th_base_0->vy.z =  0;
    th_base_0->vz.z =  1;
    th_base_0->vp.z =  0;

    printf("INFO: LWR configured to operate mounted on a table\n");

    break;
  case LWR_JIDO_BASE:
    th_base_0->vx.x =  0;
    th_base_0->vy.x =  1;
    th_base_0->vz.x =  0;
    th_base_0->vp.x =  0.165;

    th_base_0->vx.y =  0;
    th_base_0->vy.y =  0;
    th_base_0->vz.y =  -1;
    th_base_0->vp.y =  -0.1105;//-0.165;
    //th_base_0->vp.y =  0.2;//-0.165;

    th_base_0->vx.z = -1;
    th_base_0->vy.z =  0;
    th_base_0->vz.z =  0;
    th_base_0->vp.z =  0.907;

    printf("INFO: LWR configured to operate mounted on JIDO robot\n");

    break;
  default:
    printf("ERROR: Unrecogniced LWR base: %d. Null base will be used instead\n", base);
#ifdef LWR_FRI
    setupThBase0(friInst, LWR_NULL_BASE);
#else
    setupThBase0(LWR_NULL_BASE);
#endif
    return;
  }
#ifdef LWR_FRI
  friInst->setToKRLInt(CMD_CHANNEL, SET_BASE);
  friInst->setToKRLInt(ARG_CHANNEL, base);
#endif 
  return;
}

/*
 * Setup velocity, acceleration and jerk limits for both per-joint and
 * cartesian movements. fullSpeedPercent Varies from 0 (no movement) and
 * 1 (movement at the maximum allowed velocity, acceleration and jerk)
 */
void setupMaxVAJ(double fullSpeedPercent, std::vector<SM_LIMITS> & limits_q,  
                 std::vector<SM_LIMITS> & limits_x)
{

  if(fullSpeedPercent > 1.0) {
    fullSpeedPercent = 1.0;
  }
  if(fullSpeedPercent < 0) {
    fullSpeedPercent = 0;
  }
  /* Per joint maximum v, a, j */
  limits_q.clear();
  limits_q.resize(2*LWR_JOINTS);
  
  limits_q[0].maxVel = LWR_A1_MAX_VELOCITY*fullSpeedPercent;
  limits_q[1].maxVel = LWR_A2_MAX_VELOCITY*fullSpeedPercent;
  limits_q[2].maxVel = LWR_E1_MAX_VELOCITY*fullSpeedPercent;
  limits_q[3].maxVel = LWR_A3_MAX_VELOCITY*fullSpeedPercent;
  limits_q[4].maxVel = LWR_A4_MAX_VELOCITY*fullSpeedPercent;
  limits_q[5].maxVel = LWR_A5_MAX_VELOCITY*fullSpeedPercent;
  limits_q[6].maxVel = LWR_A6_MAX_VELOCITY*fullSpeedPercent;

  limits_q[7].maxVel = LWR_A1_MAX_VELOCITY*fullSpeedPercent;
  limits_q[8].maxVel = LWR_A2_MAX_VELOCITY*fullSpeedPercent;
  limits_q[9].maxVel = LWR_E1_MAX_VELOCITY*fullSpeedPercent;
  limits_q[10].maxVel = LWR_A3_MAX_VELOCITY*fullSpeedPercent;
  limits_q[11].maxVel = LWR_A4_MAX_VELOCITY*fullSpeedPercent;
  limits_q[12].maxVel = LWR_A5_MAX_VELOCITY*fullSpeedPercent;
  limits_q[13].maxVel = LWR_A6_MAX_VELOCITY*fullSpeedPercent;

  for (int i = 0; i < 2*LWR_JOINTS; i++) {
    limits_q[i].maxAcc = limits_q[i].maxVel*2.0;
    limits_q[i].maxJerk = limits_q[i].maxAcc*3.0;
  }

  /* Cartesian maximum v, a, j */

  limits_x.clear();
  limits_x.resize(12);

  limits_x[0].maxVel  = LWR_MAX_LIN_VELOCITY*fullSpeedPercent;  // position x
  limits_x[1].maxVel  = LWR_MAX_LIN_VELOCITY*fullSpeedPercent;  // position y
  limits_x[2].maxVel  = LWR_MAX_LIN_VELOCITY*fullSpeedPercent;  // position z
  limits_x[3].maxVel  = LWR_MAX_ANG_VELOCITY*fullSpeedPercent;  // rotation x
  limits_x[4].maxVel  = LWR_MAX_ANG_VELOCITY*fullSpeedPercent;  // rotation y
  limits_x[5].maxVel  = LWR_MAX_ANG_VELOCITY*fullSpeedPercent;  // rotation z

  limits_x[6].maxVel  = LWR_MAX_LIN_VELOCITY*fullSpeedPercent;  // position x
  limits_x[7].maxVel  = LWR_MAX_LIN_VELOCITY*fullSpeedPercent;  // position y
  limits_x[8].maxVel  = LWR_MAX_LIN_VELOCITY*fullSpeedPercent;  // position z
  limits_x[9].maxVel  = LWR_MAX_ANG_VELOCITY*fullSpeedPercent;  // rotation x
  limits_x[10].maxVel = LWR_MAX_ANG_VELOCITY*fullSpeedPercent;  // rotation y
  limits_x[11].maxVel = LWR_MAX_ANG_VELOCITY*fullSpeedPercent;  // rotation z

  for (int i = 0; i < 12; i++) {
    limits_x[i].maxAcc  = limits_x[i].maxVel*2.0;
    limits_x[i].maxJerk = limits_x[i].maxAcc*3.0;
  }
  return;
}


bool approachingBounds(int joint, double old_q, double new_q)
{
  if ((fabs(LWR_JOINT_RANGE_MIN[joint] - new_q) < LWR_JNT_SECURITY_MARGIN) &&
      (new_q < (old_q - 0.0001)))
    return true;

  if ((fabs(LWR_JOINT_RANGE_MAX[joint] - new_q) < LWR_JNT_SECURITY_MARGIN) &&
      (new_q > (old_q + 0.0001)))
    return true;

  return false;
}

bool approachingBoundsForArm(Gb_q7 q_old, Gb_q7 q)
{
  if(approachingBounds(0, q_old.q1, q.q1) == true) {
    return true;
  }
  if(approachingBounds(1 , q_old.q1, q.q2) == true) {
    return true;
  }
  if(approachingBounds(2 , q_old.q1, q.q3) == true) {
    return true;
  }
  if(approachingBounds(3, q_old.q1, q.q4) == true) {
    return true;
  }
  if(approachingBounds(4 , q_old.q1, q.q5) == true) {
    return true;
  }
  if(approachingBounds(5 , q_old.q1, q.q6) == true) {
    return true;
  }
  if(approachingBounds(6, q_old.q1, q.q7) == true) {
    return true;
  }
  return false;
}

double normalize(double   x, double   y, double   z,
                 double* nx, double* ny, double* nz)
{
  Gb_v3 v, nv;
  v.x = x;
  v.y = y;
  v.z = z;
  double w = Gb_v3_norme(&v, &nv);
  *nx = nv.x;
  *ny = nv.y;
  *nz = nv.z;
  return w;
}

void setMotionError(LWR_MOTION_ERROR_ENUM error)
{
  currentMotion.type  = LWR_NOT_MOVE;
  currentMotion.state = LWR_ABORTED;
  currentMotion.error = error;
}

void setMotionStart(LWR_MOTION_TYPE_ENUM type)
{
  currentMotion.type  = type;
  currentMotion.state  = LWR_STARTED;
  currentMotion.error  = LWR_OK;
  currentMotion.curTime = 0.0;
}

void setMotionEnd()
{
  currentMotion.type  = LWR_NOT_MOVE;
  currentMotion.state = LWR_FINISHED;
  currentMotion.error = LWR_OK;
}

void fillMotionReport(LWR_MOTION_ERROR_ENUM error, int *report)
{
  switch (error) {
  case LWR_OK:               *report = OK;                      break;
  case LWR_NOT_INITIALIZED:  *report = S_lwr_NOT_INITIALIZED;   break;
  case LWR_WRONG_Q_VALUE:    *report = S_lwr_WRONG_VALUE;       break;
  case LWR_SOFTMOTION_ERROR: *report = S_lwr_SOFTMOTION_ERROR;  break;
  case LWR_FRI_MONITOR_MODE: *report = S_lwr_FRI_MONITOR_MODE;  break;
  case LWR_MGI_CALC_ERROR:   *report = S_lwr_MGI_CALC_ERROR;    break;
  case LWR_MGI_SINGULARITY:  *report = S_lwr_MGI_SINGULARITY;   break;
  case LWR_JNT_OUT_OF_RANGE: *report = S_lwr_JNT_OUT_OF_RANGE;  break;
  case LWR_COLLISION_RISK:   *report = S_lwr_COLLISION_RISK;    break;
  }
}

// daniel: reste problme calcul temps pour acclration
#ifdef LWR_FRI
int lwrUpdateArmStatus(friRemote *friInst, LWR_ARM_INST* arm, POM_SE_POSTER *pomArmPos, char * pomPosterName)
#else
int lwrUpdateArmStatus( LWR_ARM_INST* arm, POM_SE_POSTER *pomArmPos, char * pomPosterName)
#endif
{
  //struct timeval s_date;
  // struct timeval s_date0 = { arm->s_date0_tv_sec, arm->s_date0_tv_usec };
  double date; 
  double time;

  //float currentCartPos[LWR_CART_FRM_DIM];
  float currentJntVals[LWR_JOINTS];
  T3D t3d;
  Gb_dep currDep;
  Gb_th currTh;

#ifdef LWR_FRI
  if( friInst != NULL) {
    memcpy(arm->currentCartPos, friInst->getMsrCartPosition(), sizeof(float)*LWR_CART_FRM_DIM);
    memcpy(arm->currentMsrJntVals, friInst->getMsrMsrJntPosition(), sizeof(float)*LWR_JOINTS);
    //    memcpy(currentCartPos, friInst->getMsrCmdCartPosition(), sizeof(float)*FRI_CART_FRM_DIM);
    friInst->getCurrentCmdJntPosition(currentJntVals);
    //memcpy(currentJntVals, friInst->getMsrMsrJntPosition(), sizeof(float)*LWR_JOINTS);

  } else {
    memcpy(currentJntVals , arm->currentJntVals, sizeof(float)*LWR_JOINTS);
  }
  // Fill posters with current position and configuration
  fri2dep_cartPos(arm->currentCartPos, &arm->currPose);
#else
  memcpy(currentJntVals , arm->currentJntVals, sizeof(float)*LWR_JOINTS);

#endif
  arm->currConf.q1 = currentJntVals[0];
  arm->currConf.q2 = currentJntVals[1];
  arm->currConf.q3 = currentJntVals[2];
  arm->currConf.q4 = currentJntVals[3];
  arm->currConf.q5 = currentJntVals[4];
  arm->currConf.q6 = currentJntVals[5];
  arm->currConf.q7 = currentJntVals[6];

#ifndef   LWR_FRI

  //similate the pose of the flange returned by the FRI
  Gb_th th07, th08;
  kukaLBR_mgd(&(arm->currConf), LWR_R3, LWR_R5, &th07);
  //  Gb_th_print(&th07, "thR7");
  Gb_th th_7_8;

  th_7_8.vx.x =  1;
  th_7_8.vy.x =  0;
  th_7_8.vz.x =  0;
  th_7_8.vp.x =  0;

  th_7_8.vx.y =  0;
  th_7_8.vy.y =  1;
  th_7_8.vz.y =  0;
  th_7_8.vp.y =  0;

  th_7_8.vx.z =  0;
  th_7_8.vy.z =  0;
  th_7_8.vz.z =  1;
  th_7_8.vp.z = LWR_R7;

  Gb_th_produit(&th07, &th_7_8, &th08); 
  th08.vp.z += LWR_R1;
  th2fri_cartPos(&th08, arm->currentCartPos);

  //std::cout << "simulated MGD th07 "<< std::endl;
  //Gb_th_print(&th07, "th07");

  //arm->currentCartPos[11] += LWR_R1;
#endif

  arm->pos.q.q1 = currentJntVals[0];
  arm->pos.q.q2 = currentJntVals[1];
  arm->pos.q.q3 = currentJntVals[2];
  arm->pos.q.q4 = currentJntVals[3];
  arm->pos.q.q5 = currentJntVals[4];
  arm->pos.q.q6 = currentJntVals[5];
  arm->pos.q.q7 = currentJntVals[6];

#ifdef LWR_FRI
  if( friInst != NULL) {
    date = friInst->getTimestamp();
  } else {
    date  = arm->timeStamp;
  }
#endif
  arm->pos.date = date;

  if (arm->date1 == -1.) {
    arm->pos.vq.q1 = 0.;
    arm->pos.vq.q2 = 0.;
    arm->pos.vq.q3 = 0.;
    arm->pos.vq.q4 = 0.;
    arm->pos.vq.q5 = 0.;
    arm->pos.vq.q6 = 0.;
    arm->pos.vq.q7 = 0.;

    arm->pos.vqold.q1 = 0.;
    arm->pos.vqold.q2 = 0.;
    arm->pos.vqold.q3 = 0.;
    arm->pos.vqold.q4 = 0.;
    arm->pos.vqold.q5 = 0.;
    arm->pos.vqold.q6 = 0.;
    arm->pos.vqold.q7 = 0.;

  }
  else {

#ifndef LWR_FRI
    time = (double)LWR_PERIOD;
#else
    if(date > 0.0 ) {
      time = (double)date - (double)arm->date1;
    } else {
      time = (double)LWR_PERIOD;
    }
    time = (double)LWR_PERIOD;
#endif

    arm->pos.vq.q1 =  (arm->pos.q.q1 - arm->pos.qold.q1) / (time);
    arm->pos.vq.q2 =  (arm->pos.q.q2 - arm->pos.qold.q2) / (time);
    arm->pos.vq.q3 =  (arm->pos.q.q3 - arm->pos.qold.q3) / (time);
    arm->pos.vq.q4 =  (arm->pos.q.q4 - arm->pos.qold.q4) / (time);
    arm->pos.vq.q5 =  (arm->pos.q.q5 - arm->pos.qold.q5) / (time);
    arm->pos.vq.q6 =  (arm->pos.q.q6 - arm->pos.qold.q6) / (time);
    arm->pos.vq.q7 =  (arm->pos.q.q7 - arm->pos.qold.q7) / (time);

  }
  if (arm->date2 == -1.) {
    arm->pos.aq.q1 = 0.;
    arm->pos.aq.q2 = 0.;
    arm->pos.aq.q3 = 0.;
    arm->pos.aq.q4 = 0.;
    arm->pos.aq.q5 = 0.;
    arm->pos.aq.q6 = 0.;
    arm->pos.aq.q7 = 0.;

    arm->pos.aqold.q1 = 0.;
    arm->pos.aqold.q2 = 0.;
    arm->pos.aqold.q3 = 0.;
    arm->pos.aqold.q4 = 0.;
    arm->pos.aqold.q5 = 0.;
    arm->pos.aqold.q6 = 0.;
    arm->pos.aqold.q7 = 0.;
  }
  else {
#ifndef LWR_FRI
    time = (double)LWR_PERIOD;
#else
    if(date > 0.0 ) {
      //time =  (date - arm->date2) * 2.0 ;
      time =  ((double)date - (double)arm->date2);
    } else {
      time = (double)LWR_PERIOD;
    }
    time = (double)LWR_PERIOD;
#endif

    arm->pos.aq.q1 = (arm->pos.vq.q1 - arm->pos.vqold.q1) / time;
    arm->pos.aq.q2 = (arm->pos.vq.q2 - arm->pos.vqold.q2) / time;
    arm->pos.aq.q3 = (arm->pos.vq.q3 - arm->pos.vqold.q3) / time;
    arm->pos.aq.q4 = (arm->pos.vq.q4 - arm->pos.vqold.q4) / time;
    arm->pos.aq.q5 = (arm->pos.vq.q5 - arm->pos.vqold.q5) / time;
    arm->pos.aq.q6 = (arm->pos.vq.q6 - arm->pos.vqold.q6) / time;
    arm->pos.aq.q7 = (arm->pos.vq.q7 - arm->pos.vqold.q7) / time;

  }



  arm->date2 = arm->date1;
  arm->date1 = date;
  /* Save previous configuration (q and vq) */ 
  arm->pos.qold  = arm->pos.q;
  arm->pos.vqold = arm->pos.vq;
  arm->pos.aqold = arm->pos.aq;

  // POM interface
  strcpy(pomArmPos->seName, pomPosterName);
  fri2th_cartPos(arm->currentCartPos, &currTh);
  t3dSetOk(&t3d);
  t3d.matrix.flags = T3D_ALLOW_CONVERSION;
  t3d.type = T3D_MATRIX;
  t3d.matrix.type = T3D_MATRIX;
  t3d.matrix.matrix[0]  = currTh.vx.x;
  t3d.matrix.matrix[1]  = currTh.vy.x;
  t3d.matrix.matrix[2]  = currTh.vz.x;
  t3d.matrix.matrix[3]  = currTh.vp.x;
  t3d.matrix.matrix[4]  = currTh.vx.y;
  t3d.matrix.matrix[5]  = currTh.vy.y;
  t3d.matrix.matrix[6]  = currTh.vz.y;
  t3d.matrix.matrix[7]  = currTh.vp.y;
  t3d.matrix.matrix[8]  = currTh.vx.z;
  t3d.matrix.matrix[9]  = currTh.vy.z;
  t3d.matrix.matrix[10] = currTh.vz.z;
  t3d.matrix.matrix[11] = currTh.vp.z;
  t3d.matrix.matrix[12] = 0;
  t3d.matrix.matrix[13] = 0;
  t3d.matrix.matrix[14] = 0;
  t3d.matrix.matrix[15] = 1;
  pomEulerVFromT3D(&t3d, &pomArmPos->seConfig);

  return 0;
}

bool lwrIsArmStopped(LWR_ARM_INST* armInst) 
{
  double vel_threshold = 0.001;


  if(fabs(armInst->pos.vq.q1) > vel_threshold) {
      return false;
  }
  if(fabs(armInst->pos.vq.q2) > vel_threshold) {
    return false;
  }
  if(fabs(armInst->pos.vq.q3) > vel_threshold) {
    return false;
  }
  if(fabs(armInst->pos.vq.q4) > vel_threshold) {
    return false;
  }
  if(fabs(armInst->pos.vq.q5) > vel_threshold) {
    return false;
  }
  if(fabs(armInst->pos.vq.q6) > vel_threshold) {
    return false;
  }
  if(fabs(armInst->pos.vq.q7) > vel_threshold) {
    return false;
  }
  
  return true;
}

#ifdef LWR_FRI
int lwrCommandArm(friRemote *friInst, GEN_BOOL simulation, GEN_BOOL isTrqCntrlSofMotion, LWR_ARM_ID_ENUM arm_id, LWR_ARM_INST* armInst, GENMANIP_MOTION_STR *curState, LWR_MOTION_STR *motion) 
#else
int lwrCommandArm(  GEN_BOOL isTrqCntrlSofMotion, LWR_ARM_ID_ENUM arm_id, LWR_ARM_INST* armInst, GENMANIP_MOTION_STR *curState, LWR_MOTION_STR *motion)
#endif
{
  float commandJntVals[LWR_JOINTS];
  float currentJntVals[LWR_JOINTS];
  float commandCartPos[LWR_CART_FRM_DIM];
  float currentCartPos[LWR_CART_FRM_DIM];
  Gb_dep dep;
  Gb_th th_R8, th_W8, th_Wt;
  int OFFSET_ARM = 0;

  std::vector<SM_COND> cond;

  if ((motion->type == LWR_POSITION) && (motion->state != LWR_ABORTED)) {
    // LWR_POSITION == control in articular space
    if(arm_id == LWR_ARM_RIGHT) {
      OFFSET_ARM =  LWR_JOINTS;
    }
    /***********************************************/
    /*         Compute The Control Law             */
    /*  output: commandJntVals[]                   */
    /***********************************************/
    switch (armInst->cntrl_mode) {
    case LWR_OPEN_LOOP_CNTRL :
      lwrOpenLoopCntrl(armInst, curState, motion, commandJntVals);
      break;
    case LWR_TRAJECTORY_CNTRL :
      lwrTrajectoryCntrl(OFFSET_ARM, armInst, curState, motion, commandJntVals);
      break;
    case LWR_GOAL_CNTRL :
      lwrGoalCntrl(OFFSET_ARM, curState, motion, commandJntVals);
      break;
    case LWR_EXTERNAL_POSITION_CNTRL :
      break;
    case LWR_EXTERNAL_VELOCITY_CNTRL :
      break;
    case LWR_FORCE_CNTRL :
      lwrJntTrqCntrl(isTrqCntrlSofMotion, OFFSET_ARM, armInst, curState, motion, commandJntVals);
      break;
    case LWR_XOPEN_LOOP_CNTRL :
      lwrXOpenLoopCntrl(armInst, curState, motion, commandJntVals);
      break;
    case LWR_XCLOSED_LOOP_CNTRL:
      lwrXClosedLoopCntrl(armInst, curState, motion, commandJntVals);
      break;
    default:
      printf("ERROR lwrCommandArm: WRONG CONTROL MODE\n");
      break;
    }
    /***********************************************/

#ifdef LWR_FRI
    if((friInst != NULL) && (simulation != GEN_TRUE)) {
      friInst->doPositionControl(commandJntVals, false); // ligne qui fait bouger le robot // apres le dodataexchange
      memcpy(armInst->currentJntVals, commandJntVals, sizeof(float)*LWR_JOINTS);
    } else {
      memcpy(armInst->currentJntVals, commandJntVals, sizeof(float)*LWR_JOINTS);
    }
#else
    memcpy(armInst->currentJntVals, commandJntVals, sizeof(float)*LWR_JOINTS);
#endif
    motion->state = LWR_MOVING;

  } /*if ((currentMotion.type[arm_id] == LWR_POSITION) && (currentMotion.state != LWR_ABORTED)) */

  else if  ((motion->type == LWR_CARTESIAN) && (motion->state != LWR_ABORTED)){

    //if(arm_id == LWR_ARM_RIGHT) {
    //  OFFSET_ARM =  LWR_CART_VEC;
    //}
    //
    //// Cartesian command: we obtain Q from the MGI
    //dep.x = motion->x[0 + OFFSET_ARM][motion->sample];
    //dep.y = motion->x[1 + OFFSET_ARM][motion->sample];
    //dep.z = motion->x[2 + OFFSET_ARM][motion->sample];
    //dep.a = normalize(motion->x[3 + OFFSET_ARM][motion->sample],
    //    	      motion->x[4 + OFFSET_ARM][motion->sample],
    //    	      motion->x[5 + OFFSET_ARM][motion->sample],
    //    	      &dep.rx, &dep.ry, &dep.rz);
    //
    //Gb_dep_th(&dep, &th_R8);
    //Gb_th_produit(&armInst->th_base_0, &th_R8, &th_W8);
    //Gb_th_produit(&th_W8, &armInst->th_8_tool, &th_Wt);
    //
    //th2fri_cartPos(&th_R8, commandCartPos);
#ifdef LWR_FRI
    //if((friInst != NULL) && (simulation != GEN_TRUE)) {
    // friInst->doCartesianImpedanceControl(commandCartPos, NULL, NULL, //currentCartStiff, currentCartDamp,
    //					 NULL, NULL, false);
    //} else {
    //TODO COMPUTE THE SUIMUALTED JNT VAL
    //}
#else
    //TODO COMPUTE THE SUIMUALTED JNT VAL
#endif
  }/*if  ((currentMotion.type == LWR_CARTESIAN) && (currentMotion.state != LWR_ABORTED)) */
  return 0;
}



bool lwrIsDepNegligeable(double error, const Gb_dep *dep) {
  if(error <= 0.0) {
    error = 0.001;
    printf("WARN lwrIsDepNegligeable: wrong value set error to 0.001\n");
  }
  if(fabs(dep->x) <   error && 
     fabs(dep->y) <   error &&
     fabs(dep->z) <   error &&
     fabs(dep->a) <   error) {
    return true;
  }
  return false;
}

bool lwrIsDistDepNegligeable(double error, const Gb_dep *dep1, const Gb_dep *dep2) {
  if(error <= 0.0) {
    error = 0.001;
    printf("WARN lwrIsDistDepNegligeable: wrong value set error to 0.001\n");
  }
  // todo check if the rotation vector is also the same when angle is the same
  if(fabs(dep2->x - dep1->x) <   error && 
     fabs(dep2->y - dep1->y) <   error &&
     fabs(dep2->z - dep1->z) <   error &&
     fabs(dep2->a - dep1->a) <   error) {
    return true;
  }
  return false;
}

int lwrOpenLoopCntrl( LWR_ARM_INST* armInst, GENMANIP_MOTION_STR *curState, LWR_MOTION_STR *motion, float *commandJntVals) {
  std::vector<SM_COND> cond;
  motion->traj.getMotionCond(motion->curTime, cond);
  for (int i = 0; i < LWR_JOINTS; i++) {
    commandJntVals[i] = cond[i+armInst->indexInTraj].x;
  }

  if(armInst->isHandUsed == GEN_TRUE && armInst->toolCfg.tool_type == LWR_DLR_HIT ) {
    armInst->hand.thumbRotation  = cond[armInst->indexHandInTraj + 0].x;
    armInst->hand.thumb.base     = cond[armInst->indexHandInTraj + 1].x;
    armInst->hand.thumb.phaProx  = cond[armInst->indexHandInTraj + 2].x;
    armInst->hand.thumb.phaMid   = cond[armInst->indexHandInTraj + 3].x;
    armInst->hand.thumb.phaDist  = cond[armInst->indexHandInTraj + 4].x;
    armInst->hand.fore.base      = cond[armInst->indexHandInTraj + 5].x;
    armInst->hand.fore.phaProx   = cond[armInst->indexHandInTraj + 6].x;
    armInst->hand.fore.phaMid    = cond[armInst->indexHandInTraj + 7].x;
    armInst->hand.fore.phaDist   = cond[armInst->indexHandInTraj + 8].x;
    armInst->hand.middle.base    = cond[armInst->indexHandInTraj + 9].x;
    armInst->hand.middle.phaProx = cond[armInst->indexHandInTraj + 10].x;
    armInst->hand.middle.phaMid  = cond[armInst->indexHandInTraj + 11].x;
    armInst->hand.middle.phaDist = cond[armInst->indexHandInTraj + 12].x;
    armInst->hand.ring.base      = cond[armInst->indexHandInTraj + 13].x;
    armInst->hand.ring.phaProx   = cond[armInst->indexHandInTraj + 14].x;
    armInst->hand.ring.phaMid    = cond[armInst->indexHandInTraj + 15].x;
    armInst->hand.ring.phaDist   = cond[armInst->indexHandInTraj + 16].x;
  }

  return 0;
}


/*
 *  @brief: th_R0_R7 goto in motion is the goal position of R7 in the frame of armBase R0: th_R0_R7
 *   need to calculate this before the call of function from th_Rob_Tool -> th_R0_R7
 *  
 */
int lwrXClosedLoopCntrl(LWR_ARM_INST* armInst, GENMANIP_MOTION_STR *curState, LWR_MOTION_STR* motion, float *commandJntVals) {
  std::vector<SM_COND> ic, fc;
  static std::vector<SM_COND> cond;

  static std::vector<SM_COND> cond_q;
  std::vector<SM_COND> ic_q, fc_q;

  Gb_th th_R0_R7_goal = motion->th_R0_R7_Goto;
  Gb_th th_R7_R8, th_R8_R7, th_R7_cur_R7;
  Gb_dep dep_R0_R7_cur, dep_R7_cur_R7;
  Gb_q7 q;

  static Gb_th th_R0_R7_old = {{1,0,0}, {0, 1, 0}, {0, 0, 1}, {0, 0, 0}};
  static Gb_th th_R0_R7_init;
  Gb_dep dep_R0_R7_old, dep_R0_R7_goal;

  static SM_TRAJ localTraj;
  static double timeOnLocalTraj = LWR_PERIOD;
  //localTraj.clear();
  // create an empty trajectory with 6 axes
  //localTraj.resize(6);

  Gb_dep gotoDep;
  Gb_th_dep(&th_R0_R7_goal, &gotoDep);
  if (! normalize( gotoDep.rx,  gotoDep.ry,  gotoDep.rz,
                   &gotoDep.rx, &gotoDep.ry, &gotoDep.rz)) {
    printf("ERROR: in function lwrXClosedLoopCntrl, wrong (maybe null) vector is provided for tool orientation\n");
    Gb_th_print(&th_R0_R7_goal, "th_R0_R7_goal" );
    Gb_dep_print(&gotoDep);
    return 1;
  }

  //get the current th_R0_R7
  Gb_th th_Base_R8_cur, th_R0_R8_cur,th_R0_R7_cur;
  Gb_quat quat_R0_R7_cur, quat_R0_R7;

  //currentCarPos is cartesian position in the frame of arm base
  fri2th_cartPos(armInst->currentCartPos, &th_Base_R8_cur);
  lwr_get_th_R0_R7_from_Gb_th_Base_R8(&th_Base_R8_cur, &th_R0_R7_cur);
  Gb_th_dep(&th_R0_R7_cur, &dep_R0_R7_cur);

  // check if the goal as moved within a threshold
  Gb_th_dep(&th_R0_R7_old, &dep_R0_R7_old);
  Gb_th_dep(&th_R0_R7_goal, &dep_R0_R7_goal);

  if(lwrIsDistDepNegligeable(0.02,  &dep_R0_R7_old,  &dep_R0_R7_goal) == false) { // || 
    //     lwrIsDistDepNegligeable(0.01,  &dep_R0_R7_cur,  &dep_R0_R7_goal) == false) {
    printf("#### GOAL HAS MOVED, REPLAN THE TRAJECTORY ######\n");
    // we need to recompute the trajectory since the goal as moved
    // update the th_R0_R7_old for the next period
    //printf("********************INFO computing a new trajectory****************\n");
    th_R0_R7_old = th_R0_R7_goal;
    th_R0_R7_init = th_R0_R7_cur;
 
    // compute the relative dep
    Gb_th_quat(&th_R0_R7_goal, &quat_R0_R7);
    Gb_th_quat(&th_R0_R7_cur, &quat_R0_R7_cur);
    Gb_quat_compute_relativeDep_to_interpole(&quat_R0_R7_cur, &quat_R0_R7, &dep_R7_cur_R7);
    //printf("dep_R7_cur_R7 used by compute traj \n");
    //Gb_dep_print(&dep_R7_cur_R7);
    
    //compute the softMotion trajectory in the cartesian frame
    cond.resize(6);
    ic.resize(6);
    fc.resize(6);
    for(unsigned int i=0; i<6; i++) {
      // todo set the right accel and velocity
      ic[i].a = cond[i].a;
      ic[i].v = cond[i].v;
      ic[i].x = 0;
    }
    fc[0].a = 0;
    fc[0].v = 0;
    fc[0].x = dep_R7_cur_R7.x;
    fc[1].a = 0;
    fc[1].v = 0;
    fc[1].x = dep_R7_cur_R7.y;
    fc[2].a = 0;
    fc[2].v = 0;
    fc[2].x = dep_R7_cur_R7.z;
    fc[3].a = 0;
    fc[3].v = 0;
    fc[3].x = dep_R7_cur_R7.rx * dep_R7_cur_R7.a;
    fc[4].a = 0;
    fc[4].v = 0;
    fc[4].x = dep_R7_cur_R7.ry * dep_R7_cur_R7.a;
    fc[5].a = 0;
    fc[5].v = 0;
    fc[5].x = dep_R7_cur_R7.rz * dep_R7_cur_R7.a;
    
    localTraj.clear();
    std::vector<SM_LIMITS> limits;
    limits.resize(6);
    for(unsigned int p=0; p<6; p++) {
      limits[p].maxJerk = currentMotion.limits_x[p+6].maxJerk;
      limits[p].maxAcc = currentMotion.limits_x[p+6].maxAcc;
      limits[p].maxVel = currentMotion.limits_x[p+6].maxVel;
    }
    localTraj.computeTraj(ic, fc, limits, SM_TRAJ::SM_INDEPENDANT);
    timeOnLocalTraj = LWR_PERIOD;
    //localTraj.print();
  }

  //printf(" localTraj.getDuration() %f timeOnLocalTraj %f \n",  localTraj.getDuration(), timeOnLocalTraj);
  if( localTraj.getDuration() > 0.0) {
    if(timeOnLocalTraj >= localTraj.getDuration()) {
      timeOnLocalTraj = localTraj.getDuration();
    }


    localTraj.getMotionCond(timeOnLocalTraj + 50*LWR_PERIOD, cond);
    timeOnLocalTraj += LWR_PERIOD;
    
    if(timeOnLocalTraj >= localTraj.getDuration()) {
      timeOnLocalTraj = localTraj.getDuration();
    }

    // compute the next R7 pose in the R0 frame
    dep_R7_cur_R7.x =  cond[0].x;
    dep_R7_cur_R7.y =  cond[1].x;
    dep_R7_cur_R7.z =  cond[2].x;
    Gb_th th_R0_R7;
    Gb_v3 rel_rot, rot;
    rel_rot.x = cond[3].x;
    rel_rot.y = cond[4].x;
    rel_rot.z = cond[5].x;
    double angle =  Gb_v3_norme ( &rel_rot, &rot);
    dep_R7_cur_R7.rx = rot.x;
    dep_R7_cur_R7.ry = rot.y;
    dep_R7_cur_R7.rz = rot.z;
    dep_R7_cur_R7.a = angle;

    //printf("Gb_dep_delta \n");
    //Gb_dep_print(&dep_R7_cur_R7);
    Gb_dep_th(&dep_R7_cur_R7, &th_R7_cur_R7);
    //th_R7_cur_R7 is the next position in cartesian space, inital condition to be 0 0 0 0 0 0
    //local R7 times this gives the goal pos for the movement
    Gb_th_produit(&th_R0_R7_init , &th_R7_cur_R7, &th_R0_R7);
    
    //MGI, and pass to commandeJntVals; 
    //TODO: write the second lwrComputeMGI with optimized J3 motion
    
    if(lwrComputeMGI(&(armInst->currConf), &th_R0_R7, &q) == 1) {
      printf("ERROR lwrComputeMGI return 1\n");
      return 1;
    }

    // compute the sub trajectory in the configuation space
    if(cond_q.size() <= 0 ) {
      cond_q.resize(7);
    }
    
    ic_q.resize(7);
    fc_q.resize(7);
   ic_q[0].a = cond_q[0].a;
   ic_q[0].v = cond_q[0].v;
   ic_q[0].x = armInst->currConf.q1;
   ic_q[1].a = cond_q[1].a;
   ic_q[1].v = cond_q[1].v;
   ic_q[1].x = armInst->currConf.q2;
   ic_q[2].a = cond_q[2].a;
   ic_q[2].v = cond_q[2].v;
   ic_q[2].x = armInst->currConf.q3;
   ic_q[3].a = cond_q[3].a;
   ic_q[3].v = cond_q[3].v;
   ic_q[3].x = armInst->currConf.q4;
   ic_q[4].a = cond_q[4].a;
   ic_q[4].v = cond_q[4].v;
   ic_q[4].x = armInst->currConf.q5;
   ic_q[5].a = cond_q[5].a;
   ic_q[5].v = cond_q[5].v;
   ic_q[5].x = armInst->currConf.q6;
   ic_q[6].a = cond_q[6].a;
   ic_q[6].v = cond_q[6].v;
   ic_q[6].x = armInst->currConf.q7;

   fc_q[0].a = 0;
   fc_q[0].v = 0;
   fc_q[0].x = q.q1;
   fc_q[1].a = 0;
   fc_q[1].v = 0;
   fc_q[1].x = q.q2;
   fc_q[2].a = 0;
   fc_q[2].v = 0;
   fc_q[2].x = q.q3;
   fc_q[3].a = 0;
   fc_q[3].v = 0;
   fc_q[3].x = q.q4;
   fc_q[4].a = 0;
   fc_q[4].v = 0;
   fc_q[4].x = q.q5;
   fc_q[5].a = 0;
   fc_q[5].v = 0;
   fc_q[5].x = q.q6;
   fc_q[6].a = 0;
   fc_q[6].v = 0;
   fc_q[6].x = q.q7;
   SM_TRAJ localTrajQ;
   localTrajQ.clear();
   std::vector<SM_LIMITS> limits_q;
   limits_q.resize(7);
   for(unsigned int p=0; p<7; p++) {
     limits_q[p].maxJerk = currentMotion.limits_q[p+6].maxJerk;
     limits_q[p].maxAcc = currentMotion.limits_q[p+6].maxAcc;
     limits_q[p].maxVel = currentMotion.limits_q[p+6].maxVel;
   }
   localTrajQ.computeTraj(ic_q, fc_q, limits_q, SM_TRAJ::SM_INDEPENDANT);

   localTrajQ.getMotionCond(LWR_PERIOD, cond_q);
   
   q.q1 =  cond_q[0].x;
   q.q2 =  cond_q[1].x;
   q.q3 =  cond_q[2].x;
   q.q4 =  cond_q[3].x;
   q.q5 =  cond_q[4].x;
   q.q6 =  cond_q[5].x;
   q.q7 =  cond_q[6].x;

  } else {
    
    // the displacement is negligeable
    // printf("Goal reached do not move\n");
    q.q1 =  armInst->currConf.q1;
    q.q2 =  armInst->currConf.q2;
    q.q3 =  armInst->currConf.q3;
    q.q4 =  armInst->currConf.q4;
    q.q5 =  armInst->currConf.q5;
    q.q6 =  armInst->currConf.q6;
    q.q7 =  armInst->currConf.q7;
    
    for(unsigned int i=0; i<cond.size(); i++) {
      cond[i].a = 0.0;
      cond[i].v = 0.0;
      cond[i].x = 0.0;
    }
    
    for(unsigned int i=0; i<cond_q.size(); i++) {
      cond_q[i].a = 0.0;
      cond_q[i].v = 0.0;
      cond_q[i].x = 0.0;
    }
  }

  
  if(approachingBoundsForArm(armInst->currConf, q)) {
    printf("WARN: approachingBoundsForArm %d\n", armInst->isLeft);
    commandJntVals[0] =  armInst->currConf.q1;
    commandJntVals[1] =  armInst->currConf.q2;
    commandJntVals[2] =  armInst->currConf.q3;
    commandJntVals[3] =  armInst->currConf.q4;
    commandJntVals[4] =  armInst->currConf.q5;
    commandJntVals[5] =  armInst->currConf.q6;
    commandJntVals[6] =  armInst->currConf.q7;
  } else {
    Gb_q7 q_acceptable;
    // Check if new q is acceptable according to the robot's kinematic constraints
    lwrCheckAndSmoothArmConfigEvolution(&armInst->currConf, &q, LWR_PERIOD, &q_acceptable);
    commandJntVals[0] =  q_acceptable.q1;
    commandJntVals[1] =  q_acceptable.q2;
    commandJntVals[2] =  q_acceptable.q3;
    commandJntVals[3] =  q_acceptable.q4;
    commandJntVals[4] =  q_acceptable.q5;
    commandJntVals[5] =  q_acceptable.q6;
    commandJntVals[6] =  q_acceptable.q7;
  }
  
  //Gb_q7_print(&q, "goalConf");
  
  return 0;
}

int lwrCheckAndSmoothArmConfigEvolution(const Gb_q7 *q_old, const Gb_q7 *q_goto, double dt, Gb_q7 *q_new)
{
  if(fabs(q_goto->q1 - q_old->q1)/dt > LWR_JOINT_VEL_MAX[0]) {
    q_new->q1 = q_old->q1 + SIGN(q_goto->q1 - q_old->q1)*dt*LWR_JOINT_VEL_MAX[0];
    printf("WARN joint q1 is saturated \n");
  } else {
    q_new->q1 = q_goto->q1;
  }
  if(fabs(q_goto->q2 - q_old->q2)/dt > LWR_JOINT_VEL_MAX[1]) {
    q_new->q2 = q_old->q2 + SIGN(q_goto->q2 - q_old->q2)*dt*LWR_JOINT_VEL_MAX[1];
    printf("WARN joint q2 is saturated \n");
  } else {
    q_new->q2 = q_goto->q2;
  }
  if(fabs(q_goto->q3 - q_old->q3)/dt > LWR_JOINT_VEL_MAX[2]) {
    q_new->q3 = q_old->q3 + SIGN(q_goto->q3 - q_old->q3)*dt*LWR_JOINT_VEL_MAX[2];
    printf("WARN joint q3 is saturated \n");
  } else {
    q_new->q3 = q_goto->q3;
  }
  if(fabs(q_goto->q4 - q_old->q4)/dt > LWR_JOINT_VEL_MAX[3]) {
    q_new->q4 = q_old->q4 + SIGN(q_goto->q4 - q_old->q4)*dt*LWR_JOINT_VEL_MAX[3];
    printf("WARN joint q4 is saturated \n");
  } else {
    q_new->q4 = q_goto->q4;
  }
  if(fabs(q_goto->q5 - q_old->q5)/dt > LWR_JOINT_VEL_MAX[4]) {
    q_new->q5 = q_old->q5 + SIGN(q_goto->q5 - q_old->q5)*dt*LWR_JOINT_VEL_MAX[4];
    printf("WARN joint q5 is saturated \n");
  } else {
    q_new->q5 = q_goto->q5;
  }
  if(fabs(q_goto->q6 - q_old->q6)/dt > LWR_JOINT_VEL_MAX[5]) {
    q_new->q6 = q_old->q6 + SIGN(q_goto->q6 - q_old->q6)*dt*LWR_JOINT_VEL_MAX[5];
    printf("WARN joint q6 is saturated \n");
  } else {
    q_new->q6 = q_goto->q6;
  }
  if(fabs(q_goto->q7 - q_old->q7)/dt > LWR_JOINT_VEL_MAX[6]) {
    q_new->q7 = q_old->q7 + SIGN(q_goto->q7 - q_old->q7)*dt*LWR_JOINT_VEL_MAX[6];
    printf("WARN joint q7 is saturated \n");
  } else {
    q_new->q7 = q_goto->q7;
  }
  return 0;
}

int lwrXOpenLoopCntrl( LWR_ARM_INST* armInst, GENMANIP_MOTION_STR *curState, LWR_MOTION_STR *motion, float *commandJntVals) {
  printf(" lwrXOpenLoopCntrl \n");
  std::vector<SM_COND> cond;
  Gb_dep dep_R7_cur_R7;
  Gb_th th_R7_cur_R7, th_R0_R7;
  Gb_q7 q;

  // Get the next conditions in the cartesian trajectory
  motion->traj_x.getMotionCond(motion->curTime, cond);

  // compute the R7 pose int the R0 frame
  dep_R7_cur_R7.x =  cond[armInst->indexCartInTraj].x;
  dep_R7_cur_R7.y =  cond[armInst->indexCartInTraj + 1].x;
  dep_R7_cur_R7.z =  cond[armInst->indexCartInTraj + 2].x;
  Gb_v3 rel_rot, rot;
  rel_rot.x = cond[armInst->indexCartInTraj + 3].x;
  rel_rot.y = cond[armInst->indexCartInTraj + 4].x;
  rel_rot.z = cond[armInst->indexCartInTraj + 5].x;
  double angle =  Gb_v3_norme ( &rel_rot, &rot);
  dep_R7_cur_R7.rx = rot.x;
  dep_R7_cur_R7.ry = rot.y;
  dep_R7_cur_R7.rz = rot.z;
  dep_R7_cur_R7.a = angle;
  Gb_dep_print(&dep_R7_cur_R7);
  Gb_dep_th(&dep_R7_cur_R7, &th_R7_cur_R7);
  Gb_th_produit(&(motion->traj_x_init_R0_R7) , &th_R7_cur_R7, &th_R0_R7);

  // compute MGI
  if(lwrComputeMGI(&(armInst->currConf), &th_R0_R7, &q) == 1) {
    return 1;
  }

  if(approachingBoundsForArm(armInst->currConf, q)) {
    printf("WARN: approachingBoundsForArm %d\n", armInst->isLeft);
    commandJntVals[0] =  armInst->currConf.q1;
    commandJntVals[1] =  armInst->currConf.q2;
    commandJntVals[2] =  armInst->currConf.q3;
    commandJntVals[3] =  armInst->currConf.q4;
    commandJntVals[4] =  armInst->currConf.q5;
    commandJntVals[5] =  armInst->currConf.q6;
    commandJntVals[6] =  armInst->currConf.q7;
  } else {
    Gb_q7 q_acceptable;
    // Check if new q is acceptable according to the robot's kinematic constraints
    lwrCheckAndSmoothArmConfigEvolution(&armInst->currConf, &q, LWR_PERIOD, &q_acceptable);
    commandJntVals[0] =  q_acceptable.q1;
    commandJntVals[1] =  q_acceptable.q2;
    commandJntVals[2] =  q_acceptable.q3;
    commandJntVals[3] =  q_acceptable.q4;
    commandJntVals[4] =  q_acceptable.q5;
    commandJntVals[5] =  q_acceptable.q6;
    commandJntVals[6] =  q_acceptable.q7;
  }
  return 0;
}

int lwrTrajectoryCntrl(int offsetArm,  LWR_ARM_INST* armInst, GENMANIP_MOTION_STR *curState, LWR_MOTION_STR *motion, float *commandJntVals){
  double commandJntDes[LWR_JOINTS];

  double timeSwitchMode;

  std::vector<SM_COND> condDes;
  std::vector<SM_COND> condCur;


  double criticalLength;
  SM_MOTION_MONO motionMono[LWR_JOINTS];
  SM_LIMITS limitsGoto;
  int TrajectoryType[LWR_JOINTS];
  SM_TIMES T_Jerk[LWR_JOINTS];
  double maxTime = 0.0;
  double sum = 0.0;
  double maxAxis = 0;

  double jerkNeed[LWR_JOINTS]; 


  double ratioJerkAllowed = 1.5;

  lwrGetCondFromGENMANIP_MOTION_STR(curState, condCur);
  motion->traj.getMotionCond(motion->curTime, condDes);

  maxTime = 0.0;
  sum = 0.0;
  maxAxis = 0;


  /* check if condDes is reachable within one tic according the limits */
  for(unsigned int i=0; i<LWR_JOINTS; i++) {
    jerkNeed[i] = (condDes[i+ offsetArm].a -  condCur[i+ offsetArm].a) / ((double)LWR_PERIOD*2.0);
    if( ABS(jerkNeed[i]) > ratioJerkAllowed*ABS( motion->limits_q[i+ offsetArm].maxJerk) ) {
      //std::cout << " WARNING JERK EXCEED (" << i << "){jmax= " <<  motion->q_J_max[i+ offsetArm] << " , jneed= " << jerkNeed[i] << "}"  << " condDes.a: " <<  condDes[i+ offsetArm].a << " condCur.a: " << condCur[i+ offsetArm].a << std::endl;;

    }
  }

  //for(unsigned int i=0 i<LWR_JOINTS; i++) {
  //  limitsGoto.maxJerk = motion->q_J_max[i+ offsetArm];
  //  limitsGoto.maxAcc  = motion->q_A_max[i+ offsetArm];
  //  limitsGoto.maxVel  = motion->q_V_max[i+ offsetArm];
  //
  //  SM_STATUS resp = sm_ComputeSoftMotion(condCur[i+ offsetArm], condDes[i+ offsetArm], limitsGoto, &(T_Jerk[i]), &(TrajectoryType[i]));
  //  if (resp != SM_OK) {
  //    printf("ERROR sm_ComputeSoftMotion\n");
  //  }
  //  sm_sum_motionTimes(&(T_Jerk[i]), &sum);
  //  
  //  if(sum > maxTime) {
  //    maxTime = sum;
  //    maxAxis = i;
  //  }
  //}
  //
  //for(int i=0; i<LWR_JOINTS_USER; i++) {
  //  sm_adjustMotionWith3seg(motion[i].IC, motion[i].FC, maxTime, &(motionMono[i]));
  //}


  



  for (int i = 0; i < LWR_JOINTS; i++) {
    commandJntVals[i] = condDes[i + offsetArm].x;
    armInst->currentCmdVelJntVals[i] = condDes[i + offsetArm].v;
    armInst->currentCmdAccJntVals[i] = condDes[i + offsetArm].a;
  }


  

  return 0;
}

int lwrGoalCntrl(int offsetArm, GENMANIP_MOTION_STR *curState, LWR_MOTION_STR *motion, float *commandJntVals){


  return 0;
}

int lwrJntTrqCntrlPosition(double force, double factor, double maxForce,SM_COND cond, float* jntVal) {
  double step = 0.0;
  step = ABS(force)*0.000375*factor  - 0.001;
  if(ABS(force)> maxForce) {
    step =  maxForce*0.000375*factor  - 0.001;
  }
  if(force >= 0.0) {
    *jntVal =  cond.x - step;
  } else {
    *jntVal =  cond.x + step;
  }  
  return 0;
}

int lwrJntTrqCntrlSoftMotion(SM_LIMITS limitsGoto, double force, double maxForce, SM_COND cond, float* jntVal, float *jntVel, float *jntAcc) {
  double velocity = 0.0, dc = 0.0;
  SM_COND IC, condDes;
  int dir =0;
  SM_TIMES T_Jerk;
  SM_STATUS resp;

  std::vector<double> I(3);
  std::vector<double> T(SM_NB_SEG);
  std::vector<double> J(SM_NB_SEG);
  std::vector<double> t(1);
  std::vector<double> a(1);
  std::vector<double> v(1);
  std::vector<double> x(1);

  if(ABS(force) > ABS(maxForce)) {
    if(force >= 0.0) {
      velocity = -limitsGoto.maxVel;
    } else {
      velocity = limitsGoto.maxVel;
    }
  } else {
    if(force >= 0.0) {
      velocity = - ABS(force * limitsGoto.maxVel / ABS(maxForce));
    } else {
      velocity = ABS(force * limitsGoto.maxVel / ABS(maxForce));
    }
  }
  condDes.a = 0.0;
  condDes.v = velocity;
  condDes.x = 0.0;
  
  IC.a = cond.a;
  IC.v = cond.v;
  IC.x = 0.0;

  /* compute the critical length */
  sm_CalculOfCriticalLength(IC, condDes, limitsGoto, &dc);
  if (isnan(dc)) {
    printf("isnan dc\n");
    printf("IC.a = %f    IC.v = %f    IC.x = %f\n",cond.a,cond.v,cond.x);
    printf("FC.a = %f    FC.v = %f    FC.x = %f\n",condDes.a, condDes.v,condDes.x);
    *jntVal = cond.x;
    *jntVel = 0.0;
    *jntAcc = 0.0;
    return 0;
  }
  condDes.x = dc;

  /* compute the motion */
  resp = sm_ComputeSoftMotion(IC, condDes, limitsGoto, &T_Jerk, &dir);
  if (resp != SM_OK) {
    printf("ERROR sm_ComputeSoftMotion\n");
  }

  /* get the  position at the next tick */       
  I[0] = cond.a;
  I[1] = cond.v;
  I[2] = cond.x;

  T[0] = T_Jerk.Tjpa;
  T[1] = T_Jerk.Taca;
  T[2] = T_Jerk.Tjna;
  T[3] = T_Jerk.Tvc;
  T[4] = T_Jerk.Tjnb;
  T[5] = T_Jerk.Tacb;
  T[6] = T_Jerk.Tjpb;

  J[0] =   dir*limitsGoto.maxJerk;
  J[1] =   0.0;
  J[2] = - dir*limitsGoto.maxJerk;
  J[3] =   0.0;
  J[4] = - dir*limitsGoto.maxJerk;
  J[5] =   0.0;
  J[6] =   dir*limitsGoto.maxJerk;

  t[0] = (double)LWR_PERIOD;    
  resp = sm_AVX_TimeVar(I, T, J, t, a, v, x);

  *jntVal = x[0];
  *jntVel  = v[0];
  *jntAcc  = a[0];

  return 0;
}


int lwrJntTrqCntrl(GEN_BOOL isTrqCntrlSofMotion, int offsetArm,  LWR_ARM_INST* armInst, GENMANIP_MOTION_STR *curState, LWR_MOTION_STR *motion, float *commandJntVals){

  std::vector<SM_COND> condCur;
  double velocity = 0.0;
  double coef = 2.0;
  double minForce[LWR_JOINTS];
  double maxForce = 20.0;
  SM_LIMITS limitsGoto;
  double force = 0.0;
  double factor[LWR_JOINTS];


  minForce[0] = 1.0;
  minForce[1] = 1.0;
  minForce[2] = 1.0;
  minForce[3] = 1.0;
  minForce[4] = 0.5;
  minForce[5] = 0.5;
  minForce[6] = 0.5;


  factor[0] = coef * 1.0;
  factor[1] = coef * 1.0;
  factor[2] = coef * 1.0;
  factor[3] = coef * 3.0;
  factor[4] = coef * 10.0;
  factor[5] = coef * 20.0;
  factor[6] = coef * 40.0;

  int nbJointStabilized = 0;
  lwrGetCondFromGENMANIP_MOTION_STR(curState, condCur);

  for(unsigned int i=0 ; i<LWR_JOINTS; i++) {    

    force = armInst->estMeanExtJntTrq[i];

    // if(ABS( armInst->desJntTrq[i] - armInst->estExtJntTrq[i] ) < ABS(minForce[i]) &&  ABS(condCur[i+ offsetArm].v) < 0.01) {

    if(ABS( armInst->desJntTrq[i] - force ) < ABS(minForce[i]) ) {

      commandJntVals[i] = condCur[i+ offsetArm].x;
      armInst->currentCmdVelJntVals[i] = 0.0;
      armInst->currentCmdAccJntVals[i] = 0.0;      

      nbJointStabilized ++;
    } else {

      if(isTrqCntrlSofMotion == GEN_FALSE) {

        //step = ABS(force)*0.000375*factor[i]  - 0.001;
        //
        //if(ABS(force)> maxForce) {
        //  step =  maxForce*0.000375*factor[i]  - 0.001;
        //}
        //if(force >= 0.0) {
        //  commandJntVals[i] =  condCur[i+ offsetArm].x - step;
        //} else {
        //  commandJntVals[i] =  condCur[i+ offsetArm].x + step;
        //}
        lwrJntTrqCntrlPosition(force, factor[i], maxForce, condCur[i+ offsetArm], &(commandJntVals[i]));

      } else { 

        limitsGoto.maxJerk = 2.0 * motion->limits_q[i+ offsetArm].maxJerk;
        limitsGoto.maxAcc  = 2.0 * motion->limits_q[i+ offsetArm].maxAcc;
        limitsGoto.maxVel  =       motion->limits_q[i+ offsetArm].maxVel;

        lwrJntTrqCntrlSoftMotion(limitsGoto, force, maxForce, condCur[i+ offsetArm], &(commandJntVals[i]), &(armInst->currentCmdVelJntVals[i]), &(armInst->currentCmdAccJntVals[i]));
      }
    }
  }

  if(nbJointStabilized == LWR_JOINTS) {
    printf("INFO: Arm is stabilized\n");
    armInst->isStabilized = GEN_TRUE;
  } else {
    armInst->isStabilized = GEN_FALSE;
  }


  return 0;
}


int lwrGetCondFromGENMANIP_MOTION_STR(GENMANIP_MOTION_STR *curState, std::vector<SM_COND> &cond) {
  SM_COND cond_i;
  cond.clear();
  for(int i=0; i<curState->size; i++){
    cond_i.a = curState->aq[i];
    cond_i.v = curState->vq[i];    
    cond_i.x = curState->q[i];
    cond.push_back(cond_i);
  }
  return 0;
}

/********************************************************/
/*                                                      */
/*               FUNCTIONS SPECIFIC TO FRI              */
/*                                                      */
/********************************************************/
#ifdef LWR_FRI

void switchCtrlScheme(friRemote * friInst, LWR_CTRL_SCHEME_ENUM scheme)
{
  switch (scheme) {
  case LWR_CTRL_POSITION:
    friInst->setToKRLInt(CMD_CHANNEL, POSIT_CTRL);
    break;
  case LWR_CTRL_CART_IMP:
    friInst->setToKRLInt(CMD_CHANNEL, CART_CTRL);
    break;
  case LWR_CTRL_JOINT_IMP:
    friInst->setToKRLInt(CMD_CHANNEL, AXIS_CTRL);
    break;
  case LWR_CTRL_GRAV_COMP:
    friInst->setToKRLInt(CMD_CHANNEL, GRAV_COMP);
    break;
  default:
    printf("ERROR: Unrecogniced control scheme: %d. Ignoring command\n", scheme);
    break;
  }
}

const char * friStateStr(int state)
{
  switch (state) {
  case FRI_STATE_OFF:
    return "inactive";
  case FRI_STATE_MON:
    return "monitor";
  case FRI_STATE_CMD:
    return "command";
  default:
    return "ERROR";
  }
}

const char * friSchemeStr(int scheme)
{
  switch (scheme) {
  case LWR_CTRL_POSITION:
    return "joint position";
  case LWR_CTRL_CART_IMP:
    return "cartesian imp.";
  case LWR_CTRL_JOINT_IMP:
    return "axis impedance";
  case LWR_CTRL_GRAV_COMP:
    return "gravity comp.";
  default:
    return "ERROR";
  }
}

const char * friQualityStr(int quality)
{
  switch (quality) {
  case FRI_QUALITY_UNACCEPTABLE:
    return "unacceptable";
  case FRI_QUALITY_BAD:
    return "bad";
  case FRI_QUALITY_OK:
    return "ok";
  case FRI_QUALITY_PERFECT:
    return "perfect";
  default:
    return "ERROR";
  }
}

/*
 * Synchronize FRI commanded position/configuration with the measured one. This
 * is required when switching between position and cartesian control schemes. If
 * not in FRI monitor state, this can cause a FRI interpolation error.
 */
void syncFriWithMeasuredPose(friRemote * friInst, LWR_CTRL_SCHEME_ENUM newScheme)
{
  static int first = 0;


  if (first < 2) {
    //printf("******************************* FIRST ************************\n");
    if (newScheme == LWR_CTRL_CART_IMP) {
      friInst->doCartesianImpedanceControl(friInst->getMsrCartPosition(), NULL, NULL, NULL, NULL, false);
    } else {
      friInst->doPositionControl(friInst->getMsrMsrJntPosition(), false);
    }
    first++;// = false;
    return;
  }
  LWR_CTRL_SCHEME_ENUM currentScheme = (LWR_CTRL_SCHEME_ENUM)friInst->getCurrentControlScheme();

  if (newScheme != currentScheme) {
    if ((newScheme == LWR_CTRL_POSITION) || (newScheme == LWR_CTRL_JOINT_IMP)) {
      friInst->doPositionControl(friInst->getMsrMsrJntPosition(), false);
    }
    else
      if (newScheme == LWR_CTRL_CART_IMP) {
        friInst->doCartesianImpedanceControl(friInst->getMsrCartPosition(),
                                             NULL, NULL, NULL, NULL, false);
      }
  }
  else
    /*
 if ((ctrlScheme == LWR_CTRL_GRAV_COMP) && (friInst->getCurrentControlScheme() == LWR_CTRL_CART_IMP))    {
 friInst->doCartesianImpedanceControl(friInst->getMsrCartPosition(), NULL, NULL, NULL, NULL, false);
 printf("*******************************CARTESIAN************************\n");}
 else
 if ((ctrlScheme == LWR_CTRL_GRAV_COMP) &&
        ((friInst->getCurrentControlScheme() == LWR_CTRL_POSITION) ||
 (friInst->getCurrentControlScheme() == LWR_CTRL_JOINT_IMP)))         {
 friInst->doPositionControl(friInst->getMsrMsrJntPosition(), false);
 printf("*******************************POSITION************************\n");}
 else
      */
    // TODO esto nuevo por si solo podria valer.....
    if (((newScheme == LWR_CTRL_POSITION) || (newScheme == LWR_CTRL_JOINT_IMP)) && (jointMaxDiff(friInst) > 0.0001)) {
      friInst->doPositionControl(friInst->getMsrMsrJntPosition(), false);
      //printf("*******************************POSITION 2 %f ************************\n", jointMaxDiff());
    }
    else {
      if ((newScheme == LWR_CTRL_CART_IMP) && (cartMaxDiff(friInst) > 0.0001)) {
        friInst->doCartesianImpedanceControl(friInst->getMsrCartPosition(), NULL, NULL, NULL, NULL, false);
        //	printf("*******************************CARTESIAN 2 %f ************************\n",cartMaxDiff());
      }
    }

  return;
}

float cartMaxDiff(friRemote * friInst)
{
  float maxDiff = 0.0;
  float cmd[FRI_CART_FRM_DIM];

  float* msr = friInst->getMsrCartPosition();
  //  float* cmd = friInst->getMsrCmdCartPosition();
  friInst->getCurrentCmdCartPosition(cmd);

  for (int i = 0; i < FRI_CART_FRM_DIM; i++) {
    //printf("%f / %f,  ", msr[i],cmd[i]);
    float diff = fabs(msr[i] - cmd[i]);
    if (diff > maxDiff)
      maxDiff = diff;
  }
  //printf("  ### %f ###   ", maxDiff);
  return maxDiff;
}

float jointMaxDiff(friRemote * friInst)
{
  float maxDiff = 0.0;

  float* msr = friInst->getMsrMsrJntPosition();
  float* cmd = friInst->getMsrCmdJntPosition();
  //friInst->getCurrentCmdJntPosition(cmd);

  for (int i = 0; i < LWR_JOINTS; i++) {
    float diff = fabs(msr[i] - cmd[i]);
    if (diff > maxDiff)
      maxDiff = diff;
  }
  return maxDiff;
}

double lwrComputeDistToEndTraj(LWR_ARM_INST *armInst) {
  

  double dx, dy, dz, dist;
  Gb_q7 endConf;
  Gb_th thCurr, thEnd;
  std::vector<SM_COND> cond;
  
  /* conpute the distance between the collision pose and the goal pose */
  if(armInst->cntrl_mode == LWR_TRAJECTORY_CNTRL || armInst->cntrl_mode == LWR_OPEN_LOOP_CNTRL){
    
    currentMotion.traj.getMotionCond(currentMotion.traj.getDuration(), cond);
    
    kukaLBR_mgd(&armInst->currConf, LWR_R3, LWR_R5, &thCurr);
    
    endConf.q1 = cond[0 + armInst->indexInTraj].x;
    endConf.q2 = cond[1 + armInst->indexInTraj].x;
    endConf.q3 = cond[2 + armInst->indexInTraj].x;
    endConf.q4 = cond[3 + armInst->indexInTraj].x;
    endConf.q5 = cond[4 + armInst->indexInTraj].x;
    endConf.q6 = cond[5 + armInst->indexInTraj].x;
    endConf.q7 = cond[6 + armInst->indexInTraj].x;
    kukaLBR_mgd(&endConf, LWR_R3, LWR_R5, &thEnd);
    
    dx =  thCurr.vp.x - thEnd.vp.x;
    dy =  thCurr.vp.y - thEnd.vp.y; 
    dz =  thCurr.vp.z - thEnd.vp.z;
    
    dist = sqrt(dx*dx + dy*dy + dz*dz);
    
  } else {
    printf("WARN lwrComputeDistToEndTraj cannot compute the distance with the actual mode\n");
    dist = -1;
  }
  
  return dist;
}
#endif

// Digest is printed in coldman-genom viewer
int lwrGenerateDigestStatus() {
#ifdef LWR_FRI
  // Prepare an human-readable resume of robot status and configuration
  if(SDI_F->armLeft.is_used == GEN_TRUE && SDI_F->armRight.is_used == GEN_TRUE ) {
    snprintf(SDI_F->status.status_str.str, 1024,
             "Simulation:\t%s\n"			    \
             "Arm power:\tLeft: %s\tRight: %s\n"		\
             "FRI status:\tLeft: %s\tRight: %s\n"		\
             "Control scheme:\tLeft: %s\tRight: %s\n"	\
             "Link quality:\tLeft: %s\tRight:%s\n"		     \
             "Configuration\n\t"					      \
             "q1:\t%3.2f\n\tq2:\t%3.2f\n\tq3:\t%3.2f\n\tq4:\t"		\
             "%3.2f\n\tq5:\t%3.2f\n\tq6:\t%3.2f\n\tq7:\t%3.2f\n"        \
             "Fx: %4.2f Fy: %4.2f Fz: %4.2f Mx: %4.2f My: %4.2f Mz: %4.2f",
             SDI_F->status.simulation?"yes":"no",
             SDI_F->armLeft.power_on?"on":"off",SDI_F->armRight.power_on?"on":"off",
             friStateStr(SDI_F->armLeft.fri_state), friStateStr(SDI_F->armRight.fri_state),
             friSchemeStr(SDI_F->armLeft.fri_scheme), friSchemeStr(SDI_F->armRight.fri_scheme),
             friQualityStr(SDI_F->armLeft.fri_quality),friQualityStr(SDI_F->armRight.fri_quality),
             rad2deg(SDI_F->motion. q[0]), rad2deg(SDI_F->motion. q[1]),
             rad2deg(SDI_F->motion. q[2]), rad2deg(SDI_F->motion. q[3]),
             rad2deg(SDI_F->motion. q[4]), rad2deg(SDI_F->motion. q[5]),
             rad2deg(SDI_F->motion. q[6]),
             SDI_F->armLeft.estExtTcpFT[0], SDI_F->armLeft.estExtTcpFT[1], SDI_F->armLeft.estExtTcpFT[2],
             SDI_F->armLeft.estExtTcpFT[3], SDI_F->armLeft.estExtTcpFT[4], SDI_F->armLeft.estExtTcpFT[5]);
  } else if( SDI_F->armLeft.is_used == GEN_TRUE ) {
    snprintf(SDI_F->status.status_str.str, 1024,
             "Simulation:\t%s\n"			    \
             "Arm power:\tLeft: %s\n"	    \
             "FRI status:\tLeft: %s\n"		\
             "Control scheme:\tLeft: %s\n"	\
             "Link quality:\tLeft: %s\n"		     \
             "Configuration\n\t"					      \
             "q1:\t%3.2f\n\tq2:\t%3.2f\n\tq3:\t%3.2f\n\tq4:\t"		\
             "%3.2f\n\tq5:\t%3.2f\n\tq6:\t%3.2f\n\tq7:\t%3.2f\n"        \
             "Fx: %4.2f Fy: %4.2f Fz: %4.2f Mx: %4.2f My: %4.2f Mz: %4.2f",
             SDI_F->status.simulation?"yes":"no",
             SDI_F->armLeft.power_on?"on":"off",
             friStateStr(SDI_F->armLeft.fri_state),
             friSchemeStr(SDI_F->armLeft.fri_scheme),
             friQualityStr(SDI_F->armLeft.fri_quality),
             rad2deg(SDI_F->motion. q[0]), rad2deg(SDI_F->motion. q[1]),
             rad2deg(SDI_F->motion. q[2]), rad2deg(SDI_F->motion. q[3]),
             rad2deg(SDI_F->motion. q[4]), rad2deg(SDI_F->motion. q[5]),
             rad2deg(SDI_F->motion. q[6]),
             SDI_F->armLeft.estExtTcpFT[0], SDI_F->armLeft.estExtTcpFT[1], SDI_F->armLeft.estExtTcpFT[2],
             SDI_F->armLeft.estExtTcpFT[3], SDI_F->armLeft.estExtTcpFT[4], SDI_F->armLeft.estExtTcpFT[5]);
  } else if ( SDI_F->armRight.is_used == GEN_TRUE ) {
    snprintf(SDI_F->status.status_str.str, 1024,
             "Simulation:\t%s\n"			    \
             "Arm power:\tRight: %s\n"	    \
             "FRI status:\tRight: %s\n"		\
             "Control scheme:\tRight: %s\n"	\
             "Link quality:\tRight: %s\n"		     \
             "Configuration\n\t"					      \
             "q1:\t%3.2f\n\tq2:\t%3.2f\n\tq3:\t%3.2f\n\tq4:\t"		\
             "%3.2f\n\tq5:\t%3.2f\n\tq6:\t%3.2f\n\tq7:\t%3.2f\n"        \
             "Fx: %4.2f Fy: %4.2f Fz: %4.2f Mx: %4.2f My: %4.2f Mz: %4.2f",
             SDI_F->status.simulation?"yes":"no",
             SDI_F->armRight.power_on?"on":"off",
             friStateStr(SDI_F->armRight.fri_state),
             friSchemeStr(SDI_F->armRight.fri_scheme),
             friQualityStr(SDI_F->armRight.fri_quality),
             rad2deg(SDI_F->motion. q[LWR_JOINTS+0]), rad2deg(SDI_F->motion. q[LWR_JOINTS+1]),
             rad2deg(SDI_F->motion. q[LWR_JOINTS+2]), rad2deg(SDI_F->motion. q[LWR_JOINTS+3]),
             rad2deg(SDI_F->motion. q[LWR_JOINTS+4]), rad2deg(SDI_F->motion. q[LWR_JOINTS+5]),
             rad2deg(SDI_F->motion. q[LWR_JOINTS+6]),
             SDI_F->armRight.estExtTcpFT[0], SDI_F->armRight.estExtTcpFT[1], SDI_F->armRight.estExtTcpFT[2],
             SDI_F->armRight.estExtTcpFT[3], SDI_F->armRight.estExtTcpFT[4], SDI_F->armRight.estExtTcpFT[5]);
  }
#else
  if(SDI_F->armLeft.is_used == GEN_TRUE || SDI_F->armRight.is_used == GEN_TRUE ) {
    // Prepare an human-readable resume of robot status and configuration
    snprintf(SDI_F->status.status_str.str, 1024,
             "Simulation:\t%s\n"			    \
             "Configuration armLeft:\n\t"					\
             "q1:\t%3.2f\n\tq2:\t%3.2f\n\tq3:\t%3.2f\n\tq4:\t"		\
             "%3.2f\n\tq5:\t%3.2f\n\tq6:\t%3.2f\n\tq7:\t%3.2f\n"
             "Configuration armRight:\n\t"					\
             "q1:\t%3.2f\n\tq2:\t%3.2f\n\tq3:\t%3.2f\n\tq4:\t"		\
             "%3.2f\n\tq5:\t%3.2f\n\tq6:\t%3.2f\n\tq7:\t%3.2f\n",
             "yes",
             rad2deg(SDI_F->motion. q[0]), rad2deg(SDI_F->motion. q[1]),
             rad2deg(SDI_F->motion. q[2]), rad2deg(SDI_F->motion. q[3]),
             rad2deg(SDI_F->motion. q[4]), rad2deg(SDI_F->motion. q[5]),
             rad2deg(SDI_F->motion. q[6]),
             rad2deg(SDI_F->motion. q[LWR_JOINTS+0]), rad2deg(SDI_F->motion. q[LWR_JOINTS+1]),
             rad2deg(SDI_F->motion. q[LWR_JOINTS+2]), rad2deg(SDI_F->motion. q[LWR_JOINTS+3]),
             rad2deg(SDI_F->motion. q[LWR_JOINTS+4]), rad2deg(SDI_F->motion. q[LWR_JOINTS+5]),
             rad2deg(SDI_F->motion. q[LWR_JOINTS+6])
             );
  } else if( SDI_F->armLeft.is_used == GEN_TRUE ) {
    snprintf(SDI_F->status.status_str.str, 1024,
             "Simulation:\t%s\n"			    \
             "Configuration arm Left\n\t"					\
             "q1:\t%3.2f\n\tq2:\t%3.2f\n\tq3:\t%3.2f\n\tq4:\t"		\
             "%3.2f\n\tq5:\t%3.2f\n\tq6:\t%3.2f\n\tq7:\t%3.2f\n",
             "yes",
             rad2deg(SDI_F->motion. q[0]), rad2deg(SDI_F->motion. q[1]),
             rad2deg(SDI_F->motion. q[2]), rad2deg(SDI_F->motion. q[3]),
             rad2deg(SDI_F->motion. q[4]), rad2deg(SDI_F->motion. q[5]),
             rad2deg(SDI_F->motion. q[6]));

  } else if ( SDI_F->armRight.is_used == GEN_TRUE ) {
    snprintf(SDI_F->status.status_str.str, 1024,
             "Simulation:\t%s\n"			    \
             "Configuration armRight\n\t"					\
             "q1:\t%3.2f\n\tq2:\t%3.2f\n\tq3:\t%3.2f\n\tq4:\t"		\
             "%3.2f\n\tq5:\t%3.2f\n\tq6:\t%3.2f\n\tq7:\t%3.2f\n",
             "yes",
             rad2deg(SDI_F->motion. q[LWR_JOINTS+0]), rad2deg(SDI_F->motion. q[LWR_JOINTS+1]),
             rad2deg(SDI_F->motion. q[LWR_JOINTS+2]), rad2deg(SDI_F->motion. q[LWR_JOINTS+3]),
             rad2deg(SDI_F->motion. q[LWR_JOINTS+4]), rad2deg(SDI_F->motion. q[LWR_JOINTS+5]),
             rad2deg(SDI_F->motion. q[LWR_JOINTS+6]));
  }
#endif
  return 0;
}

int lwrComputeMGI(Gb_q7* currConf, Gb_th *gotoTh, Gb_q7 *gotoConf)
{
  int e1, e2, e3;
  kukaLBR_gete1e2e3(LWR_R3, LWR_R5, currConf, &e1, &e2, &e3);


  Gb_statusMGI result =
      kukaLBR_mgi_q_e(gotoTh, currConf, LWR_R3, LWR_R5, LWR_MGI_EPSILON,
                      e1, e2, e3, gotoConf);

  switch (result) {
  case MGI_ERROR:
    printf("WARN: MGI calculation failed. Coldman model is probably wrong\n");
    return 1;
    break;
  case MGI_SINGULAR:
    printf("WARN: Singularity on MGI calculation. Colman model is probably wrong\n");
    return 2;
    break;
  case MGI_APPROXIMATE:
    printf("WARN: Approximate MGI calculation. Colman model can be inexact\n");
    return 3;
    break;
  case MGI_OK:
    //printf("INFO: MGI OK\n");
    ;
  }
  return 0;
}

int lwrComputeMGIQ3Opt(Gb_q7* currConf, Gb_th *gotoTh, Gb_q7 *gotoConf)
{
  int e1, e2, e3;
  kukaLBR_gete1e2e3(LWR_R3, LWR_R5, currConf, &e1, &e2, &e3);


  Gb_q7 QMax;
  Gb_q7 QMin;

  QMax.q1 = LWR_A1_RANGE_MAX;
  QMax.q2 = LWR_A2_RANGE_MAX;
  QMax.q3 = LWR_E1_RANGE_MAX;
  QMax.q4 = LWR_A3_RANGE_MAX;
  QMax.q5 = LWR_A4_RANGE_MAX;
  QMax.q6 = LWR_A5_RANGE_MAX;
  QMax.q7 = LWR_A6_RANGE_MAX;


  QMin.q1 = LWR_A1_RANGE_MIN;
  QMin.q2 = LWR_A2_RANGE_MIN;
  QMin.q3 = LWR_E1_RANGE_MIN;
  QMin.q4 = LWR_A3_RANGE_MIN;
  QMin.q5 = LWR_A4_RANGE_MIN;
  QMin.q6 = LWR_A5_RANGE_MIN;
  QMin.q7 = LWR_A6_RANGE_MIN;
  
  double gain = 0.05;
  int nbMaxLoop = 10;


  Gb_statusMGI result =
      kukaLBR_mgi_q3midle_e(gotoTh, currConf, LWR_R3, LWR_R5, &QMax, &QMin, gain, nbMaxLoop, LWR_MGI_EPSILON,
                            e1, e2, e3, gotoConf);

  switch (result) {
  case MGI_ERROR:
    printf("WARN: MGI calculation failed. Coldman model is probably wrong\n");
    return 1;
    break;
  case MGI_SINGULAR:
    printf("WARN: Singularity on MGI calculation. Colman model is probably wrong\n");
    return 2;
    break;
  case MGI_APPROXIMATE:
    printf("WARN: Approximate MGI calculation. Colman model can be inexact\n");
    return 3;
    break;
  case MGI_OK:
    //printf("INFO: MGI OK\n");
    ;
  }
  return 0;
}
