/*
 * Copyright (c) 2010 LAAS/CNRS                       --  Oct 2010
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
/*------------------------------------------------------------
 * 
 *                     --  Module  LWR  --
 * 
 *  Description:
 *  Creation date : Wed May 19 22:35:29 2010 
 *  Author: Xavier Broquere
 *  LAAS-CNRS, Toulouse FRANCE
 *------------------------------------------------------------*/

#ifndef LWR_FUNCTION_STRUCT_H
#define LWR_FUNCTION_STRUCT_H

#include "genManipStruct.h"

typedef struct LWR_MOTION_STR {
  LWR_MOTION_TYPE_ENUM  type;
  LWR_MOTION_STATE_ENUM state;
  LWR_MOTION_ERROR_ENUM error;

  double duration;

  double curTime;
  SM_COND timeScaleCond;
  int trajId;

  // for articular trajectory
  SM_TRAJ traj;

  // for cartesian trajectory
  // traj_x describes the evolution of a virtual point
  SM_TRAJ traj_x;
  // since traj_x is relative to the initital pose, we store the initial pose of the trajectory in the R0 frame --> MGI is compute between R0 and R7
  Gb_th traj_x_init_R0_R7;
  // this matrix describes the transform between the virtual point of traj_x and the flange (R8 frame)
  // virtual point is the one of which we want to control the pose during the motion
  Gb_th traj_x_R8_Rtraj;

  // for the closed-loop control in cartesian space
  Gb_th th_R0_R7_Goto; 

  // Per-joint and cartesian velocity, acceleration and jerk limits
  std::vector<SM_LIMITS> limits_q;
  std::vector<SM_LIMITS> limits_x;
} LWR_MOTION_STR;

#endif
